#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'mixlib/authentication/signatureverification'

#
# => Base controller. All controlers inherits from this.
#    It controls several validations like certs, user setup,
#    licenses....
#
class ApplicationController < ActionController::Base
  include Redborder::Rbdruid
  include Redirections
  helper Redborder::MenuManager::MenuHelper

  #
  # => Prevent CSRF attacks by raising an exception.
  #    For APIs, you may want to use :null_session instead.
  #
  protect_from_forgery with: :exception

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  around_action :switch_locale

  # Authentication
  before_action :configure_permitted_parameters, if: :devise_controller?

  #rescue_from DeviseLdapAuthenticatable::LdapException do |exception|
  #  render text: exception, status: 500
  #end

  if Rails.env.production?
    unless Rails.application.config.consider_all_requests_local
      rescue_from Exception, with: :handle_exception
    end
  end

  before_action :authenticate_user!
  before_action :user_setup
  before_action :set_timezone
  before_action :set_time_machine
  DittoCode::Exec.if 'ENTERPRISE' do
    before_action :get_alerts, if: proc { request.format.html? }
  end

  after_action :flash_to_headers
  before_action :init_zendesk_tab

  DittoCode::Exec.if 'ENTERPRISE' do
    before_action :check_blocked
  end

  before_action :set_global_filters

  ## ======================================================================
  # => Errors
  ## ======================================================================

  #
  # Redirect the user to a 404 page.
  #
  def not_found
    error_404("#{I18n.t('error_404')}")
  end

  #
  # => 500 error
  #
  def internal_server_error
    if user_signed_in?
      render template: 'errors/internal_server_error', status: 500
    else
      render template: 'errors/internal_server_error_nolayout',
             status: 500, layout: false
    end
  end

  ## =====================================================================
  # => Helper functions to show errors
  ## =====================================================================

  #
  # Define methods for some status codes. It will respond in the same
  # request format.
  #
  # == Parameters:
  # error::
  #   An string with the error to display
  #
  [400, 401, 403, 404, 422, 500, 503].each do |status|
    define_method("error_#{status}") do |error|
      respond_to do |format|
        format.html do
          template = case status
                     when 404 then 'errors/not_found'
                     when 403 then 'errors/forbidden'
                     else 'errors/internal_server_error'
                     end
          render template: template, status: status
        end
        format.js do
          render js: '', status: status
        end
        format.json do
          render json: {
            query: false,
            error: error
          }, status: status
        end
        format.csv { render nothing: true, status: status }
      end
    end
  end

  ## ======================================================================
  # => Methods and helpers
  ## ======================================================================

  protected

  def set_global_filters
    @global_filters = session[:global_filters] || {}
    @custom_global_filters =
      if File.exist?('config/global_filters.yml')
        YAML.load_file(
          Rails.root.join('config/global_filters.yml')
        ).symbolize_keys
      else
        {}
      end
  end

  #
  #
  #
  def should_skip_empty
    @should_skip_empty = params[:skip_empty] == '0' ? 'false' : 'true'
  end

  #
  # Init inactive settings to help the rendering of view.
  # At this state, a @source must be initialized.
  #
  def set_module_settings
    @inactive_settings = {}
    Setting.where('name LIKE ? AND value=\'0\'', "#{@source.id}_%").map do |s|
      @inactive_settings[s.name] = ''
    end
  end

  def configure_permitted_parameters
    devise_parameter_sanitizer.permit(:sign_in, keys: [:login, :username, :email, :password, :remember_me])
  end

  # The default url to be used after loging in
  #
  # == Parameters:
  # _resource::
  #   resource to be used to get the root path
  #
  # == Returns:
  #   path to be used after loging in user.
  #
  #
  def after_sign_in_path_for(_resource)
    sensors_num = User.current_user.sensor.children.count
    if current_user.admin? && sensors_num.zero?
      sensors_path
    else
      current_user.root_page
    end
  end

  #
  # Load filters from URL and store them if params[:search] or param[:filters]
  # exist.
  #
  # ==Parameters:
  # filter::
  #   Global filter must be initialized!
  #
  def filters_from_url
    [:search, :filters].each do |p|
      next if params[p].nil?
      # Parse filters
      @filter.parse(params[p], true)
      # Don't save if it's called from widget
      save_filters(@filter.filters) if params[:w].nil?
    end
  end

  def handle_exception(exception)
    Rails.logger.error exception.inspect
    Rails.logger.error exception.backtrace.join("\n")
    case exception
    when ActionController::RoutingError, ActionController::UnknownController,
         ActiveRecord::RecordNotFound
      not_found
    else
      internal_server_error
    end
  end

  #
  # Check if the current user is a super admin or redirect it
  #
  # == Parameters:
  # options::
  #   Hash with options. Available options:
  #     - redirect_to: path to redirect the user.
  #
  # == Returns:
  # Redirect the user if he isn't admin
  #
  def require_super_administrative_privileges(options = {})
    return true if user_signed_in? && current_user.super_admin?
    no_enough_privileges options
  end

  #
  # Check if the current user is a super reader or redirect it
  #
  # == Parameters:
  # options::
  #   Hash with options. Available options:
  #     - redirect_to: path to redirect the user.
  #
  # == Returns:
  # Redirect the user if he isn't super reader
  #
  def require_super_reader_privileges(options = {})
    return true if user_signed_in? && current_user.super_reader?
    no_enough_privileges options
  end

  #
  # Check if the current user is an admin or redirect it
  #
  # == Parameters:
  # options::
  #   Hash with options. Available options:
  #     - redirect_to: path to redirect the user.
  #
  # == Returns:
  # Redirect the user if he isn't admin
  #
  def require_administrative_privileges(options = {})
    return true if user_signed_in? && current_user.admin?
    no_enough_privileges options
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def check_blocked
      return unless ENV['BLOCKED'] == 'true'
      render file: "#{Rails.root}/public/license_failed.html", layout: false
    end
  end

  def init_zendesk_tab
    section = request.fullpath.split('/').delete_if(&:empty?).first
    return unless section != 'users' || request.fullpath == ''
    @zendesk_tab = true
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    #
    # => Gets active alerts
    #
    def get_alerts
      @alerts = Alert.active_alerts session[:dismissed_alerts]
    end
  end

  #
  # => Setup the user adding sessions and default configs
  #
  def user_setup
    return unless user_signed_in?

    if current_user.enabled
      User.current_user = current_user

      Redborder::Module.each_with_interface do |_id, properties|
        current_user.properties[properties[:session_id].to_sym] ||= {}
      end

      # set session filter for user
      Redborder::Module.keys_with_interface.each do |id|
        send('set_session_values_' + id.to_s)
      end
      session[:rule]          ||= { filter: {} }
      session[:object]        ||= :net
      session[:sensor_view]   ||= :tree
    else
      sign_out current_user
      redirect_to new_user_session_path,
                  notice: "#{I18n.t('account_disabled')}"
    end
  end

  #
  # => Set the current timezone
  #
  def set_timezone
    Time.zone = browser_timezone.nil? ? 'UTC' : browser_timezone
  end

  def set_time_machine
    if params[:time_machine_now].present?
      current_user
        .update(time_machine_now: Time.zone.at(params[:time_machine_now].to_i))
      current_user.reload
    elsif !params[:time_machine_now].nil? && params[:time_machine_now].blank?
      current_user.update(time_machine_now: nil)
      current_user.reload
    end
  end

  def in_time_machine(&block)
    if current_user.nil? || current_user.time_machine_now.nil?
      yield
    else
      Timecop.travel(Time.at(current_user.time_machine_now), &block)
    end
  end

  def validate_cert
    begin
      authenticator = Mixlib::Authentication::SignatureVerification.new(request)
      @username = authenticator.user_id
      user = Chef::ApiClient.load(@username)
      user_key = OpenSSL::PKey::RSA.new(user.public_key)
      @auth_user = user
      authenticator.authenticate_request(user_key)
    rescue Mixlib::Authentication::MissingAuthenticationHeader => e
      return error_401(e.message)
    rescue StandardError => se
      # raise Unauthorized, "Failed to authenticate.
      # Ensure that your client key is valid."
      return error_401(se.message)
    end

    if !authenticator.valid_request? && authenticator.valid_timestamp?
      # raise Unauthorized, "Failed to authenticate.
      # Ensure that your client key is valid."
      return error_401("#{I18n.t('error_401')}")
    elsif !authenticator.valid_request?
      # raise Unauthorized, "Failed to authenticate.
      # Please synchronize the clock on your client"
      return error_401("#{I18n.t('error_401')}")
    end
  end

  #
  # Return forbidden if the IP that make the request is not a private IP address.
  #
  def network_authentication
    request_ip = (request.headers['X-Forwarded-For'] ||
              request.headers['X-Real-IP'] ||
              request.remote_ip).split(',').first

    if IPAddress.valid_ipv4?(request_ip)
      ip_address = IPAddress::IPv4.new(request_ip)
    elsif IPAddress.valid_ipv6?(request_ip)
      ip_address = IPAddress::IPv6.new(request_ip)
    elsif request_ip == "::1"
      ip_address = IPAddress::IPv4.new("127.0.0.1")
    else
      return error_401("#{I18n.t('failed')}")
    end
    return error_401("#{I18n.t('failed')}") unless ip_address.private? || ip_address.loopback?
  end

  private

  def switch_locale(&action)
    if current_user
      I18n.locale = current_user.try(:preferred_locale) || I18n.default_locale
      I18n.with_locale(I18n.locale, &action)
    else
      I18n.with_locale(I18n.locale, &action)
    end
  end

  def flash_to_headers
    return unless request.xhr?
    msg = flash_message
    return unless msg

    response.headers['X-Message'] = msg
    response.headers['X-Message-Type'] = flash_type.to_s
    flash.discard
  end

  def flash_message
    [:error, :warning, :notice].each do |type|
      return flash[type] unless flash[type].blank?
    end

    nil
  end

  def flash_type
    [:error, :warning, :notice].each do |type|
      return type unless flash[type].blank?
    end

    nil
  end

  def set_range
    type = send('set_type_' + params[:controller].to_s)

    if params[:range] && params[:range] != 'custom'
      @filter.remove(:start_time)
      @filter.remove(:end_time)
      save_filters(@filter.filters) if params[:w].nil?
      @range = params[:range]
    elsif params[:range]
      if params[:range] == 'custom' && !@filter.exist_any?(:start_time) &&
         !@filter.exist_any?(:end_time)
        @range = 'last_1_hour'
      else
        @range = params[:range]
      end
    else
      @range = (@global_filters || {}).fetch(:range, 'last_1_hour')
    end

    end_time = start_time = Time.now

    in_time_machine do
      end_time = Time.now
      start_time = end_time - 1.hour
      if @filter.exist_any?(:start_time) && @filter.exist_any?(:end_time)
        @range = 'custom'

        start_time = Time.at(@filter.get(:start_time).keys.first)

        end_time = Time.at(@filter.get(:end_time).keys.first)
      end

      @end_time = Time.now.change(sec: 0)
    end

    case @range.to_sym
    when :live
      @start_time = @end_time - 3.minutes
    when :last_5_minute
      @start_time = @end_time - 5.minutes
    when :last_15_minute
      @start_time = @end_time - 15.minutes
    when :last_30_minute
      @start_time = @end_time - 30.minutes
    when :last_1_hour
      @start_time = @end_time - 1.hour
    when :last_2_hour
      @start_time = @end_time - 2.hour
    when :last_3_hour
      @start_time = @end_time - 3.hour
    when :last_12_hour
      @start_time = (@end_time - 12.hour).beginning_of_hour
    when :last_24
      @start_time = (@end_time - 24.hour).beginning_of_hour
    when :last_week
      @start_time = (@end_time - 1.week).beginning_of_hour
    when :custom
      @end_time   = end_time
      @start_time = start_time
    else
      @start_time = @end_time - 1.hour
    end

    @start_time_for_slider = @start_time
    @end_time_for_slider = @end_time

    granularities = Redborder::Rbdruid
                    .granularity_list(start_time: @start_time,
                                      end_time: @end_time)
    granularity = if params[:granularity].present?
                    params[:granularity]
                  else
                    granularities.first
                  end

    step = Redborder::Rbdruid::GRANULARITY_TYPE[granularity.to_sym][:step]
    roundto = step / 1000
    @start_time = Time.at((@start_time.to_i / roundto) * roundto)

    # Multiple return lines as granularity.to_sym should not be executed
    # if params[:ganularity] is not a valid one
    return if Redborder::Rbdruid::GRANULARITY_TYPE.keys.map(&:to_s).exclude?(granularity)
    return if granularity.to_sym.eql?(:all)

    
  end

  def set_chart_type
    send('default_chart_type_' + params[:controller].to_s)
  end

  def set_granularity
    granularities = Redborder::Rbdruid.granularity_list(start_time: @start_time, end_time: @end_time) +
                    Redborder::Rbdruid.slow_granularity_list(start_time: @start_time, end_time: @end_time)

    params[:granularity] =  if granularities.map(&:to_s).include?(params[:granularity])
                              params[:granularity]
                            else
                              granularities.first
                            end
  end

  def set_total_options
    type = send('set_type_' + params[:controller].to_s)
    @hide_total = (session[type][:hide_total] || 'true')
  end

  def browser_timezone
    return nil if cookies[:tzoffset].blank?
    @browser_timezone ||= begin
      zone = cookies[:tzoffset].to_s
      ActiveSupport::TimeZone[zone]
    end
  end

  #
  # Return the palette or the color for given number
  #
  def color_palette(number = nil)
    palette = [
      '#638bdf', '#e89c4f', '#946acc',
      '#1758b5', '#c74531', '#3e812d', '#d87e34', '#734ca2',
      '#63a0e7', '#f4988a', '#a2bb77', '#edbc6c', '#be90fd',
      '#0f3742', '#783234', '#55633f', '#fdc52b', '#371c39',
      '#38606b', '#996062', '#6f795f', '#878787', '#8a3b8f',
      '#5d8d9a', '#ae9596', '#a8b298', '#bcbcbc', '#eb80f1'
    ]
    # Return the palette or the number
    number ? palette[number] : palette
  end

  #
  # Covert received times to Time Object if the parameters
  # are an integer or a String.
  #
  # == Parameters:
  # start_time::
  #   Start time from the request
  # end_time::
  #   End time from the request
  #
  # == Returns:
  # Set the type of variables to Time object.
  #
  def int_times
    params[:start_time] = if params[:start_time] =~ /^\d+$/
                            Time.at(params[:start_time].to_i)
                          elsif !params[:start_time].nil?
                            params[:start_time].to_time
                          elsif @filter && @filter.exist_any?(:start_time)
                            Time.at(@filter.get(:start_time).keys.first)
                          end

    params[:end_time] = if params[:end_time] =~ /^\d+$/
                          Time.at(params[:end_time].to_i)
                        elsif !params[:end_time].nil?
                          params[:end_time].to_time
                        elsif @filter && @filter.exist_any?(:end_time)
                          Time.at(@filter.get(:end_time).keys.first)
                        end
  end

  def back_without(*params)
    referer = URI.parse(request.referer)
    params_query = Rack::Utils.parse_query(referer.query)

    params.each do |param|
      params_query.delete(param)
    end

    referer.path + '?' + params_query.to_query
  end

  #
  # Sets the flash session variable
  # If the request is a XHR request, it does nothing
  # The default flash type is a notice, but it can be changed
  # using the option type
  #
  # == Parameters:
  # msg::
  #   String of the message
  # options::
  #   A hash with options. The optional parameter now it's type. The type
  #   can be: :notice, :alert, :info.
  #
  def rb_flash(msg, options = {})
    #return if request.xhr?
    options[:type] ||= :notice

    # If its an error, use full_messages instead
    msg = msg.full_messages if msg.respond_to? :full_messages

    # msg must be an array
    msg = [msg].flatten

    # If it only have one element show it, else use a list
    if msg.length == 1
      flash[options[:type]] = msg.join
    else
      flash[options[:type]] = msg.map { |m| "<li>#{m}</li>" }.join
    end
  end
end
