#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class IncidentsController < ApplicationController
  helper Redborder::ModulesHelper

  include ApplicationHelper

  include IncidentsHelper

  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_incident, only: [:edit, :edit_description, :update, :destroy, :show,
                                      :next_phase, :previous_phase, :close, :edit_report_modal,
                                      :update_report_section, :create_note, :create_response_action_note,
                                      :create_response_action_execute, :update_note, :destroy_note,
                                      :generate_ai_reports]
  before_action :set_sensor_names, only: [:index, :bulk_delete, :bulk_update_status, :update, :destroy, :close,
                                          :bulk_create_links, :search, :show, :create_note, :create_response_action_note,
                                          :create_response_action_execute, :update_note, :destroy_note,:generate_ai_reports, :update_report_section]
  before_action :check_incident_sensor_privileges, only: [:show, :update, :destroy, :close,
                                                          :create_note, :create_response_action_note,
                                                          :create_response_action_execute, :update_note, :destroy_note,
                                                          :generate_ai_reports, :update_report_section]
  before_action :check_admin_privileges, only: [:update, :destroy, :close, :bulk_delete, :bulk_update_status,
                                                :bulk_create_links, :unlink, :assign_user, :unassign_user,
                                                :create_note, :create_response_action_note,
                                                :create_response_action_execute, :update_note,
                                                :destroy_note, :generate_ai_reports, :update_report_section]
  before_action :set_priorities, only: [:index, :show, :search, :update]
  before_action :set_statuses, only: [:index, :show, :search, :update]
  before_action :assign_first_user_and_open!, only: [:show]
  before_action :set_users, only: [:show]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all incidents in a table
  #
  def index
    @page = params[:page] || 1
  
    # Get only the incidents where the sensor_id is among those allowed
    @incidents = Incident.where(visible: true, sensor_id: @sensor_names.keys).order(created_at: :desc)

    # Apply search filters if they exist
    if params[:search].present? || params[:name].present?
      @incidents = search # Reuse search method logic
    end

    # To show statistics on index view
    @all_visible_incidents_count = @incidents.count
    @total_incidents_count = @incidents.where(created_at: 1.hour.ago..Time.current).count
    @new_incidents_count = @incidents.where(incident_status_id: 5).count
    @open_incidents_count = @incidents.left_joins(:incident_users).where(incident_users: { incident_id: nil }).count
  
    # Paginate results
    @incidents = @incidents.paginate(page: @page, per_page: current_user.per_page_count)
  
    @priorities = Priority.all
    @incident_statuses = IncidentStatus.all
  end

  #
  # => Render the form to create a new incident
  #
  def new
    @incident = Incident.new(priority_id: Priority.first.id)
    @incident.incident_status = IncidentStatus.find_by(name: IncidentStatus::NEW)
  end

  #
  # => Receive the data to create a new incident. Try to create a new incident
  #    and redirect to incidents_path if it's ok
  #
  def create
    @incident = Incident.new(incident_params)

    respond_to do |format|
      if @incident.save
        format.html do
          redirect_to incidents_path,
                      notice: "#{I18n.t('controllers.incident_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_created')}" }
        format.json { head :no_content }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @incident.errors.full_messages.map do
            |msg| "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @incident.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Render the form to edit an incident
  #
  def edit
    respond_to do |format|
      format.js
    end
  end

  #
  # => Render the form to edit an incident
  #
  def edit_description
    respond_to do |format|
      format.js
    end
  end

  #
  # => Receive the data to update a new incident. Try to update a new incident
  #    and redirect to incident_path if it's ok
  #
  def update
    @page = params[:page] || 1
    original_name = @incident.name
    original_description = @incident.description
  
    respond_to do |format|
      if @incident.update(incident_params)
        
        # Delete AI Generation Record for name if it exists
        if incident_params[:name] && incident_params[:name] != original_name
          AiGenerationRecords.where(table_name: 'incidents', field_name: 'name', record_id: @incident.id).destroy_all
        end
        
        # Delete AI Generation Record for description if it exists
        if incident_params[:description] && incident_params[:description] != original_description
          AiGenerationRecords.where(table_name: 'incidents', field_name: 'description', record_id: @incident.id).destroy_all
        end
  
        format.html do
          notice_message = "#{I18n.t('controllers.incident_updated')}"
  
          if request.referer && request.referer.include?(incident_path(@incident))
            redirect_to incident_path(@incident), notice: notice_message
          else
            redirect_to incidents_path(page: @page), notice: notice_message
          end
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_updated')}" }
        format.json { render json: @incident }
      else
        format.html { render action: 'show' }
        format.js do
          flash[:error] = @incident.errors.full_messages.map do |msg|
            "<li>#{msg}</li>"
          end.join
        end
        format.json { render json: @incident.errors, status: :unprocessable_entity }
      end
    end
  end

  #
  # => Destroy an incident
  #
  def destroy
    @incident.destroy
    respond_to do |format|
      format.html { redirect_to incidents_path }
      format.js
      format.json { head :no_content }
    end
  end

  def next_phase
    @current_phase = @incident.playbook.phases.current || @incident.playbook.phases.order(:id).first
    @current_phase.set_as_current if @current_phase

    if @current_phase && @next_phase = @current_phase.next_phase
      @next_phase.set_as_current
    end

    respond_to do |format|
      format.html { redirect_to incident_path(@incident, anchor: "response") }
      format.js
    end
  end

  def previous_phase
    @current_phase = @incident.playbook.phases.current || @incident.playbook.phases.order(:id).first
    @current_phase.set_as_current if @current_phase

    if @current_phase && @previous_phase = @current_phase.previous_phase
      @previous_phase.set_as_current
    end

    respond_to do |format|
      format.html { redirect_to incident_path(@incident, anchor: "response") }
      format.js
    end
  end

  def close
    respond_to do |format|
      if @incident.update(incident_status: IncidentStatus.find_by(name: IncidentStatus::CLOSED))
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), notice: "#{I18n.t('controllers.incident_status_update')}"
        end
        format.js { flash.now[:notice] = "#{I18n.t('controllers.incident_status_update')}" }
        format.json { render json: @incident }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), alert: "#{I18n.t('controllers.error_incident_status_update')}"
        end
        format.js { flash.now[:error] = @incident.errors.full_messages.join(", ") }
        format.json { render json: @incident.errors, status: :unprocessable_entity }
      end
    end
  end

  def bulk_delete
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)
    
    if incident_ids.any?
      Incident.where(id: incident_ids, sensor_id: @sensor_names.keys).destroy_all
      flash[:notice] = "#{I18n.t('controllers.incident_delete')}"
    else
      flash[:alert] = "#{I18n.t('controllers.incident_no_incidents_delete')}"
    end

    redirect_to incidents_path(page: page)
  end

  def bulk_update_status
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)
    status = IncidentStatus.find_by(name: params[:status])
    status_id = status.id if status
    incidents = Incident.where(id: incident_ids, sensor_id: @sensor_names.keys)

    if status && incidents.update_all(incident_status_id: status_id)
      flash[:notice] = "#{I18n.t('controllers.incidents_updated')}"
    else
      flash[:alert] = "#{I18n.t('controllers.error_incidents_updated')}"
    end

    redirect_to incidents_path(page: page)
  end

  def bulk_create_links
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)
  
    if incident_ids.size > 1
      incidents = Incident.where(id: incident_ids, sensor_id: @sensor_names.keys)
      successfully_linked = []
      failed_links = []
  
      incidents.each do |incident|
        remaining_incidents = incidents.where.not(id: incident.id)  # Exclude the current incident
        
        remaining_incidents.each do |child_incident|
          begin
            unless IncidentLink.link_exists?(incident, child_incident)
              # Create the bidirectional link
              incident.linked_incidents << child_incident
              create_note_link_incident(incident, child_incident, true)
              successfully_linked << "#{incident.id} - #{child_incident.id}"
            end
          rescue ActiveRecord::RecordInvalid => e
            failed_links << "#{incident.id} - #{child_incident.id} (#{e.message})"
          end
        end
      end
  
      if successfully_linked.any?
        flash[:notice] = "#{I18n.t('controllers.incident_create_link')}: #{successfully_linked.join(", ")}"
      end
  
      if failed_links.any?
        flash[:alert] = "#{I18n.t('controllers.incident_create_link_fail')}: #{failed_links.join(", ")}"
      end
    else
      flash[:alert] = I18n.t('controllers.incident_create_link_warning')
    end
  
    redirect_to incidents_path(page: page)
  end

  def unlink
    incident = Incident.find(params[:incident_id]) 
    linked_incident = Incident.find(params[:linked_incident_id])

    # Destroy the original link
    incident.linked_incidents.delete(linked_incident)
    # Destroy the reverse link
    linked_incident.linked_incidents.delete(incident)

    # Create notification and register in worklog
    flash[:notice] = "#{linked_incident.name} #{I18n.t('controllers.incident_unlink')}"
    create_note_link_incident(incident, linked_incident, false)
  
    redirect_to incident_path(incident)
  end

  #
  # Search on incidents
  #
  def search
    @page = params[:page] || 1
    @incidents = Incident.where(visible: true, sensor_id: @sensor_names.keys)
  
    # Filter by name
    if params[:name].present?
      @incidents = @incidents.where("incidents.name ILIKE ?", "%#{params[:name]}%")
    end
  
    # Filter by priority
    if params[:search][:priority].present?
      @incidents = @incidents.where(priority_id: params[:search][:priority])
    end
    
    # Filter by minimum priority
    if params[:search][:minimum_priority].present?
      selected_priority_id = params[:search][:minimum_priority].to_i
      higher_priority_ids = Priority.where('id <= ?', selected_priority_id).pluck(:id)
      @incidents = @incidents.where(priority_id: higher_priority_ids)
    end
  
    # Filter by source
    if params[:search][:source].present?
      @incidents = @incidents.where('LOWER(source) IN (?)', params[:search][:source].map(&:downcase))
    end
  
    # Filter by assigned user
    if params[:search][:assigned].present?
      user_ids = User.where(id: params[:search][:assigned]).pluck(:id)
      @incidents = @incidents.joins(:incident_users).where(incident_users: { user_id: user_ids })
    end
  
    # Filter by date
    if params[:search][:created_at].present?
      time_range = time_range_for_filter(params[:search][:created_at])
      if time_range
        @incidents = @incidents.where(created_at: time_range)
      end
    end
  
    # Filter by status
    if params[:search][:status].present?
      @incidents = @incidents.where(incident_status_id: params[:search][:status])
    end
  
    # Handle sorting
    order_by = params[:search][:order_by]
    direction = params[:search][:order_direction] == 'desc' ? 'desc' : 'asc'
  
    if params[:search][:order_by].present?
      if order_by == 'assigned_user'
        # Join the incident_users and users tables to sort by the users name
        @incidents = @incidents.joins(:incident_users)
                               .joins('LEFT JOIN users ON users.id = incident_users.user_id')
                               .order("users.name #{direction}")
      else
        # For other fields, apply safe sorting
        @incidents = @incidents.order(Incident.arel_table[order_by].send(direction))
      end
    else
      # Fallback to a default order if the column is not allowed
      @incidents = @incidents.order("created_at DESC")
    end
  
    # Paginate results
    @incidents = @incidents.paginate(page: @page, per_page: current_user.per_page_count)
  end

  def search_user
    incident = Incident.find(params[:id])
    query = params[:q]

    users = User.where("name LIKE ?", "%#{query}%")

    # Filter users that are not already in the incident
    filtered_users = users
      .where(id: current_user.users_to_assign.pluck(:id))
      .where.not(id: incident.users.pluck(:id))

    render json: filtered_users.map { |user| { id: user.id, name: user.name } }
  end

  def show
    @actions = gather_all_actions(@incident)
    @worklog_data = sort_worklog_data(gather_all_worklog_data(@incident), :desc)
    @actions_taken_data = fetch_actions_taken_data(@incident)
    @actions_taken_data = sort_worklog_data(@actions_taken_data, :desc)

    search_params = params[:detection_search] || {}
    search_params = search_params.reject { |key, value| value.blank? }
    @elements = get_incident_druid_data(@incident, search_params)

    @detection_unique_sources = @elements.map { |log| log[:source] }.uniq
    @detection_unique_priorities = @elements.map { |log| log[:priority] }.uniq

    @incident_nodes = []
    @incident_links_bidirectional = []

    @elements.each do |element|
      # Assets
      if element[:assets] && !element[:assets].empty?
        element[:assets].each do |asset_element|
          asset_element.each do |key, value|
           @incident_nodes.push({ name: key, type: value, image: get_asset_image(value), asset: true })
          end
        end
      end

      # Observables
      if element[:observables] && !element[:observables].empty?
        element[:observables].each do |observable_element|
          observable_element.each do |key, value|
            @incident_nodes.push({ name: key, type: value, image: get_observable_image(value), observable: true })
           end
        end
      end
    end

    # Sort elements by count
    grouped_assets = @incident_nodes.select { |node| node[:asset] == true }.group_by { |e| e }
    @assets = grouped_assets.map { |key, value| { key: key, count: value.size } }.sort_by { |e| -e[:count] }

    grouped_observables = @incident_nodes.select { |node| node[:observable] == true }.group_by { |e| e }
    @observables = grouped_observables.map { |key, value| { key: key, count: value.size } }.sort_by { |e| -e[:count] }

    @incident_nodes = @incident_nodes.uniq

    @incident_nodes.each do |node|
      related_objects = []
      @elements.each do |element|
        objects = (element[:assets] + element[:observables]).map(&:keys).flatten
        if objects.include? node[:name]
          related_objects += objects
        end
      end

      # Make the link
      related_objects.uniq.each do |object|
        related_node = @incident_nodes.find { |hash| hash[:name] == object }

        next unless related_node

        @incident_links_bidirectional.push({source: @incident_nodes.find_index(node), target: @incident_nodes.find_index(related_node) ,value: 1})
      end
    end

    @incident_links = []

    @incident_links_bidirectional.each do |link|
      unless @incident_links.any? { |existing_link| (existing_link[:source] == link[:target] && existing_link[:target] == link[:source]) }
        @incident_links << link
      end
    end


    # Indicators
    @indicators = Hash.new(0)

    grouped_elements = @elements.group_by { |event| [event[:msg], event[:source], event[:sig_id]] }
    @indicators = grouped_elements.map do |(msg, source, sig_id), entries|
      {
        source: source,
        msg: msg,
        count: entries.size,
        sig_id: sig_id
      }
    end

    @indicators = @indicators.sort_by { |entry| -entry[:count] }

    # Get Mitre techniques / subtechniques
    @tactics = []
    @techniques = []
    @subtechniques = []

    @indicators.each do |indicator|
      next if indicator[:sig_id] == "N/A"

      sid = indicator[:sig_id]
      tactics = Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_tactic = ?", [sid], true)
      @tactics.concat(tactics).uniq! { |tactic| tactic.tactic_id }

      techniques = Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_technique = ?", [sid], true)
      @techniques.concat(techniques).uniq! { |technique| technique.technique_id }

      subtechniques = Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_subtechnique = ?", [sid], true)
      @subtechniques.concat(subtechniques).uniq! { |subtechnique| subtechnique.subtechnique_id }
    end

    # Incident Reports
    @incident_report = @incident.incident_reports

    # Define current phase
    unless @incident.playbook.phases.current
      @incident.playbook.phases.order(:id).first.set_as_current
    end

    @current_phase = @incident.playbook.phases.current

    # Paginate results
    @page = params[:page] || 1
    @elements = @elements.paginate(page: params[:page], per_page: current_user.per_page_count)
  end

  def new_user
    @incident = Incident.find(params[:id])
    @users = current_user.users_to_assign
    respond_to do |format|
      format.js
    end
  end

  # Assign a user to the incident
  def assign_user
    incident = Incident.find(params[:id])
    user = User.find(params[:user_id])
    incident.users << user

    gravatar_uri = get_gravatar_uri(user.email)

    if incident.save
      notification = Notification.create(
        user_id: user.id,
        comment: I18n.t('notifications.incident.assigned', incident_id: incident.id),
        incident_id: incident.id
      )
      notification.save
    end

    render json: { id: incident.id, user_id: user.id, user_name: user.name, user_email: user.email, gravatar_uri: gravatar_uri }
  end

  def unassign_user
    incident = Incident.find(params[:id])
    user = User.find(params[:user_id])

    if incident.users.delete(user)
      respond_to do |format|
        format.html { redirect_to incident_path(incident), notice: "#{I18n.t('controllers.incident_remove_user')}" }
        format.js
      end
      if incident.save
        notification = Notification.create(
          user_id: user.id,
          comment: I18n.t('notifications.incident.unassigned', incident_id: incident.id),
          incident_id: incident.id
        )
        notification.save
      end
    else
      respond_to do |format|
        format.html { redirect_to incident_path(incident), alert: "#{I18n.t('controllers.error_incident_remove_user')}" }
        format.js
      end
    end
  end

  def sidebar_linked_incidents_content
    @incident = Incident.find(params[:incident_id]) 
    
    render partial: 'sidebar_linked_incidents_content', locals: { incident: @incident }
  end

  def sidebar_observable_content
    @observable = params[:observable]
    @observable_links, @incident_nodes = related_observables(params[:incident_id], params[:observable][:name])
    @events = related_events(params[:incident_id], params[:observable][:name])

    render partial: 'sidebar_observable_content', locals: { observable: @observable, observable_links: @observable_links, incident_nodes: @incident_nodes, events: @events }
  end

  def sidebar_content
    @incident = Incident.find(params[:id])

    @elements = get_incident_druid_data(@incident, {})
    @assets = Hash.new(0)

    @elements.each do |element|
      # Assets
      if element[:assets] && !element[:assets].empty?
        element[:assets].each do |asset|
          asset.each do |key, value|
            @assets[key] += 1
           end
        end
      end
    end

    # Sort elements by count
    @assets = @assets.sort_by { |asset, count| -count }

    Rails.logger.error "@Assets is #{@assets}"
    render partial: 'sidebar_content', locals: { incident: @incident, assets: @assets }
  end

  def new_note
    @incident = Incident.find(params[:id])
    @note = Note.new

    respond_to do |format|
      format.js
    end
  end

  def new_response_action_note
    @incident = Incident.find(params[:id])
    @task = Task.find(params[:task_id])
    @note = Note.new

    respond_to do |format|
      format.js
    end
  end

  def new_response_action_execute
    @incident = Incident.find(params[:id])
    @task = Task.find(params[:task_id])

    Rails.logger.info "Task is #{@task.id}"

    @action = nil

    if @task.execution
      actions = IntegrationService.get_actions_by_integration(@task.integration_service)
      @action = actions[@task.action] rescue nil
    end

    Rails.logger.info "Action is #{@action}"

    respond_to do |format|
      format.js
    end
  end

  def create_note
    @note = Note.new(content: note_params[:content])
    @note.user = current_user
    @note.incident = @incident

    respond_to do |format|
      if @note.save

        @worklog_note = sort_worklog_data(gather_notes_worklog_data(@incident), :desc).first

        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "#{I18n.t('controllers.incident_note_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_note_created')}" }
        format.json { head :no_content }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), alert: "#{@note.errors.full_messages.join(", ")}"
        end
        format.js { flash[:error] = @note.errors.full_messages.join(", ") }
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_response_action_note
    @task = Task.find(note_params[:task_id])
    @phase = @task.phase
    
    @note = Note.new(content: note_params[:content])
    @note.user = current_user
    @note.response_action =  @response_action
    
    respond_to do |format|
      if @note.save
        # TODO: make this atomic
        @response_action = ResponseAction.new
        @response_action.task = @task
        @response_action.user = current_user
        @response_action.notes.append(@note)
        @response_action.save

        @worklog_response_action_note = sort_worklog_data(gather_response_action_notes_worklog_data(@incident), :desc).first
        @action_taken_data = sort_worklog_data(fetch_actions_taken_data(@incident), :desc).first

        format.html do
          redirect_to incident_path(@incident, anchor: "response"), notice: "#{I18n.t('controllers.incident_note_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_note_created')}" }
        format.json { head :no_content }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), alert: "#{@note.errors.full_messages.join(", ")}"
        end
        format.js { flash[:error] = @note.errors.full_messages.join(", ") }
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_response_action_execute
    @task = Task.find(params[:task_id])

    @action = {}

    if @task.execution
      actions = IntegrationService.get_actions_by_integration(@task.integration_service)
      @action = actions[@task.action] rescue {}
    end

    action_params = {}
    @action.keys.each { |key| action_params[key] = params[key] unless params[key].empty? || params[key].nil? }

    # TODO: make this atomic
    @response_action = ResponseAction.new
    @response_action.task = @task
    @response_action.user = current_user
    @response_action.save

    Delayed::Job.enqueue(
      Redborder::Jobs::ResponseActionJob.new(true, current_user.id, @response_action.id, @task.integration_service, @task.action, action_params),
      priority: 10,
      run_at: Time.now + 2.seconds
    )

     #stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{@response_action.id}%").first
   
    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "response"), notice: "#{I18n.t('controllers.incident_action_enqueued')}"
      end
      format.js { flash[:notice] = "#{I18n.t('controllers.incident_action_enqueued')}" }
      format.json { head :no_content }
    end
  end

  def edit_note
    @incident = Incident.find(params[:id])
    @note = Note.find(params[:note_id])

    respond_to do |format|
      format.js
    end
  end

  # TODO: only creator of node can update his note
  def update_note
    @note = Note.find(params[:note_id])

    respond_to do |format|
      if @note.update(content: note_params[:content])
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "#{I18n.t('incident_note_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('incident_note_updated')}" }
        format.json { render json: @incident }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @note.errors.full_messages.join(", ")
        end
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def destroy_note
    @note = Note.find(params[:note_id])

    @note.destroy

    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "worklog"), notice: "#{I18n.t('incident_note_destroyed')}"
      end
      format.js
      format.json { head :no_content }
    end
  end

  def worklog_search
    @incident = Incident.find(params[:id])
    search_params = params[:worklog_search] || {}

    type = search_params[:type].presence || 'All'
    user_id = search_params[:user]
    order = search_params[:order] == 'Newest' ? :desc : :asc

    @worklog_data = fetch_worklog_data(type, @incident)
    @worklog_data = filter_worklog_data_by_user(@worklog_data, user_id) if user_id.present?
    @worklog_data = sort_worklog_data(@worklog_data, order)

    @worklog_data
  end

  def detection_search
    @page = params[:page] || 1
    @incident = Incident.find(params[:id])

    search_params = params[:detection_search] || {}
    search_params = search_params.reject { |key, value| value.blank? }
    @elements = get_incident_druid_data(@incident, search_params)

    @detection_unique_sources = @elements.map { |log| log[:source] }.uniq
    @detection_unique_priorities = @elements.map { |log| log[:priority] }.uniq

    # Paginate results
    @page = params[:page] || 1
    @elements = @elements.paginate(page: params[:page], per_page: current_user.per_page_count)
  end

  ## ======================================================================
  # => AI Generation Content Methods
  ## ======================================================================
  def check_ai_job_status
    section = params[:section]
    incident_id = params[:incident_id]
    running = is_ai_running?(section, incident_id)
  
    render json: { running: running }
  end

  def generate_ai_reports
    section = params[:section]
    events = get_incident_druid_data(@incident, {})
  
    # Define the function name and content based on the section
    case section
    when 'executive_summary'
      function_name = 'generate_incident_executive_summary'
      content = "\n- Incident title: #{@incident.name}\n- Incident date: #{@incident.created_at}\n Status: #{@incident.incident_status.name}\n- Assigned users: #{@incident.users.map(&:name).join(', ')}\n- Priority: #{@incident.priority.name}\n- Number of events: #{events.count}\n- Reported by: #{@incident.source}"
      functions_params = { content: content }
    when 'event_summary'
      function_name = 'generate_incident_event_summary'
      content = ""
      events.each_with_index do |event, index|
        observables = event[:observables].map { |obs| obs.keys.first }.join(", ")
        content += "Event #{index + 1}: \nMessage: #{event[:msg]} \nObservables: #{observables} \nSource Address: #{event[:url_filters][:src]} \nDestination Address: #{event[:url_filters][:dst]}\n"
      end
      functions_params = { content: content }
    else
      redirect_to incident_path(@incident, anchor: "report"), alert: "Invalid section"
      return
    end
  
    # Define the model params
    model_params = {
      model_name: 'IncidentReports',
      name_field: section,
      record_id: @incident.incident_reports.id
    }
    # Enqueue the AI job
    Delayed::Job.enqueue(
      Redborder::Jobs::AIGenerateContentJob.new(true, current_user.id, function_name, functions_params, model_params),
      priority: 10,
      run_at: Time.now
    )
  
    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "report"), notice: "#{I18n.t('controllers.incident_generate_ai_' + section)}"
      end
    end
  end

  ## ======================================================================
  # => Report methods
  ## ======================================================================

  #
  # => Create PDF Report 
  #
  def create_incident_report
    @incident = Incident.find(params[:id])
    @incident_report = @incident.incident_reports
    @time = Time.now

    respond_to do |format|
      format.pdf do
        render pdf: "#{@incident.name} report",
               template: "incidents/incident_report.pdf.erb",
               page_size: "A4", 
               javascript_delay: 3000,
               orientation: "Portrait",
               margin: { top: 5, bottom: 5, left: 5, right: 5 }
      end
    end
  end

  #
  # => Render edit report modal
  #
  def edit_report_modal
    @incident_report = @incident.incident_reports
    @section = params[:report_section]
  
    # Fetch the current content of the section from @incident_report
    @section_content = @incident_report[@section]
  
    respond_to do |format|
      format.js
    end
  end

  #
  # => Update report section (post)
  #
  def update_report_section
    @incident_report = @incident.incident_reports
    
    section = params[:section]
    content = params[:content]
    original_content = @incident_report[section]
  
    # Update the corresponding section in the incident report
    respond_to do |format|
      if @incident_report.update(section => content)
        # Delete AI Generation Record for the section if the content has changed
        if content && content != original_content
          AiGenerationRecords.where(table_name: 'incident_reports', field_name: section, record_id: @incident_report.id).destroy_all
        end
        
        format.html do
          redirect_to incident_path(@incident, anchor: "report"), notice: "#{I18n.t('controllers.edit_report')}"
        end
      else
        format.js { render status: :unprocessable_entity }
      end
    end
  end

  private

  # Set @sensor_names for all actions
  def set_sensor_names
    @sensor_names = current_user.domains_with_permission.map { |d| [d.id, d.path_name] }.to_h
  end

  #
  # => Init the incident if the user has permissions
  #
  def set_incident
    @incident = Incident.find(params[:id])
  end

  def set_priorities
    @priorities = Priority.all.order(:id)
  end

  def set_statuses
    @incident_statuses = IncidentStatus.all
  end

  # Those are the users the current_user can manage to add into the incident
  def set_users
    @users = current_user.users_to_share
  end

  def check_admin_privileges
    current_user.admin?
  end

  # Assign firs user that visits a new incident and open it
  def assign_first_user_and_open!
    if @incident.users.empty? && current_user && @incident.new?
      @incident.users.push(current_user)
      if @incident.save
        notification = Notification.create(
          user_id: current_user.id,
          comment: I18n.t('notifications.incident.assigned', incident_id: @incident.id),
          incident_id: @incident.id
        )
        notification.save
        unless @incident.open!
          Rails.logger.error "Failed to open the incident: #{@incident.errors.full_messages.join(', ')}"
        end
      else
        Rails.logger.error "Failed to save the incident: #{@incident.errors.full_messages.join(', ')}"
      end
    end
  end

  def gather_all_actions(incident)
    response_actions = incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions)
    all_actions = (response_actions + response_actions).sort_by(&:created_at)
    all_actions
  end

  def gather_all_worklog_data(incident)
    incident_changes = gather_incident_audit_worklog_data(incident)
    notes = gather_response_action_notes_worklog_data(incident) + gather_notes_worklog_data(incident)
    executions = gather_response_action_executions_worklog_data(incident)
    all_worklog_data = (incident_changes + notes + executions)
  end

  def gather_incident_audit_worklog_data(incident)
    incident.audits.map do |audit|
      formatted_content = audit.comment

      if audit.audited_changes['incident_status_id']
        # Obtener los valores antiguos y nuevos del estado
        old_status_id, new_status_id = audit.audited_changes['incident_status_id']

        # Buscar los nombres de los estados antiguos y nuevos
        old_status = IncidentStatus.find_by(id: old_status_id).try(:name) ? IncidentStatus.find_by(id: old_status_id).name : old_status_id
        new_status = IncidentStatus.find_by(id: new_status_id).try(:name) ? IncidentStatus.find_by(id: new_status_id).name : new_status_id

        # Reemplazar el texto en el comentario
        # TODO: refactor this to cover all languages..
        # English
        formatted_content = formatted_content.gsub(/incident_status_id from '(\d+)' to '(\d+)'/) do
          "status from '#{old_status}' to '#{new_status}'"
        end
        # Spanish
        formatted_content = formatted_content.gsub(/incident_status_id desde '(\d+)' a '(\d+)'/) do
          "estado desde '#{old_status}' a '#{new_status}'"
        end
      end

      {
        action: audit.action,
        user: audit.user,
        created_at: audit.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: formatted_content
      }
    end
  end

  def gather_response_action_executions_worklog_data(incident)
    incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map do |response_action|
      if response_action.notes.any?
        []
      else
        stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{response_action.id}%").first
        stored_job_id = stored_job.present? ? stored_job.id : nil
        [{
          action: "Executed Action",
          user: response_action.user,
          created_at: response_action.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
          source: "[RESPONSE TASK] #{response_action.task.name}",
          stored_job_id: stored_job_id
        }]
      end
    end.compact
  end

  def gather_response_action_notes_worklog_data(incident)
    return incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map(&:notes).map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        source: "[RESPONSE TASK] #{note.response_action.task.name}",
        content: note.content
      }
    end
  end

  def gather_notes_worklog_data(incident)
    return incident.notes.map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: note.content
      }
    end
  end

  def fetch_worklog_data(type, incident)
    case type
    when 'Notes'
      (gather_response_action_notes_worklog_data(incident) +  gather_notes_worklog_data(incident))
    when 'Incident Changes'
      gather_incident_audit_worklog_data(incident)
    when 'Response Actions'
      gather_response_action_notes_worklog_data(incident) + gather_response_action_executions_worklog_data(incident)
    else
      gather_all_worklog_data(incident)
    end
  end

  def format_actions_taken_note(note)
    action = note[:action]
    if note[:action] == "Create"
      action = "Added a note for"
    end
    user = note[:user]
    source = note[:source].sub("[RESPONSE TASK] ", "")
    content = note[:content]
    created_at = note[:created_at]

    { action: action, user: user, created_at: created_at, source: source, content: content }
  end

  def format_actions_taken_execution(execution)    
    action = execution[:action]
    if execution[:action] == "Executed Action"
      action = "Executed Action for "
    end
    user = execution[:user]
    source = execution[:source].sub("[RESPONSE TASK] ", "")
    # content = execution[:content]
    created_at = execution[:created_at]
    stored_job_id = execution[:stored_job_id]

    { action: action, user: user, created_at: created_at, source: source, stored_job_id: stored_job_id }
  end

  def fetch_actions_taken_data(incident)
    data = []
    notes = gather_response_action_notes_worklog_data(incident)
    notes.each do |note|
      data.push(format_actions_taken_note(note))
    end
    
    executions = gather_response_action_executions_worklog_data(incident)

    executions.each do |execution|
      data.push(format_actions_taken_execution(execution))
    end

    data
  end

  def filter_worklog_data_by_user(worklog_data, user_id)
    user = User.find_by(id: user_id)
    return worklog_data unless user

    worklog_data.select { |item| item[:user]['id'] == user.id }
  end

  def sort_worklog_data(worklog_data, order)
    date_format = "%B %d, %Y %I:%M:%S %p"

    converted_data = worklog_data.map do |item|
      item[:created_at_datetime] = DateTime.strptime(item[:created_at], date_format)
      item
    end

    sorted_data = converted_data.sort_by { |item| item[:created_at_datetime] }

    sorted_data.reverse! if order == :desc

    sorted_data.each { |item| item.delete(:created_at_datetime) }

    sorted_data
  end

  # Dynamic mapping method
  def map_dynamic_fields(data, field_map)
    field_map.each do |key, possible_fields|
      possible_fields.each do |possible_field|
        if data.methods.include?(possible_field)
          value = data.send(possible_field)
          return value unless value.nil?
        end
      end
    end
    'N/A'
  end

  def get_incident_druid_data(incident, search_params)
    data = []
  
    # Mapping of fields that can change depending on the module or data source
    field_map = {
      priority: [:priority, :syslogseverity_text], # Priority can be one of these fields
      msg: [:msg, :message],                       # Msg can be one of these fields
    }

    @objects = RedborderObject.where(user_id: current_user.id)

    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }

    incident_modules.each do |mod| 
      source = Redborder::DruidSource.get(source: mod.first)
      filter = Redborder::Filter.new(source, {})
      filter.add(:incident_uuid, incident.uuid)

      start_time = incident.created_at - 12.hour
      end_time = start_time + 12.hour

      # group_by: @source.default_columns_raw,
      data += source.where({
        query_type: :group_by,
        range: :custom,
        start_time: start_time,
        end_time: end_time,
        granularity: :minute,
        aggregation: source.default_aggregation,
        group_by: source.default_columns_incidents,
        search: filter.get
      }).map do |t|
        url_filters = {}
        url_filters[:start_time] = start_time
        url_filters[:end_time] = end_time

        # We use the map_dynamic_fields method to map the fields correctly
        priority = map_dynamic_fields(t, priority: field_map[:priority])
        msg = map_dynamic_fields(t, msg: field_map[:msg])

        source_name = mod.last[:incident_source_name] rescue "N/A"
        
        module_name = mod.first

        timestamp = t.send(:timestamp) 

        # Observables are define in the columns of each module
        observables = []
        assets = []
        mod.last[:columns].select { |column, cfg| cfg[:observable_type] }.each do |column, cfg|
          asset = nil
          url_filters[column] = t.send(column)

          # We check if this is an asset
          if cfg[:object_type] && t.methods.include?(column) && t.send(column)
            asset = RedborderObject.find_by(value: t.send(column), type: cfg[:object_type], user: current_user) 
          end

          if asset
            assets.push({ t.send(column) => cfg[:object_type] }) if t.methods.include?(column) && t.send(column)
          else
            observables.push( { t.send(column) => cfg[:observable_type] }) if t.methods.include?(column) && t.send(column)
          end
        end

        url_filters["incident_uuid"] = incident.uuid

        { timestamp: timestamp, priority: priority, msg: msg, source: source_name, assets: assets, observables: observables, module_name: module_name, url_filters: url_filters }
      end
    end
    
    # Filter by search
    if search_params && search_params[:source]
      data = data.select { |element| element[:source] == search_params[:source] }
    end

    if search_params && search_params[:priority]
      data = data.select { |element| element[:priority] == search_params[:priority] }
    end

    data
  end

  # TODO: each module should define this
  def get_module_source_name(m)
    case m
    when :ips
      "redBorder Intrusion"
    when :vault
      "redBorder Vault"
    else
      m
    end
  end

  def get_asset_image(type)
    view_context.asset_path_from_controller("icons/assets/#{type}.svg")
  end

  def get_observable_image(type)
    view_context.asset_path_from_controller("icons/observables/#{type}.svg")
  end

  #
  # => Return a hash with the params to create/update an incident
  #
  def incident_params
    params
      .require(:incident)
      .permit(:name, :source, :description, :status, :user,
              :priority_id, :incident_status_id, :page, :search, :fast_search)
  end

  def note_params
    params.require(:note).permit(:id, :content, :task_id)
  end

  def related_observables(incident_id, observable_name)
    incident = Incident.find(incident_id)
    elements = get_incident_druid_data(incident, {})
    incident_nodes = []
    incident_links_bidirectional = []
  
    # Extract observable nodes from the incident data
    elements.each do |element|
      if element[:observables] && !element[:observables].empty?
        element[:observables].each do |observable_element|
          observable_element.each do |key, value|
            incident_nodes.push({ name: key, type: value, image: get_observable_image(value), observable: true })
          end
        end
      end
    end
  
    # Remove duplicate nodes
    incident_nodes = incident_nodes.uniq
  
    # Build bidirectional links between nodes
    incident_nodes.each do |node|
      related_objects = []
  
      elements.each do |element|
        objects = (element[:assets] + element[:observables]).map(&:keys).flatten
        if objects.include? node[:name]
          related_objects += objects
        end        
      end
  
      related_objects.uniq.each do |object|
        related_node = incident_nodes.find { |hash| hash[:name] == object }
        next unless related_node
  
        incident_links_bidirectional.push({
          source: incident_nodes.find_index(node),
          target: incident_nodes.find_index(related_node),
          value: 1
        })
      end
    end
  
    # Ensure that each link is unique and does not duplicate the reverse direction
    incident_links = []
  
    incident_links_bidirectional.each do |link|
      unless incident_links.any? { |existing_link| (existing_link[:source] == link[:target] && existing_link[:target] == link[:source]) }
        incident_links << link
      end
    end

    # Filter links involving the specified observable
    observable = incident_nodes.find { |node| node[:name] == observable_name }
  
    observable_links = incident_links.select do |link|
      link[:source] == incident_nodes.find_index(observable) || link[:target] == incident_nodes.find_index(observable)
    end
  
    [observable_links, incident_nodes]
  end

  def related_events(incident_id, observable_name)
    incident = Incident.find(incident_id)
    elements = get_incident_druid_data(incident, {})
    
    filtered_elements = elements.select do |element|
      element[:observables].any? do |observable|
        observable.key?(observable_name)
      end
    end

    filtered_elements
  end

  def time_range_for_filter(selected_range)
    case selected_range
    when "Last Hour"
      1.hour.ago..Time.current
    when "Last 2 hours"
      2.hours.ago..Time.current
    when "Last 3 hours"
      3.hours.ago..Time.current
    when "Last 12 hours"
      12.hours.ago..Time.current
    when "Last 24 hours"
      24.hours.ago..Time.current
    when "Last Week"
      1.week.ago..Time.current
    when "Last 2 Weeks"
      2.weeks.ago..Time.current
    else
      nil
    end
  end

  ## ======================================================================
  # => Incident links methods
  ## ======================================================================

  def create_note_link_incident(incident, linked_incident, is_create)
    if is_create
      # Create note for the original link
      Note.create(
        incident_id: incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_create_link_note')}: 
      <a href='/incidents/#{linked_incident.id}'>#{linked_incident.id}</a>.
      </div>",
      )

      # Create note for the reverse link
      Note.create(
        incident_id: linked_incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_create_link_note')}: 
      <a href='/incidents/#{incident.id}'>#{incident.id}</a>.
      </div>",
      )
    else
      Note.create(
        incident_id: incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_unlink_note')}: 
      <a href='/incidents/#{linked_incident.id}'>#{linked_incident.id}</a>.
      </div>",
      )

      Note.create(
        incident_id: linked_incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_unlink_note')}: 
      <a href='/incidents/#{incident.id}'>#{incident.id}</a>.
      </div>",
      )
    end
  end

  def check_incident_sensor_privileges
    if !@sensor_names.include?(@incident.sensor.id)
      respond_to do |format|
        format.html do
          redirect_to root_path,
                      alert: "#{I18n.t('controllers.no_privileges')}"
        end
        format.json do
          render json: {error: "#{I18n.t('controllers.no_privileges')}"},
                 status: :unauthorized
        end
        format.js { head 403 }
      end
    end

    true
  end
end
