#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Control the Jobs and workers.
#
class JobsController < ApplicationController
  ## ======================================================================
  # => Controller properties
  ## ======================================================================

  respond_to :html
  respond_to :js, :json, only: [
    :index, :destroy, :destroy_stored_job, :info
  ]

  layout false, only: [:info]

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_filter :require_super_administrative_privileges,
                except: [:index, :info, :status]

  before_filter :require_super_reader_privileges,
                only: [:status, :index, :info]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def index
    @node_name = "#{`hostname`.delete("\n")}"

    @stored_jobs =
    if current_user.admin? || current_user.super_reader?
      domains_with_manage_privileges = Sensor.where(type: Sensor::DOMAIN)
                                       .order('name ASC')
                                       .select { |s| current_user.can? :read, s }

      # Search for stored_delayed_jobs of users that belongs to domains where
      # the user has admin permissions.
      stored_delayed_job_ids =
      StoredDelayedJob.includes(user: [:sensor])
      .where(users: { sensor_id: domains_with_manage_privileges.map(&:id) })
      .map(&:id)
      # Search for current_user's stored_delayed_jobs
      stored_delayed_job_ids += StoredDelayedJob.where(user: current_user)

      StoredDelayedJob.where(id: stored_delayed_job_ids)
      .where.not(content: nil)
      .order('stored_delayed_jobs.created_at DESC')
      .paginate(page: params[:page], per_page: current_user.per_page_count)
    else
      StoredDelayedJob.where(user: current_user)
      .where.not(content: nil)
      .order('created_at DESC')
      .paginate(page: params[:page], per_page: current_user.per_page_count)
    end

    if current_user.super_admin? || current_user.super_reader?
      @jobs = DelayedJob.all
      @workers = DelayedJob.process
    else
      @jobs = DelayedJob.all.select { |dj| dj.user_id.equal? current_user.id }
    end
    respond_with @jobs, @workers, location: jobs_path
  end

  #
  # => TODO: status action will be replaced by index
  #
  def status
    @node_name = "#{`hostname`.delete("\n")}"
    @jobs = DelayedJob.all
    @stored_jobs =
      if current_user.admin? || current_user.super_reader?
        domains_with_manage_privileges =
          Sensor.where(type: Sensor::DOMAIN).order('name ASC')
                .select { |s| current_user.can? :read, s }
  
        # Search for stored_delayed_jobs of users that belongs to domains where
        # the user has admin permissions.
        stored_delayed_job_ids =
          StoredDelayedJob.includes(user: [:sensor])
                        .where(users: { sensor_id: domains_with_manage_privileges.map(&:id) })
                        .map(&:id)
        # Search for current_user's stored_delayed_jobs
        stored_delayed_job_ids += StoredDelayedJob.where(user: current_user).pluck(:id)
  
        StoredDelayedJob.where(id: stored_delayed_job_ids)
                       .where.not(content: nil)
                       .where('job ILIKE ?', "%#{params[:search]}%") if params[:search].present?
                       .order('stored_delayed_jobs.created_at DESC')
                       .paginate(page: params[:page], per_page: current_user.per_page_count)
      else
        StoredDelayedJob.where(user: current_user)
                       .where.not(content: nil)
                       .where('job ILIKE ?', "%#{params[:search]}%") if params[:search].present?
                       .order('created_at DESC')
                       .paginate(page: params[:page], per_page: current_user.per_page_count)
      end
    @workers = DelayedJob.process
  end  

  #
  # => Show the info of a executed delayed job
  #
  def info
    @stored_job = StoredDelayedJob.find(params[:id])
    return unless (!current_user.admin && !current_user.super_reader? && current_user.id != @stored_job.user_id)
    no_enough_privileges redirect_to: jobs_path
  end

  #
  # => search for Stored Jobs
  #
  def search
    search_term = "%#{params[:search].try(:downcase)}%"
    @stored_jobs = StoredDelayedJob.where('LOWER(job) LIKE ?', search_term)
                                   .order(created_at: :desc)
                                   .paginate(page: params[:page], per_page: current_user.per_page_count)
  end  
  #
  # => Destroy a delayed job
  #
  def destroy
    @job = DelayedJob.find(params[:id])
    if @job.locked_at.nil?
      @job.destroy
    else
      flash[:alert] = "#{I18n.t('controllers.job_running')}"
    end
    respond_to do |format|
      format.html { redirect_to jobs_url }
      format.json { head :no_content }
    end
  end

  #
  # => Destroy the info of a stored job
  #
  def destroy_stored_job
    @stored_job = StoredDelayedJob.find(params[:id])
    @stored_job.destroy
    respond_to do |format|
      format.html { redirect_to jobs_url }
      format.json { head :no_content }
    end
  end

  #
  # => Starts a delayed job if it isn't running
  #
  def start
    DelayedJob.start unless DelayedJob.running?
    redirect_to jobs_path
  end

  #
  # => Stop a delayed job if it's running
  #
  def stop
    DelayedJob.stop if DelayedJob.running?
    redirect_to jobs_path
  end

  #
  # => Restart a delayed job if it's running
  #
  def restart
    DelayedJob.restart if DelayedJob.running?
    redirect_to jobs_path
  end

  #
  # => Starts an alarm update job
  #
  def start_alarm
    DelayedJob.alarm.destroy! if DelayedJob.alarm?
    Delayed::Job.enqueue(Redborder::Jobs::CheckAlarmsJob
    .new(true), priority: 10, run_at: Time.now + 5.seconds)

    redirect_to jobs_path
  end

  def start_checks
    DelayedJob.check_corrupt_data.destroy! if DelayedJob.check_corrupt_data?
    Delayed::Job.enqueue(
      Redborder::Jobs::CheckCorruptDataJob.new(true),
      priority: 5,
      run_at: Time.now + 5.seconds
    )

    redirect_to jobs_path
  end

  #
  # => Starts a report update job
  #
  def start_report
    DelayedJob.report.destroy! if DelayedJob.report?
    Delayed::Job.enqueue(Redborder::Jobs::CheckReportsJob
    .new(true), priority: 10, run_at: Time.now + 5.seconds)

    redirect_to jobs_path
  end

  #
  # => Starts a GeoIP update job
  #
  def start_geoip_update
    DelayedJob.geoip_update.destroy! if
    DelayedJob.geoip_update? && !DelayedJob.geoip_update.running?

    Delayed::Job.enqueue(Redborder::Jobs::GeoipUpdateJob
    .new(true), priority: 10, run_at: Time.now + 10.seconds)

    redirect_to jobs_path
  end

  #
  # => Wakeup chef
  #
  def start_rb_wakeup_chef_cluster
    Delayed::Job.enqueue(Redborder::Jobs::RbWakeupChefClusterJob
    .new(true, current_user.id), priority: 5) unless
    DelayedJob.rb_wakeup_chef_cluster?

    render layout: false
  end

  #
  # => Update Vulnerabilities database
  #
  def start_rb_vulnerabilities_update_db
    DelayedJob.rb_vulnerabilities_update_db.destroy! if
      DelayedJob.rb_vulnerabilities_update_db? && !DelayedJob.rb_vulnerabilities_update_db.running?

    Delayed::Job.enqueue(Redborder::Jobs::RbVulnerabilitiesUpdateDBJob
                           .new(current_user.id, true), priority: 10)

    redirect_to jobs_path
  end

  #
  # => Starts a title/description generation job
  #
  def start_update_content_incidents
    DelayedJob.update_content_incidents.destroy! if
      DelayedJob.update_content_incidents? && !DelayedJob.update_content_incidents.running?
    Delayed::Job.enqueue(Redborder::Jobs::UpdateContentIncidentsJob
                        .new(true, current_user.id), priority: 10, run_at: Time.now + 5.minutes)

    redirect_to jobs_path
  end

  def start_glpi_inventory_job
    DelayedJob.glpi_inventory_job.destroy! if
      DelayedJob.glpi_inventory_job? && !DelayedJob.glpi_inventory_job.running?

    Delayed::Job.enqueue(Redborder::Jobs::GlpiInventoryJob
                           .new(current_user.id, true), priority: 10)
    redirect_to jobs_path                  
  end

  def start_incident_to_syslog_job
    DelayedJob.incident_to_syslog.destroy! if
      DelayedJob.incident_to_syslog? && !DelayedJob.incident_to_syslog.running?

    Delayed::Job.enqueue(Redborder::Jobs::IncidentToSyslogJob
                           .new(current_user.id, true), priority: 10)
    redirect_to jobs_path                  
  end

  def start_delete_stored_jobs
    Delayed::Job.enqueue(Redborder::Jobs::DeleteStoredJobs.new(current_user.id, true), 
                        priority: 10, run_at: 1.day.from_now.beginning_of_day) unless DelayedJob.delete_stored_jobs?

    redirect_to jobs_path                  
  end
end
