#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class RedborderObjectsController < ApplicationController
  # -------------------------------------------------------------
  # Before actions
  # -------------------------------------------------------------

  before_action :set_object, only: [:edit, :update, :destroy]

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  #
  # Render all the objects inside tabs.
  #
  def index
    @net_objects        = current_user
                          .net_objects
                          .order(:name).paginate(
                            page: 1, per_page: current_user.per_page_count)
    @mac_objects        = current_user
                          .mac_objects
                          .order(:name).paginate(
                            page: 1, per_page: current_user.per_page_count)
    @vlan_objects       = current_user
                          .vlan_objects
                          .order(:name).paginate(
                            page: 1, per_page: current_user.per_page_count)
    @service_objects    = current_user
                          .service_objects
                          .order(:name).paginate(
                            page: 1, per_page: current_user.per_page_count)
    @application_objects = ApplicationObject.all.order(:name)
                           .paginate(page: 1, per_page: current_user.per_page_count)
  end

  #
  # Render the form to create a new object
  #
  def new
    return unless RedborderObject::OBJECT_CLASSES.include? params[:class]
    @object = params[:class].constantize.new(value: params[:value])
    render layout: false
  end

  #
  # Render the form to edit a object
  #
  def edit
    render layout: false
  end

  #
  # Receive the POST data of an object an try to save it
  #
  def create
    # Store the object
    @object = RedborderObject.new(object_params)
    @object.user = current_user
    return unless @object.save
    # Get objects
    @objects = current_user.send(@object.type.pluralize.underscore)
               .paginate(page: 1, per_page: current_user.per_page_count) if
               current_user.send(@object.type.pluralize.underscore).one?
  end

  #
  # Receive the POST data of an object an try to update it
  #
  def update
    @object.update(object_params)
  end

  #
  # Destroy an object
  #
  def destroy
    @object.destroy if @object
  end

  #
  # Store session tab of the user
  #
  def change_session_tab
    session[:object] = params[:id].to_sym
    render js: ''
  end

  # TODO: THIS NEED TO BE DELETED
  # def apply
  #   @sensor.apply_objects
  #   Manager.apply_objects
  #   respond_to do |format|
  #     format.html { redirect_to :back, notice: 'Objects has been applied.' }
  #     format.json { head :no_content }
  #   end
  # end

  #
  # Search in objects
  #
  def search
    filter = ['( LOWER(name) LIKE :search or ' \
              'LOWER(comment) LIKE :search or LOWER(value) LIKE :search)',
              search: "%#{params[:search].try(:downcase)}%"]

    begin
      object_type = RedborderObject.from_string(params[:tab])
    rescue RedborderObject::Unknown
      error_400("#{I18n.t('controllers.no_object')}") && return
    end

    @objects =
      current_user.send(object_type)
      .where(filter).order(:name)
      .paginate(page: params[:page], per_page: current_user.per_page_count)
  end

  #
  # Export all objects as CSS
  #
  def export
    tab = params[:tab]
    export_objects_csv tab
  end

  #
  # Show a modal to upload a file to import objects.
  #
  def import
    @type = params[:type]
    render layout: false
  end

  #
  # Import objects from a given CSV file.
  #
  def import_file
    tab = params[:tab]
    content_file = params[:redborder_object][:csv_file]

    if params[:add_vars] == 'true'
      vars_count = RedborderObject.add_vars_from_file(content_file, tab)
      msg = "#{vars_count[0]} #{I18n.t('controllers.objects_added')} #{vars_count[1]} #{I18n.t('controllers.objects_updated')}"
    else
      vars_count = RedborderObject.replace_vars_from_file(content_file, tab)
      msg = "#{vars_count} #{I18n.t('controllers.objects_added')}"
    end

    redirect_to objects_path, notice: msg
  end

  private

  #
  # Set the object.
  #
  def set_object
    type = RedborderObject.from_string(params[:object][:type])
    @object = current_user.send(type).where(id: params[:id]).first
    return if @object

    # The object doesn't exist!
    rb_flash "#{I18n.t('controllers.object_not_exist')}", type: :error
  end

  #
  # Export objects as CSV
  #
  def export_objects_csv(tab)
    filename = in_time_machine { Time.now.strftime('%Y%m%d%H%M%S') +"-#{tab.capitalize}-Objects.csv" }
    content = objects_to_csv_format(tab)
    send_data content, filename: filename
  end

  #
  # Convert current Objects to CSV
  #
  def objects_to_csv_format(tab)
    search_term = params[:search].presence || ""
    search_term = "" if search_term == "undefined"

   
    filter = ['(LOWER(name) LIKE :search OR LOWER(comment) LIKE :search OR LOWER(value) LIKE :search)',
              { search: "%#{search_term.downcase}%" }]

    objects = if current_user.super_admin?
                RedborderObject.where(user: [current_user, nil], type: "#{tab.capitalize}Object")
              else
                RedborderObject.where(user: current_user, type: "#{tab.capitalize}Object")
              end

    # Generate CSV content within the same method.
    objects.where(filter).order(:name).map do |o|
      if o.is_a?(MacObject)
        object_type = o.object_type.presence || "N/A"
        "#{o.name};#{o.value};#{o.comment};#{object_type};#{o.operating_system};#{o.ip_addresses}"
      else
        "#{o.name};#{o.value};#{o.comment}"
      end
    end.join("\n")
  end

  #
  # Allowed params when create an object
  #
  def object_params
    params.require(:object).permit(:name, :value, :comment, :type, :object_type)
  end
end
