#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'zk'
require 'getopt/std'
require 'tempfile'

class ReportsController < ApplicationController
  ## ======================================================================
  # => INCLUDES
  ## ======================================================================

  include AlarmsHelper
  helper GroverHelper::GroverAssetsHelper

  ## ======================================================================
  # => Controller properties
  ## ======================================================================

  respond_to :html
  respond_to :js, :json,
             only: [:create, :update, :destroy, :save_composition, :update]

  layout false, only: [:import, :send_now, :new_user, :new_sensor, :generate,
                       :remove_page, :remove_page_modal, :sensor_childrens]

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  # Check if the user has a report
  before_action :check_reports, only: :index

  # Set the current report
  before_action :set_report,
                only: [:show, :edit, :update, :destroy, :composition,
                       :save_composition, :export, :send_now, :generate,
                       :add_page, :remove_page_modal, :remove_page]

  # Check if the user can read a report
  before_action :check_read_permissions,
                only: [:show, :edit, :composition, :export, :generate]

  # Check if the user have permission to do some actions
  before_action :check_write_permissions,
                only: [:update, :save_composition, :send_now, :add_page,
                       :remove_page_modal, :remove_page]

  # Remove owner and current_user from params
  # before_action :remove_user_and_owner, only: [:create, :update]

  # Set the current permissions
  before_action :set_write_permissions, only: [:index]

  # Initialize the session report_pages
  before_action :initialize_report_pages,
                only: [:composition, :save_composition]

  # Initialize the times when generate a report
  before_action :int_times,
                only: [:show]

  before_action :check_collaborators, only: [:create, :update]

  ## ======================================================================
  # => Actions
  ## ======================================================================
  helper_method :sort_column, :sort_direction
  helper_method :report_sort_column, :report_sort_direction

  DittoCode::Exec.if 'COMMUNITY' do
    #
    # => Return all the reports of a user. Return the reports
    #    of his sensors and group them by sensor
    #
    def index
      # Get all the reports
      all_reports = current_user.reports.order(:name)

      if all_reports.size > 1
        @user_reports = [all_reports.pop]
        all_reports.destroy_all
      else
        @user_reports = all_reports
      end

      # Get collaborative reports
      @reports_from_others = current_user.reports_from_others.uniq.reorder(report_sort_column(params[:others_reports_sort]) + ' ' + report_sort_direction(params[:others_reports_direction]))
                            .where(
                              'LOWER(name) LIKE :search OR LOWER(description) LIKE :search',
                              search: "%#{params[:query].try(:downcase)}%")
                            .paginate(page: params[:others_reports_page], per_page: current_user.per_page_count)
      @reports_from_sensors = []
    end
    
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def index
      #This method uses .reorder() in order to override the default scope
      @user_reports = current_user.reports.reorder(report_sort_column(params[:user_reports_sort]) + ' ' + report_sort_direction(params[:user_reports_direction]))
                            .where(
                              'LOWER(name) LIKE :search OR LOWER(description) LIKE :search',
                              search: "%#{params[:query].try(:downcase)}%")
                            .paginate(page: params[:user_reports_page], per_page: current_user.per_page_count/3)
      @reports_from_others = current_user.reports_from_others.reorder(report_sort_column(params[:others_reports_sort]) + ' ' + report_sort_direction(params[:others_reports_direction]))
                            .where(
                              'LOWER(name) LIKE :search OR LOWER(description) LIKE :search',
                              search: "%#{params[:query].try(:downcase)}%")
                            .paginate(page: params[:others_reports_page], per_page: current_user.per_page_count/3)
      @reports_from_sensors = current_user.reports_from_sensors.reorder("reports." + report_sort_column(params[:sensor_reports_sort]) + ' ' + report_sort_direction(params[:sensor_reports_direction]))
                            .where(
                              'LOWER(name) LIKE :search OR LOWER(description) LIKE :search',
                              search: "%#{params[:query].try(:downcase)}%")
                            .paginate(page: params[:sensor_reports_page], per_page: current_user.per_page_count/3)
      respond_to do |format|
        format.html
        format.js
      end
    end

    def history_data
      @user_historicals = complete_history
        .select("reports_historicals.*, reports.name as report_name")
        .joins(:report)
        .where(
          'LOWER(reports.name) LIKE :search OR LOWER(reports_historicals.generated_by) LIKE :search',
          search: "%#{params[:query] ? params[:query].downcase : ''}%",
          user_id: current_user.id
        )
        .order("#{sort_column} #{sort_direction}")
        .paginate(page: params[:page], per_page: current_user.per_page_count)

      respond_to do |format|
        format.html { render 'historical_index' }
        format.js { render 'historical_index' }
      end
    end

    def history_records
      @report = Report.find(params[:id])
      @user_historicals = @report.reports_historicals.reorder(sort_column + ' ' + sort_direction)
                                 .paginate(
                                   per_page: current_user.per_page_count, page: params[:page]
                                 )
      respond_to do |format|
        format.html { render 'historical_index' }
        format.js { render 'historical_index' }
      end
    end

    #
    # => Initiate the report. Load the users and the sensors
    #    the user can see.
    #
    #    @route    /reports/new
    #    @helper   new_report_path
    #
    def new
      @report = current_user.reports.build
    end

    #
    # => Create a new Report
    #
    def create
      # Get all the params
      to_save = create_report_params

      # Default values
      to_save.merge!(
        number_pages: 1
      )

      # Save the collaborators and the sensors
      collaborators = to_save.delete(:collaborators)
      sensors_collaborators = to_save.delete(:sensors_collaborators)

      # Create the Report
      @report = Report.new(to_save)

      # Associate the user
      @report.owner = current_user

      if @report.save
        save_collaborators @report, collaborators
        save_sensors_collaborators @report, sensors_collaborators
        rb_flash I18n.t('report_success')
      end

      respond_with @report, location: composition_report_path(@report)
    end
  end

  #
  # Show a modal to generate a report at this time or in the past
  #
  def generate
  end

  #
  # => Destroy the report if the current user is the owner
  #    or leave if it's a collaborative report for the user.
  #
  def destroy
    if @report.owner == current_user
      @report.destroy
    else
      @report.report_users.find_by(user_id: current_user.id).try(:destroy)
    end

    respond_with @report, location: reports_url
  end

  #
  # => Render the report in PDF.
  #
  def show
    @start_time = params[:start_time]
    @end_time = params[:end_time]

    if !self.date_evaluation(@start_time, @end_time)
      # Get the granularity
      @granularity = Redborder::Rbdruid.granularity_list(
        start_time: @start_time,
        end_time: @end_time
      )[1]
      # We must apply the range time
      @blocks = @report.blocks_data(
        start_time: @start_time,
        end_time: @end_time,
        granularity: @granularity
      )
    else
      @blocks = @report.blocks_data
    end
    
    # Used to render widgets
    @parent = @report

    #Generate PDF and Historical
    respond_to do |format|
      format.html do
        render template: "reports/show.pdf.erb", locals: { report: @report }
      end
      format.pdf do
        grover = Grover.new(
          render_to_string(
            template: "reports/show.pdf.erb", 
            locals: { report: @report }
          ),
          orientation: @report.orientation_name
        )
        
        # Generate the PDF data
        pdf_data = grover.to_pdf

        Tempfile.open(['temp', '.pdf']) do |file|
          file.binmode
          file.write(pdf_data)
          file.rewind
          create_report_historical(@report.id, file)
        end
        show_pdf(pdf_data, @report.name)
      end
    end
  end

  def create_report_historical(report_id , file)
    @reports_historical = ReportsHistorical.new
    @reports_historical.report_id = report_id
    @reports_historical.generated_by = current_user.name
    @reports_historical.pdf_file = file
    if @reports_historical.save
      Rails.logger.info "Report: '#{report_id}' generated."
    else
      Rails.logger.error "Report '#{report_id}' couldn't be generated."
    end
  end

  def visualize_pdf
    @reports_historical = ReportsHistorical.find(params[:id])
    pdf_data = open(@reports_historical.pdf_file_uri,
                        ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE).read
    show_pdf(pdf_data, @reports_historical.report.name)
  end

  def show_pdf(pdf_data, filename)
    send_data pdf_data, filename: "#{filename}.pdf",
                        type: 'application/pdf',
                        disposition: 'inline'
  end

  def destroy_report_historical
    @reports_historical = ReportsHistorical.find(params[:id])
    destroy_report_historical_if_authorized
    respond_to_action(params[:history_action])
  end

  # 
  # Function to check the dates that are passed 
  # by parameter to avoid null fields and fields in 
  # non-evaluable ranges of dates.
  # 
  def date_evaluation(init, final)
    @invalid_date = false
    today = DateTime.now
    if init.nil? || final.nil?
      @invalid_date = true
    elsif  init.blank? || final.blank? 
      @invalid_date = true
    elsif today < init
      @invalid_date = true
    elsif today < final
      @invalid_date = true
    elsif final < init
      @invalid_date = true
    end
    return @invalid_date
  end

  #
  # => Edit the general info of a Report
  #
  #    @route
  #    @helper
  #
  def edit
  end

  #
  # => Render the view that allow the user to edit the composition of a Report
  #
  #    @route
  #    @helper
  #
  def composition
    # TODO: SAVE THE CURRENT PAGE OF A USER!
    @current_page = 0
    @pages = @report.pages # Get the pages

    # This is used for render widgets. Set the parent and the manage permission
    # (always true on reports)
    @parent = @report
    @manage = true
  end

  #
  # Add a page to the curren report
  #
  def add_page
    @report.number_pages += 1
    # Update teh page of widgets. We can add a page in the place we choose.
    after = params[:after].to_i
    @report.widgets.each do |widget|
      if widget.page.to_i > after
        widget.page = (widget.page.to_i + 1).to_s
        widget.save
      end
    end

    @report.save

    render json: { status: 'ok' }
  end

  #
  # Show a modal to remove a page
  #
  def remove_page_modal
    @page = params[:page]
  end

  #
  # Remove a page. It receive by parameters teh option to move the blocks
  # to another page, or delete them
  #
  def remove_page
    @page = params[:page].to_i
    @move_page = params[:move_page].to_i
    # Check if we must delete the blocks
    if params[:move_blocks].nil?
      # Delete the block
      @report.widgets.each do |w|
        next if w.page < @page

        if w.page == @page
          w.destroy
        else
          w.page = (w.page.to_i - 1).to_s
          w.save
        end
      end
    else
      # Move blocks
      @report.widgets.each do |w|
        next if w.page < @page

        if w.page == @page
          w.page = @move_page
        else
          w.page = (w.page.to_i - 1).to_s
        end
        w.save
      end
    end

    @report.number_pages -= 1
    @report.save
  end

  #
  # => Update the current report with the new params
  #    The sensors and the users must be saved
  #
  def update
    # Get all the params
    to_save = update_report_params

    # Save the collaborators and the sensors
    collaborators = to_save.delete(:collaborators)
    sensors_collaborators = to_save.delete(:sensors_collaborators)

    if @report.update(to_save)
      save_collaborators @report, collaborators
      save_sensors_collaborators @report, sensors_collaborators
      rb_flash I18n.t('report_update')

      if params[:redirect_to_composition].present?
        redirect_to composition_report_path(@report)
      else
        redirect_to reports_path
      end
    else
      render action: :edit
    end
  end

  #
  # => Send the report to receivers now by a job.
  #

  def send_now
    Delayed::Job.enqueue(Redborder::Jobs::SendReportJob
      .new(@report.id, true), priority: 10, run_at: Time.now)
  end

  #
  # => Export a current report into a tar.gz
  #
  def export
    send_file @report.export, filename: "#{@report.name}.tar.gz"
  end

  #
  # => Load the form to upload a exported report
  #
  def import
    @report = current_user.reports.build
  end

  #
  # => Process the uploaded exported report
  #
  def save_import

    @report = Report.import(import_params)
    @report.owner = current_user


    if @report.save
      session[:report_id] = @report.id
      rb_flash I18n.t('report_success')
      redirect_to reports_path
    else
      rb_flash @report.errors, type: :error
      redirect_to action: :index
    end
    rescue => e
      rb_flash "#{e.message} \n\n", type: :error
      redirect_to action: :index
  end

  def new_user
    @users = current_user.users_to_share
  end

  def new_sensor
    @sensors = if current_user.super_admin?
                 current_user.sensor.children.where(type: Sensor::DOMAIN)
                 .reject { |s| (s.unassigned? && s.children.count.zero?) || s.is_root? }
               else
                 [current_user.sensor]
               end
  end

  private

  def remove_user_and_owner
    user_id = @report.nil? ? current_user.id : @report.user_id
    params[:report][:collaborators].delete_if do |user|
      user['user_id'] == user_id
    end
  end

  def check_reports
    return unless current_user.reports.blank?
    current_user.reports.create(
      name: I18n.t('default'),
      description: I18n.t('default_report'),
      orientation: Report::ORIENTATION_PORTRAIT
    )

    current_user.reload
  end

  def set_report
    @report = Report.find(params[:id])
  end

  #
  # => It checks the user's permissions
  #    If the user hasn't got permissions, he will
  #    be redirected to the index
  #
  def check_read_permissions
    available_reports = current_user.readable_reports
    return if available_reports.include? @report
    no_enough_privileges redirect_to: reports_path
  end

  #
  # => It checks the user's permissions
  #    If the user hasn't got permissions, he will
  #    be redirected to the index
  #
  def check_write_permissions
    available_reports = current_user.writable_reports
    return if available_reports.include? @report
    no_enough_privileges redirect_to: reports_path
  end

  #
  # => Check if the session[:report_pages] is
  #    a Hash. If it isn't, this method initialize it
  #
  def initialize_report_pages
    # Initialize the session report pages
    return unless session[:report_pages].nil? || !session[:report_pages].is_a?(Hash)
    session[:report_pages] = {}
  end

  #
  # Set the list of reports ids the user can edit
  #
  def set_write_permissions
    @writable_reports = current_user.writable_reports
  end

  #
  # => Never trust parameters from the scary internet
  #    only allow the white list through.
  #
  def create_report_params
    params.require(:report)
      .permit(:name, :description, :orientation,
              :schedule_type, :email_list, :schedule_days, :schedule_hour,
              collaborators: [:user_id, :can_edit],
              sensors_collaborators: [:sensor_id, :can_edit])
  end

  def update_report_params
    params.require(:report)
      .permit(:name, :description, :number_pages,
              :schedule_type, :email_list, :schedule_days, :schedule_hour,
              collaborators: [:user_id, :can_edit],
              sensors_collaborators: [:sensor_id, :can_edit])
  end

  def pages_param
    JSON.parse(params.require(:report)[:pages])
  end

  #
  # => Get the current page
  #
  def current_page_param
    params.require(:report)[:current_page]
  end

  def import_params
    post = params[:report]

    # Exit if the file doesnt exist
    if post.nil?
      rb_flash I18n.t('specify'), type: :error
      redirect_to reports_path
    end

    post[:file]
  end

  #
  # => Get the common blocks from deleted pages to destroy them
  #
  def delete_blocks_params
    params.require(:report)[:to_delete] || []
  end

  def save_collaborators(report, collaborators)
    # Remove all relations, except those with the owner and the current user
    report.report_users.each do |x|
      x.delete unless [x.report.user_id, current_user.id].include? x.user_id
    end

    return if collaborators.nil?

    # Dont save relations with the owner or the current user
    collaborators.each do |user|
      next if user['user_id'] == current_user.id ||
              user['user_id'] == report.user_id

      report.report_users.create(
        user_id: user['user_id'],
        can_edit: (user['can_edit'] == 'on')
      )
    end
  end

  def save_sensors_collaborators(report, sensors_collaborators)
    # Remove all relations
    report.report_sensors.destroy_all
    return if sensors_collaborators.nil?

    # Save all relations
    sensors_collaborators.each do |_index, sensor|
      next unless [:all, :admin, :none].include? sensor['can_edit'].to_sym

      report.report_sensors.create(
        sensor_id: sensor['sensor_id'],
        can_edit: sensor['can_edit']
      )
    end
  end

  #
  # Saves the pages. We use the intermediate table Report_composition
  # to set the page and the position of the blocks
  #
  def save_pages
    # Disasociated previous blocks
    @report.report_blocks.update_all(page: nil)

    # Iterate over the pages
    pages_param.each do |page|
      # Iterate over the blocks in the page
      page.last.each_with_index do |block, position|
        next unless numeric? block['id'] # Check if it's and ID

        # Get the block
        report_block = ReportBlock.find(block['id'].to_i)
        next unless report_block.report == @report

        # Set the data
        report_block.page = page.first
        report_block.position = position
        report_block.save
      end
    end
  end

  #
  # Destroy the blocks of a deleted page. We must delete the blocks here
  # because the user can cancel the changes, and he must can recover the pages
  #
  def destroy_blocks
    # TODO: Check the sensor and the user
    items = delete_blocks_params.split(',') || []
    items.each do |id|
      ReportBlock.find(id).destroy
    end
  end

  def numeric?(obj)
    !obj.to_s.match(/\A[+-]?\d+?(\.\d+)?\Z/).nil?
  end

  def check_collaborators
    user_collaborators = User.where(id: (params[:report][:collaborators] || []).map(&:values).flatten)
    domain_collaborators = Sensor.where(type: Sensor::DOMAIN,
                                        id: (params[:report][:sensors_collaborators] || {}).values.map { |s| s['sensor_id'] })
    return error_403('') if domain_collaborators.select { |d| current_user.cannot?(:read, d) }.any?
    return error_403('') if (user_collaborators - current_user.users_to_share).any?
  end

  def sort_column
    params[:sort] || 'created_at'
  end

  def sort_direction
    %w[asc desc].include?(params[:direction]) ? params[:direction] : 'desc'
  end

  def report_sort_column(column)
    Report.column_names.include?(column) ? column : 'name'
  end

  def report_sort_direction(direction)
    %w[asc desc].include?(direction) ? direction : 'asc'
  end

  def destroy_report_historical_if_authorized
    if @reports_historical.generated_by == current_user.name || current_user.admin?
      @reports_historical.delete_file @reports_historical.id
      @reports_historical.destroy
    end
  end

  def respond_to_action(action)
    params[:id] = params[:report_id] if params[:report_id].present?
    case action
    when 'history_data'
      respond_with @reports_historical, location: history_data_reports_path
    when 'history_records'
      respond_with @reports_historical, location: history_records_report_path(id: params[:id])
    else
      respond_with @reports_historical, location: reports_url
    end
  end

  def complete_history
    report_ids = current_user.reports.pluck(:id) +
                 current_user.reports_from_others.pluck(:id) +
                 current_user.reports_from_sensors.pluck(:id)

    ReportsHistorical.where(report_id: report_ids).distinct
  end
end
