#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# Controller to perform actions in users like edit, create or destroy.
# To perfrom some actions, the user must have permissions. For example:
#
# - Only a super admin can show, edit and destroy another super admin.
# - To show, edit and destroy an user, current user must have admin
#   permission on its domain.
#
class UsersController < ApplicationController
  # -------------------------------------------------------------
  # Before actions
  # -------------------------------------------------------------

  before_action :require_administrative_privileges,
                only: [:new, :index, :letter_search, :search, :load_more, :destroy]
  before_action :set_user_and_check_permission,
                only: [:edit, :update, :destroy, :show]
  before_action :check_domain_permission, only: :update
  before_action :require_recreate_api_key_privileges, only: [:recreate_api_key]

  # -------------------------------------------------------------
  # Controller properties
  # -------------------------------------------------------------

  layout false, only: [:list]

  # Number of users showed per page
  USERS_PAGE = 20

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  #
  # Return the list of users.
  #
  def index
    DittoCode::Exec.if 'ENTERPRISE' do
      @users = users_can_see
               .includes(:sensor)
               .order('LOWER(name) ASC').take(USERS_PAGE)
    end

    DittoCode::Exec.if 'COMMUNITY' do
      @users = User.unscoped.order('LOWER(name) ASC').take(USERS_PAGE)
    end

    # The current page and current letter
    @page = 0
    @current_letter = ''

    respond_to do |format|
      format.html {}
      format.js {}
      format.json { render json: @users }
    end
  end

  #
  # Show info of an user.
  #
  def show
    @user = User.find(params[:id])
  end

  #
  # Show a form to create a new user.
  #
  def new
    @user = User.new(sensor_id: current_user.sensor.id)
  end

  #
  # Show a view with a form to edit the user
  #
  def edit
    # Check if the user to edit is the current user
    @minimum_password_length = (Setting.find_by(name: 'minimum_password_length').try(:value) || 12).to_i
    redirect_to edit_user_registration_path if @user == current_user
  end

  #
  # Update an user.
  #
  def update
    if params[:user][:password].blank? && params[:user][:password_confirmation].blank?
      params[:user].delete(:password)
      params[:user].delete(:password_confirmation)
    end
    if @user.update(user_params)
      @user.update_attribute(:pwd_changed_at, Time.now)
      redirect_and_notice "#{I18n.t('controllers.user_updated')}", redirect_to: users_path
    else
      @user.password = @user.password_confirmation = nil
      redirect_and_error @user, redirect_to: edit_user_path(@user)
    end
  end

  #
  # Destroy current user
  #
  def destroy
    @user.destroy
    redirect_and_notice "#{I18n.t('controllers.user_deleted')}", redirect_to: users_path
  end

  #
  # Show users that can be shared
  #
  def list
    @users = current_user.users_to_share
  end

  #
  # Load more users based on number page
  #
  def load_more
    # Get the page
    @page = params[:page].to_i
    @current_letter = params[:last_letter]

    @current_letter = '' if @current_letter.nil?

    # Get the users
    if current_user.super_admin?
      @users = User.unscoped.offset(@page * USERS_PAGE)
               .order('LOWER(name) ASC').take(USERS_PAGE)
    else
      @users = User.unscoped
               .where(sensor_id: current_user.domains_with_permission.map(&:id))
               .offset(@page * USERS_PAGE).order('LOWER(name) ASC')
               .take(USERS_PAGE)
    end

    respond_to do |format|
      format.js
      format.json { render json: @users }
    end
  end

  #
  # Search the user based on the first letter. It's receive a letter.
  #
  def letter_search
    # Get the letter
    @letter = params[:id]
    @current_letter = ''

    # Check the letter
    if [*'A'..'Z'].include? @letter
      @users = User.unscoped.find_by_first_letter(@letter)
               .where(sensor_id: current_user.domains_with_permission.map(&:id))
               .order('LOWER(name) ASC')
      render :search
    else
      # Something went wrong
      redirect_to users_path
    end
  end

  #
  # Search a user based on a given String.
  #
  def search
    @search = params[:search]
    @users =
      users_can_see
      .includes(:sensor)
      .where(search_user_query, search: "%#{@search.downcase.split.join('%')}%")
      .references(:sensor).order('LOWER(users.name) ASC')
  end

  #
  # Force recreate API Key of an user
  #
  def recreate_api_key
    @user_edited.reset_authentication_token!
  end

  #
  # Show a time selector modal to update Time Machine of the user
  #
  def time_selector
    render layout: false
  end

  #
  # Save columns of user
  #
  def save_columns
    controller = params[:url][:controller].singularize.to_sym
    action  = if session[controller][:default_action].present?
                session[controller][:default_action][:action].to_sym
              else
                :tops
              end

    columns = case action
              when :index
                :active_columns
              when :tops
                :active_tabs
              when :cardinality
                :active_cardinality_tabs
              end

    current_user.properties[controller][action] = session[controller][columns]
    if current_user.save!
      flash[:notice] = "#{I18n.t 'controllers.tabs_saved'}"
    else
      flash[:alert] = "#{I18n.t 'controllers.tabs_not'}"
    end

    render js: "$('li.open').removeClass('open');"
  end

  private

  #
  # String to search the users by name, email or sensor name
  #
  # == Returns:
  # String of SQL query
  #
  def search_user_query
    'LOWER(users.name) LIKE :search OR LOWER(users.email)
     LIKE :search OR LOWER(sensors.name) LIKE :search'
  end

  #
  # Users can see current user
  #
  # == Returns:
  # ActiveRelation of Users.
  #
  def users_can_see
    return User.unscoped if current_user.super_admin?
    # Check permissions
    User.unscoped
      .where(sensor_id: current_user.domains_with_permission.map(&:id))
  end

  #
  # Filter params to edit/create an user.
  #
  # == Returns:
  # Hash with fields based on request params.
  #
  def user_params
    allowed_parameters = [
      :name, :login, :email, :password, :password_confirmation,
      :current_password, :per_page_count, :root_page, :properties,
      :sensor_id, :letter_size, :preferred_locale, :pwd_changed_at,
      :failed_attempts, :locked_at, :force_pwd_change
    ]

    allowed_parameters << :enabled if current_user.super_admin?

    if current_user.super_admin? ||
       (current_user.admin? &&
        current_user.sensor.self_and_descendants.include?(@user.sensor))
      allowed_parameters << :admin
    end

    params.require(:user).permit(*allowed_parameters)
  end

  #
  # Check if the current user has permissions on the top
  # domain to assign to another user.
  #
  # == Returns:
  # Redirect the user to #index with an error message if current_user
  # has not permissions.
  #
  def check_domain_permission
    return unless params[:user][:sensor_id]
    return if current_user
              .domains_with_manage_permission
              .exists?(params[:user][:sensor_id].to_i)
    no_enough_privileges redirect_to: users_path
  end

  #
  # Check if current_user has permission to edit user.
  #
  # == Returns:
  # Redirect the user to #index with an error message if current_user
  # has not permissions.
  #
  def set_user_and_check_permission
    @user = User.unscoped.find(params[:id])
    # Only super admins can destroy another super_admins
    return if (current_user.super_admin? && @user.super_admin?) ||
              (current_user.admin_on?(@user.sensor) && !@user.super_admin?)
    # User has not privileges
    return error_404("#{I18n.t('controllers.user_dont_exist')}")
  end

  #
  # To recreate the API key check if the user is current or it's a super admin
  #
  # == Returns:
  # True or redirect the user to #index with an error message
  #
  def require_recreate_api_key_privileges
    @user_edited = params[:user] ? User.find(params[:user]) : current_user
    return true if current_user.super_admin? || current_user.eql?(@user_edited)
    # User has not privileges
    no_enough_privileges redirect_to: users_path
  end
end
