#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module IncidentsHelper
  def incidents_uniq_created_at 
    ["Last Hour", "Last 2 hours", "Last 3 hours", "Last 12 hours", "Last 24 hours", "Last Week", "Last 2 Weeks"]
  end

  def incidents_uniq_users
    User.joins(:incident_users).distinct
  end

  def incidents_uniq_sources
    Incident.pluck(:source).uniq
  end

  def incidents_order_by
    [['ID', 'id'], [I18n.t('incidents_helper.name'), 'name'], [I18n.t('incidents_helper.source'), 'source'], [I18n.t('incidents_helper.created_at'), 'created_at'], [I18n.t('incidents_helper.assigned'), 'assigned_user']]
  end

  def incidents_order_direction
    [[I18n.t('incidents_helper.ascending'), 'asc'], [I18n.t('incidents_helper.descending'), 'desc']]
  end

  def worklog_uniq_types(user = current_user)
    [['Notes','Notes'], ['Incident Changes', 'Incident Changes'], ['Response Actions', 'Response Actions']]
  end

  def worklog_uniq_users(worklog_data, user = current_user)
    filtered_worklog_data = worklog_data.reject { |entry| entry[:user].nil? || entry[:user].is_a?(String)}
    filtered_worklog_data.pluck(:user).uniq || []
  end

  def worklog_uniq_orders(user = current_user)
    [['Newest','Newest'], ['Oldest', 'Oldest']]
  end

  def priority_color(priority_name)
    case priority_name.downcase
    when "low"
      "#ffffff"
    when "medium"
      "#ffffff"
    when "high"
      "#ffffff"
    when "critical"
      "#ffffff"
    when "none"
      "#ffffff"
    when "unknow"
      "#ffffff"
    when "info"
      "#ffffff"
    when "warning"
      "#ffffff"
    when "notice"
      "#ffffff"
    when "error"
      "#ffffff"
    else
      "#ffffff"
    end
  end

  def priority_bgcolor(priority_name)
    case priority_name.downcase
    when "low"
      "#89a54e"
    when "medium"
      "#c69958"
    when "high"
      "#aa4643"
    when "critical"
      "#ff4643"
    when "none"
      "#30BF93"
    when "unknow"
      "#919191"
    when "info"
      "#5eca5e"
    when "warning"
      "#ffa500"
    when "notice"
      "#3399ff"
    when "error"
      "#ff0000"
    else
      "#000"
    end
  end

  def get_module_name_by_source(source)
    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:incident_source_name] && v[:incident_source_name] == source }

    incident_modules.first
  end

  def get_incident_filters_link(incident)
    return nil unless incident.source

    mod = get_module_name_by_source(incident.source)

    return nil unless mod

    module_name = mod.first

    hash = {}

    hash[0] = {
        "column" => "incident_uuid",
        "operator" => "is",
        "value" => incident.uuid,
        "enabled" => true
    }

    start_time = incident.created_at - 1.hour
    
    hash[1] = {
      "column" => "start_time",
      "operator" => "is",
      "value" => start_time,
      "enabled" => true
    }
    
    end_time = incident.created_at + 1.day
    hash[2] = {
      "column" => "end_time",
      "operator" => "is",
      "value" => end_time,
      "enabled" => true
    }

    json_string = hash.to_json
    encoded_json = CGI.escape(json_string)
      
    "/#{module_name}/tops?search=#{encoded_json}"
  end

  def get_detection_element_filters_link(element)
    hash = {}

    element[:url_filters].each_with_index do |e, index|
      field = e.first
      value = e.last

      hash[index] = {
        "column" => field,
        "operator" => "is",
        "value" => value,
        "enabled" => true
      }
    end

    json_string = hash.to_json
    encoded_json = CGI.escape(json_string)
    
    "/#{element[:module_name]}/tops?search=#{encoded_json}"
  end

  # Dynamic mapping method
  def map_dynamic_fields(data, field_map)
    field_map.each do |key, possible_fields|
      possible_fields.each do |possible_field|
        if data.methods.include?(possible_field)
          value = data.send(possible_field)
          return value unless value.nil?
        end
      end
    end
    'N/A'
  end

  def get_incident_druid_data(incident, search_params, user_id)
    # set user_id
    user = User.find_by(id: user_id) || User.first

    data = []

    # Mapping of fields that can change depending on the module or data source
    field_map = {
      priority: [:priority, :syslogseverity_text], # Priority can be one of these fields
      msg: [:msg, :message],                       # Msg can be one of these fields
    }

    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }
    incident_modules.each do |mod|
      source = Redborder::DruidSource.get(source: mod.first)
      filter = Redborder::Filter.new(source, {})
      filter.add(:incident_uuid, incident.uuid)

      start_time = incident.created_at - 1.hour
      end_time = start_time + 1.day

      # group_by: @source.default_columns_raw,
      data += source.where({
        query_type: :group_by,
        range: :custom,
        start_time: start_time,
        end_time: end_time,
        granularity: :minute,
        aggregation: source.default_aggregation,
        group_by: source.default_columns_incidents,
        search: filter.get,
        user: user

      }).map do |t|
        url_filters = {}
        url_filters[:start_time] = start_time
        url_filters[:end_time] = end_time

        # We use the map_dynamic_fields method to map the fields correctly
        priority = map_dynamic_fields(t, priority: field_map[:priority])
        msg = map_dynamic_fields(t, msg: field_map[:msg])

        sig_id = 'N/A'
        if t.methods.include?(:sig_id)
          sig_id = t.send(:sig_id) || 'N/A'
        end

        source_name = mod.last[:incident_source_name] rescue 'N/A'
        module_name = mod.first
        timestamp = t.send(:timestamp)

        # Observables are define in the columns of each module
        observables = []
        assets = []
        mod.last[:columns].select { |column, cfg| cfg[:observable_type] }.each do |column, cfg|
          asset = nil
          url_filters[column] = t.send(column)

          # We check if this is an asset
          if cfg[:object_type] && t.methods.include?(column) && t.send(column)
            asset = RedborderObject.find_by(value: t.send(column), type: cfg[:object_type], user: User.find_by(id: user_id))
          end

          if asset
            assets.push({ t.send(column) => cfg[:object_type] }) if t.methods.include?(column) && t.send(column)
          else
            observables.push( { t.send(column) => cfg[:observable_type] }) if t.methods.include?(column) && t.send(column)
          end
        end

        url_filters['incident_uuid'] = incident.uuid

        { timestamp: timestamp, sig_id: sig_id, priority: priority, msg: msg, source: source_name, assets: assets, observables: observables, module_name: module_name, url_filters: url_filters }
      end
    end

    # Filter by search
    if search_params && search_params[:source]
      data = data.select { |element| element[:source] == search_params[:source] }
    end

    if search_params && search_params[:priority]
      data = data.select { |element| element[:priority] == search_params[:priority] }
    end

    data
  end

  def syslog_configured?
    syslog_strategy = Integrations::SyslogStrategy.new
    params = syslog_strategy.saved_params
    required_keys = ["syslog_server"]
  
    required_keys.all? { |key| params[key.to_sym].present? }
  end

  def is_ai_running?(section, incident_id)
    Delayed::Job.where("handler LIKE ?", "%Redborder::Jobs::AIGenerateContentJob%")
                .where("handler LIKE ?", "%#{section}%")
                .where("handler LIKE ?", "%#{incident_id}%").exists?
  end
end