#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ObjectType < ApplicationRecord
  class Unknown < StandardError; end

  # -------------------------------------------------------------
  # Constants
  # ---------------------------------------------------------------
  OBJECT = %w(Name) 

  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------
  
  validates :name,
            presence: true,
            format: { with: /\A[\p{L}0-9\s_\-\/\.]*\z/,
                      message: "#{I18n.t('models.alphanumerical')}" }
  validates_uniqueness_of :name, message: "#{I18n.t('models.name_unique')}"
  validates_uniqueness_of :comment, message: "#{I18n.t('models.comment_unique')}"

  validates :comment,
            presence: true,
            format: { with: /\A[\p{L}0-9\s_\-\/\.]*\z/,
                      message: "#{I18n.t('models.alphanumerical')}" }

  


  # -------------------------------------------------------------
  # Class methods
  # -------------------------------------------------------------

  #
  # Create object type from a file. This method open the file and process all
  # elements. If exist an object with same name, it will be updated.
  #
  # == Parameters:
  # content_file::
  #   String with the path of the file
  # tab::
  #   String with current tab to set the correct type of the objects.
  #
  # == Returns:
  # Array with number of created elements and updated.
  #
  def self.replace_vars_from_file(content_file)
    created_count = 0
    updated_count = 0
  
    CSV.foreach(content_file.path, col_sep: ';', headers: false, encoding: 'UTF-8') do |row|
      # Ignorar filas vacías
      next if row.empty? || row[0].nil?
  
      name = row[0].try(:strip)
      comment = row[1].try(:strip) 
  
      next if name.nil? || name.empty?
  
      object = ObjectType.find_or_initialize_by(name: name)
  
      if object.new_record?
        object.comment = comment
        object.save
        created_count += 1
      else
        if object.comment != comment
          object.update(comment: comment)
          updated_count += 1
        end
      end
    end
  
    [created_count, updated_count]
  end

  #
  # Replace objects from a file. This method open the file and process all
  # elements.
  #
  # == Parameters:
  # content_file::
  #   String with the path of the file
  # tab::
  #   String with current tab to set the correct type of the objects.
  #
  # == Returns:
  # Array with number of created elements and updated.
  #
  def self.add_vars_from_file(content_file)
    created_count = 0
    updated_count = 0
  
    CSV.foreach(content_file.path, col_sep: ';', headers: false, encoding: 'UTF-8') do |row|
      # Ignorar filas vacías
      next if row.empty? || row[0].nil?
  
      name = row[0].try(:strip)
      comment = row[1].try(:strip)
  
      next if name.nil? || name.empty?
  
      object = ObjectType.find_or_initialize_by(name: name)
      
      if object.new_record?
        object.comment = comment
        object.save
        created_count += 1
      elsif object.comment != comment
        object.update(comment: comment)
        updated_count += 1
      end
    end
  
    [created_count, updated_count]
  end


  #
  # Find elements based on a query parameter in name and value fields
  #
  # == Parameters:
  # query::
  #   String to search
  # user::
  #   User needed in isolated caches
  #
  # == Returns:
  # ActiveRelation
  #
  def self.search(query, user = nil)
    return ObjectType.none if isolated? && user.nil?
    term = "%#{query.downcase.split.join('%')}%"

    if isolated?
      where('(LOWER(name) like ? OR LOWER(value) LIKE ?) AND user_id = ?',
            term, term, user.id)
    else
      where('LOWER(name) like ? OR LOWER(value) LIKE ?', term, term)
    end
  end

  #
  #
  #
  def self.from_string(object_type)
    if ObjectType.include?(object_type)
      object_type.underscore.pluralize
    else
      fail Unknown
    end
  end

  #
  # Save in chef
  #
  def update_app_databag
    UpdateObjectDatabag.call(
      { name.to_s => { name: name,
                       value: value
                     } },
       self.class,
       self.class.to_s.gsub("Object","s").downcase,
      destroyed?
    )
  end
end
