#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Report < ApplicationRecord
  include Exportable

  # ==========================================================
  # Custom attributes
  # ==========================================================

  attr_accessor :collaborators
  attr_accessor :sensors_collaborators
  exportable_opts relations: [:widgets]

  # ==========================================================
  # Relations
  # ==========================================================

  # Has many sensors and users
  has_many :report_users, dependent: :destroy
  has_many :report_sensors, dependent: :destroy
  has_many :users, through: :report_users
  has_many :sensors, through: :report_sensors

  # Created by an user
  # Alias
  belongs_to :owner, class_name: 'User', foreign_key: :user_id,
                     inverse_of: :dashboards
  belongs_to :user

  # Has many report blocks.
  has_many :widgets, as: :parent, dependent: :destroy

  # Has many report historicals.
  has_many :reports_historicals

  # We receive the days like a serialized array
  serialize :schedule_days

  # ================================================================
  # Callbacks
  # ================================================================

  before_save :save_collaborators
  before_save :save_sensors_collaborators

  after_initialize :defaults

  # ================================================================
  # Constants
  # ================================================================

  ORIENTATION_PORTRAIT = 1
  ORIENTATION_LANDSCAPE = 2

  # ================================================================
  # Validations
  # ================================================================

  # The name can't be empty
  validates_presence_of :name

  # Only allow one report per user on community version
  DittoCode::Exec.if 'COMMUNITY' do
    validates_uniqueness_of :user_id
  end

  # ================================================================
  # URL helpers
  # ================================================================

  delegate :url_helpers, to: 'Rails.application.routes'

  # ================================================================
  # Methods
  # ================================================================

  def pages
    (1..number_pages).each_with_object([]) do |page, memo|
      memo << widgets.select { |w| w.page == page }
    end
  end

  def each_page
    (1..number_pages).each do |page|
      yield page
    end
  end

  def blocks_data(options = {})
    blocks = {}
    each_page do |page|
      blocks[page] = widgets.select { |w| w.page == page }.map do |widget|
        if options[:start_time] && widget.product_type != 'none'
          widget.granularity = options[:granularity] if widget.respond_to?('granularity')
          if widget.product_type != 'all'
            source = Redborder::DruidSource.get(source: widget.product_type)
            f = Redborder::Filter.new(source)
            f.parse(widget.filters, true)
            f.add(:start_time, options[:start_time])
            f.add(:end_time, options[:end_time])
            widget.filters = f.serialize(true)
          else
            series_hash = JSON.parse widget.series
            series_hash.each_key do |serie|
              source = Redborder::DruidSource.get(source: series_hash[serie]['mod'])
              f = Redborder::Filter.new(source)
              f.parse(JSON.generate(series_hash[serie]['filters']), true)
              f.add(:start_time, options[:start_time])
              f.add(:end_time, options[:end_time])
              series_hash[serie]['filters'] = f.serialize(true)
            end
            widget.series = JSON.generate series_hash
          end
          widget
        else
          widget
        end
      end
    end
    blocks
  end

  # Get the name of the orientation
  def orientation_name
    case orientation
    when ORIENTATION_PORTRAIT
      'Portrait'
    when ORIENTATION_LANDSCAPE
      'Landscape'
    end
  end

  # Check if the current report has a portrait orientation
  def portrait?
    (orientation == ORIENTATION_PORTRAIT)
  end

  # Check if the current report has a portrait orientation
  def landscape?
    (orientation == ORIENTATION_LANDSCAPE)
  end

  # ================================================================
  # Define the type for widgets
  # ================================================================

  #
  # Check if the class is a dashbaord or a report
  #
  # == Returns:
  # False, this is a report
  #
  def dashboard?
    false
  end

  #
  # Check if the class is a dashbaord or a report
  #
  # == Returns:
  # True, this is a report
  #
  def report?
    true
  end

  # ================================================================
  # Permissions for users. We MUST implements this methods.
  # They are required for widgets
  # ================================================================

  #
  # Check if the user has permissions on this report.
  #
  # == Paramenters:
  # user:
  #   User who wants to edit a report
  #
  # == Returns:
  # A boolean
  #
  def permission?(user = self.user)
    return false if user.nil?
    user.writable_reports.include? self
  end

  #
  # Check if the user has read permission on this report.
  #
  # == Paramenters:
  # user:
  #   User who wants to edit a report
  #
  # == Returns:
  # A boolean
  #
  def read_permission?(user = self.user)
    return false if user.nil?
    user.readable_reports.include? self
  end

  #
  # Main URL to redirect when the permissions are wrong
  #
  # == Returns:
  # An url to redirect
  #
  def main_path
    url_helpers.reports_path
  end

  # ================================================================
  # Import/export methods for exportable concern
  # ================================================================

  #
  # => The attributes of this object that will be exported
  #
  def export_json
    to_json(except: [
      :id, :user_id, :created_at, :updated_at, :image_content_type,
      :image_file_size, :image_updated_at, :content_type, :time
    ])
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  def save_collaborators
    # Remove all relations, except those with the owner and the current user
    report_users.destroy_all
    return if collaborators.nil?

    # Dont save relations with the owner or the current user
    collaborators.each do |user|
      report_users.build(
        user_id: user['user_id'],
        can_edit: (user['can_edit'] == 'on')
      )
    end
  end

  def save_sensors_collaborators
    # Remove all relations
    report_sensors.destroy_all
    return if sensors_collaborators.nil?

    # Save all relations
    sensors_collaborators.each do |_, sensor|
      report_sensors.build(
        sensor_id: sensor['sensor_id'],
        can_edit: sensor['can_edit']
      )
    end
  end

  def defaults
    self.name          ||= 'Default'
    self.description   ||= 'Default Report'
    self.number_pages  ||= 1
    self.orientation   ||= Report::ORIENTATION_PORTRAIT
  end
end
