#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Widget < ApplicationRecord
  include StorageFile

  # ==========================================================
  # Relations
  # ==========================================================
  belongs_to :parent, polymorphic: true

  # ================================================================
  # Callbacks
  # ================================================================

  before_save :remove_disabled_filters
  before_create :default_positions
  after_initialize :defaults

  # Set pages always integers on reports
  before_save :check_page, if: proc { |w| w.report? }

  # ================================================================
  # Constants
  # ================================================================

  COMMON_PROPERTIES = [
    :widget_type, :product_type, :graph_type, :click_action,
    :filters, :click_dashboard, :position_x, :position_y,
    :position_width, :position_height, :transparent, :page, :series
  ]

  # Default positions for dashboards
  POSITIONS = {
    position_x: {
      columns: 1,
      absolute: '15px',
      standard: 200
    },
    position_y: {
      columns: 1,
      absolute: '15px',
      standard: 120
    },
    position_width: {
      columns: 2,
      absolute: '400px',
      standard: 200
    },
    position_height: {
      columns: 2,
      absolute: '240px',
      standard: 120
    }
  }

  NOT_EXPORTABLE_ATTRIBUTES = [
    :dashboard_id, :created_at, :updated_at,
    :media_content_type, :media_file_size,
    :media_file_name, :media_updated_at
  ].freeze

  NO_RANGE_WIDGET = ['compare'].freeze

  # Constants
  DASHBOARD_COLUMNS = 8

  # ==========================================================
  # Custom attributes and scopes
  # ==========================================================

  store :property, accessors: COMMON_PROPERTIES
  attr_accessor :time_now, :graph, :wtype, :stream_columns_ordered
  cattr_accessor :sections

  self.sections = {
    infrastructure: {
      name: 'Infrastructure',
      elements: [
        {
          id: :sensor,
          name: 'Sensor',
          icon: 'fa-eye',
          report: false
        }, {
          id: :cluster,
          name: 'Cluster',
          icon: 'fa-sitemap',
          report: false,
          super_user: true
        }, {
          id: :alarm,
          name: 'Alarm',
          icon: 'fa-bell-o',
          report: false
        }
      ]
    },
    mitre: {
      name: 'Mitre ATT&CK',
      elements: [
        {
          id: :tactics,
          name: 'Tactics',
          icon: 'fa-tasks',
          report: false
        },
        {
          id: :techniques,
          name: 'Techniques',
          icon: 'fa-sort-amount-asc',
          report: false
        },
        {
          id: :subtechniques,
          name: 'Subtechniques',
          icon: 'fa-sort-desc',
          report: false
        },
      ]
    }
  }

  # ================================================================
  # Validations
  # ================================================================

  validates :click_action, inclusion: { in: %w(details dashboard none) }

  validates :graph_type,
            inclusion: { in: proc { WidgetGraph.all.keys.map(&:to_s) } }

  validates :widget_type,
            inclusion: { in: proc { WidgetType.all.keys.map(&:to_s) } }

  validates :parent_id, :parent_type, presence: true

  # ================================================================
  # Validations methods
  # ================================================================

  #
  # Check page is an integer
  #
  def check_page
    self.page = page.to_i
  end

  # ================================================================
  # Class methods
  # ================================================================

  def self.properties
    @properties ||= []
    COMMON_PROPERTIES + @properties
  end

  def self.add_properties(new_properties = [])
    @properties ||= []
    @properties += new_properties
  end

  def self.array_properties
    @array_properties ||= []
  end

  def self.add_array_properties(new_properties = [])
    @array_properties ||= []
    @array_properties += new_properties
  end

  # ================================================================
  # Public methods
  # ================================================================

  def cleaned_name
    ret = name.downcase
    ret = ret.tr(/\s+/, '_')
    ret = ret.tr(/\.|\,/, '-')
    ret = ret.tr(/[^a-zA-Z0-9_-]/, '')
    ret
  end

  def dup
    new_widget = super
    new_widget.media = media
    new_widget
  end

  def print_filters
    return '' unless respond_to? "print_filters_#{product_type}"
    send("print_filters_#{product_type}")
  end

  def graph_type=(which)
    graph_type_or_default = (which || :none).to_sym
    graph_class =
      WidgetGraph.all.fetch(graph_type_or_default,
                            WidgetGraph.all[:none])[:class]
    self.graph = graph_class.new(self)
    super(which)
  end

  def widget_type=(which)
    widget_type_or_default = (which || :none).to_sym
    widget_type_class =
      WidgetType.all.fetch(widget_type_or_default,
                           WidgetType.all[:none])[:class]
    self.wtype = widget_type_class.new(self)
    super(which)
  end

  def combinational_serie?
    self.graph_type == 'combination'
  end

  def as_json(opts = {})
    hash = super(opts).merge(
      'graph' => graph.as_json(opts),
      'wtype' => wtype.as_json(opts)
    ).dup

    hash['property'] = hash['property'].dup
    hash['property']['filters'] =
      JSON.parse(hash['property']['filters'] || '{}')
    hash['property']['series'] =
      JSON.parse(hash['property']['series'] || '{}')
    hash
  end

  def move_to(x, y)
    if dashboard? && parent.absolute?
      self.position_x = "#{x}px"
      self.position_y = "#{y}px"
    else
      self.position_x = x
      self.position_y = y
    end

    [position_x, position_y]
  end

  def displace(x, y)
    move_to(position_x.to_i + x, position_y.to_i + y)
  end

  def data
    return if graph.data.nil?
    data = {}
    graph.data.each do |d|
      begin
        data.merge!(wtype.send(d.to_s, self.parent.owner, { include_all_namespaces: true }))
      rescue ArgumentError
        # If wrong number of arguments try to call without params
        data.merge!(wtype.send(d.to_s))
      end
    end
    # Return the data
    data
  end

  #
  # Check if this widget can be exported to a report
  #
  def can_report?
    # Now, only standard widgets can be added to reports.
    return true if %w(top raw compare single_unique grouped_unique monitor
                      map heatmap hardwarestatus hardwareinfo).include?(widget_type)
    return true if widget_type == 'shape' && graph_type != 'custom'

    # Can't add to a report
    false
  end

  # ================================================================
  # Filters
  # ================================================================

  #
  # Check if this widget is filtrable based on product_type
  #
  # == Returns:
  # True if the widget is filtrable
  #
  def filtrable?
    return true if product_type != 'none' || product_type == 'all'
    false
  end

  #
  # Get allowed columns to filter based on the product_type of the widget
  #
  # == Returns:
  # An array of Symbols with the allowed columns
  #
  def filter_columns
    fail "#{I18n.t('models.not_filtrable')}" unless filtrable?
    source = Redborder::DruidSource.get(source: product_type)
    source.allowed_columns
  end

  #
  # Get the filters converted to Druid format.
  #
  # == Returns:
  # A Hash with the filters to send to druid.
  #
  def fetch_filters
    fail "#{I18n.t('models.not_filtrable')}" unless filtrable?

    if product_type.eql?('all')
      JSON.parse(series).values.each do |value|
        source = Redborder::DruidSource.get(source: value['mod'])
        f = Redborder::Filter.new(source)
        # Load filters
        return {} unless value['filters']
        f.parse(value['filters'].to_json)
      end
    else
      source = Redborder::DruidSource.get(source: product_type)
      f = Redborder::Filter.new(source)
      # Load filters
      return {} unless filters
      f.parse(filters)
    end
  end

  def apply_global_filters(globals)
    return if globals.nil?
    filters = globals.deep_dup

    if filters[:range] && NO_RANGE_WIDGET.exclude?(widget_type)
      self.range = filters.delete(:range)
      self.granularity = WidgetGraph::Base::DEFAULT_GRANULARITY[range.to_sym]
    end

    begin
      source = Redborder::DruidSource.get(source: product_type)
      f = Redborder::Filter.new(source)
      filters.each do |k, v|
        if k == :advanced
          # Advanced filters
          v.each do |_i, adv_filter|
            next unless adv_filter['enable']
            f.add(adv_filter['column'], [adv_filter['value']])
          end
        else
          # Regular filters
          f.add(k.to_s, [v.to_s])
        end
      end
      w_filters = JSON.parse(self.filters || '{}')
      w_filters.merge!(f.dump(index: w_filters.keys.size))
      self.filters = w_filters.to_json
    rescue NotImplementedError
      # If there is no source for the product type skip the widget's filters
    end
  end

  # ================================================================
  # Define the type for the parent. This is an alias for parent
  # methods
  # ================================================================

  #
  # Get the height of the block for widgets
  #
  # == Parameters:
  # p:
  #   The parent of thw widget. This is user do prevetn an extra SQL query.
  #
  # == Returns:
  # The height for these blocks
  #
  def grid_height(p = parent)
    p.dashboard? ? 110 : 73
  end

  #
  # Check if the parent is a dashboard
  # @see Dashboard#dashboard?
  #
  # == Returns:
  # Boolean, true if parent is a dashboard
  #
  def dashboard?
    parent.dashboard?
  end

  #
  # Check if the parent is a report
  # @see Report#report?
  #
  # == Returns:
  # Boolean, true if parent is a report
  #
  def report?
    parent.report?
  end

  # ================================================================
  # Import/export methods for exportable concern
  # ================================================================

  #
  # => The attributes of this object that will be exported
  #
  def export
    attributes.except(NOT_EXPORTABLE_ATTRIBUTES)
  end

  #
  # => The attributes of this object that will be exported in JSON
  #
  def export_json
    to_json(no_methods: true, except: NOT_EXPORTABLE_ATTRIBUTES)
  end

  #
  # => Called when the object will be exported to get a media file
  #    (img, video, audio, etc) that will be exported as well
  #
  def export_media
    return unless %(hardwarestatus image video).include?(graph_type) && media?
    open(media_uri, ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE).read
  end

  #
  # => This method is called when a Dashboard is imported, for every
  #    widget in that dashboard, so we can customize the method used
  #    to save the widgets into the database
  #
  def import_callback(params, media)
    params = params.merge(params['property']).except('property', 'id')

    if %(image video).include?(params['graph_type']) && !media.nil?
      self.media = media
    end

    params
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  def defaults
    self.graph_type      = graph_type  || 'none'
    self.widget_type     = widget_type || 'raw'
    self.name            ||= self.widget_type.titleize
    self.click_action    ||= 'details'
    self.transparent     ||= false
  end

  def remove_disabled_filters
    return unless filters
    temp_filters = JSON.parse(filters) rescue filters

    enabled_filters = temp_filters.select do |_key, value|
      value['enabled']
    end

    self.filters = enabled_filters.to_json
  end

  # Set the next position in the current parent
  #
  def next_default_column_position
    # Get last positioned widget
    lw =
      parent.widgets.sort_by { |w| [w.position_y + w.position_height, w.position_x] }.last
    # The default width is 2, check if the element enter in the last row.
    if lw.nil?
      self.position_x ||= POSITIONS[:position_x][:columns]
      self.position_y ||= POSITIONS[:position_y][:columns]
    elsif lw.position_x + lw.position_width <
          DASHBOARD_COLUMNS - (POSITIONS[:position_width][:columns] - 1)
      # We can set the widget in the same row
      self.position_x ||= lw.position_x + lw.position_width
      self.position_y ||= lw.position_y
    else
      # Next row
      self.position_x ||= POSITIONS[:position_x][:columns]
      self.position_y ||= lw.position_y + lw.position_height
    end
    self.position_width ||= POSITIONS[:position_width][:columns]
    self.position_height ||= POSITIONS[:position_height][:columns]
  end

  #
  # => Sets the position variables.
  #
  def default_positions
    if parent.report? || parent.layout == 'columns'
      clear_positions if position_x.is_a?(String) || position_y.is_a?(String)
      next_default_column_position
    else
      clear_positions if position_width.is_a?(Fixnum) || position_height.is_a?(Fixnum)
      self.position_x      ||= "#{50 + rand(100)-50}px"
      self.position_y      ||= "#{100 + rand(100)-50}px"
      self.position_width  ||= "400px"
      self.position_height ||= "180px"
    end
  end

  def clear_positions
    self.position_x = nil
    self.position_width = nil
    self.position_y = nil
    self.position_height = nil
  end
end
