#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# Import required libraries
require 'resolv'
require 'rest-client'
require 'json'
require 'time'
require 'logger'
require 'active_record'
require 'zk'
require 'set'

# Class for application logging
class AppLogger
  def self.logger
    # Creates a logger that outputs to the console and sets the log level to DEBUG
    @logger ||= Logger.new(STDOUT).tap { |log| log.level = Logger::DEBUG }
  end
end

class Formatter
  def self.break_line
    "-" * 80 + "\n"
  end
end

# Main class responsible for asset discovery
class AssetsDiscovery
  # Retrieves the Druid broker URL
  def self.druid_url
    @druid_url ||= retrieve_druid_broker_url
  end

  # Fetches the Druid broker URL from Zookeeper
  def self.retrieve_druid_broker_url
    # Reads the Zookeeper configuration from the YAML config file
    zk_data = YAML.load(File.read('config/rbdruid_config.yml'))[Rails.env].symbolize_keys
    zk_connect = zk_data[:zk_connect]
    discovery_path = '/druid/discoveryPath/broker'

    # Retrieves the available brokers
    brokers = fetch_brokers(zk_connect, discovery_path)
    raise "No active Druid Broker found." if brokers.empty?

    # Returns the first available broker
    brokers.first
  end

  # Queries Zookeeper for active brokers
  def self.fetch_brokers(zk_connect, discovery_path)
    brokers_names = []

    # Opens a connection to Zookeeper and fetches brokers from the discovery path
    ZK.open(zk_connect) do |zk|
      zk.children(discovery_path).shuffle.each do |broker|
        zktdata = YAML.load(zk.get("#{discovery_path}/#{broker}").first)
        # Checks if the broker has a valid address and port
        if zktdata["address"] && zktdata["port"]
          brokers_names << "http://#{zktdata['address']}:#{zktdata['port']}/druid/v2/?pretty"
        end
      end
    end

    brokers_names
  end

  def initialize
    # Initialize control variables
    @messages_printed = Hash.new(false)
    @mac_ip_mapping = {}
    @result = ""
  end

  # Main method to execute the asset inventory job
  def execute
    @result += "Starting Assets Inventory Job...\n"

    # Establishes a connection to the database
    conn = establish_db_connection

    if conn
      # Fetches data sources from sensors
      data_sources = fetch_data_sources_from_sensor
      # Processes data from the sources and returns the total number of records processed
      total_count = process_data(conn, data_sources)
      @result += "A total of #{total_count} records (MACs and IPs) have been processed.\n"
    else
      @result += "No database connection was established.\n"
    end

    # Closes the database connection if it was established
    conn.close if conn
    @result
  rescue StandardError => e
    # Logs error message and backtrace in case of an exception
    AppLogger.logger.error("Error: #{e.message}")
    AppLogger.logger.error(e.backtrace.join("\n"))
    "Error: #{e.message}"
  end

  private

  # Establishes a connection to the PostgreSQL database using ActiveRecord
  def establish_db_connection
    ActiveRecord::Base.connection_pool.with_connection do
      @result += "Connection to PostgreSQL established successfully.\n"
      ActiveRecord::Base.connection
    end
  rescue ActiveRecord::ConnectionNotEstablished => e
    @result += "Failed to connect to PostgreSQL: #{e.message}"
    nil
  end

  # Fetches data sources from the sensors
  def fetch_data_sources_from_sensor
    sensors = Sensor.where(domain_type: 8).pluck(:uuid, :name)
    
    sensors.map { |uuid, name| ["rb_flow_#{uuid}", name] }.unshift(["rb_flow", "root"])
  end

  # Processes all the data without unnecessary loops
  def process_data(conn, data_sources)
    start_time, end_time = calculate_time_interval
    @result += "Time range for the query: #{start_time} to #{end_time}\n"
    @result += Formatter.break_line
  
    total_mac_insert_count, total_ip_insert_count = 0, 0
  
    all_mac_addresses = Set.new
    all_ips = Set.new
  
    data_sources.each do |data_source, name|
      # @result += "Searching in domain: #{name}\n"
      AppLogger.logger.debug("Searching in domain: #{name}\n")
      responses = fetch_and_process_data(data_source, start_time, end_time)
  
      responses.each do |dimension, response_json|
        response_json.each do |item|
          client_mac = item.dig('event', 'client_mac')
          lan_ip = item.dig('event', 'lan_ip')
          wan_ip = item.dig('event', 'wan_ip')
  
          if client_mac
            all_mac_addresses.add(client_mac)
            @mac_ip_mapping[client_mac] = resolve_ip(lan_ip || wan_ip)
          end
          
          all_ips.add(lan_ip) if lan_ip
          all_ips.add(wan_ip) if wan_ip
        end
      end
    end
  
    total_mac_insert_count = insert_mac_records(conn, all_mac_addresses)
    total_ip_insert_count = insert_ip_records(conn, all_ips)
  
    @result += "Total MACs inserted: #{total_mac_insert_count}\n"
    @result += "Total IPs inserted: #{total_ip_insert_count}\n"
    @result += Formatter.break_line
  
    total_mac_insert_count + total_ip_insert_count
  end
  
  
  # Fetches data from Druid and processes the results
  def fetch_and_process_data(data_source, start_time, end_time)
    dimensions = %w[lan_ip wan_ip]
    query_responses = {}

    dimensions.each do |dimension|
      query = build_query(data_source, dimension, start_time, end_time)
      druid_url = AssetsDiscovery.druid_url

      begin
        # Sends the query to Druid and parses the JSON response
        response = RestClient.post(druid_url, query.to_json, { content_type: :json })
        parsed_response = JSON.parse(response.body)
        query_responses[dimension.to_sym] = parsed_response unless parsed_response.empty?
      rescue RestClient::ExceptionWithResponse, JSON::ParserError => e
        AppLogger.logger.error("Error querying Druid: #{e.message}")
        AppLogger.logger.error(e.response.body) if e.respond_to?(:response)
        raise
      end
    end

    query_responses
  end

  # Builds the Druid query
  def build_query(data_source, dimension, start_time, end_time)
    {
      queryType: "groupBy",
      dataSource: data_source,
      granularity: "all",
      dimensions: ["client_mac", dimension, "organization_uuid"],
      context: { timeout: 90000, skipEmptyBuckets: true },
      limitSpec: { type: "default", limit: 10000, columns: [{ dimension: "client_mac", direction: "ascending" }] },
      intervals: ["#{start_time}/#{end_time}"],
      filter: {
        type: "regex",
        dimension: dimension,
        pattern: /^(10\.|172\.(1[6-9]|2\d|3[01])\.|192\.168\.|fc[0-9a-fA-F]{2}:|fd[0-9a-fA-F]{2}:)/
      }
    }
  end

  # Inserts MAC records into the database
  def insert_mac_records(conn, mac_addresses)
    mac_addresses = mac_addresses.to_a
    mac_addresses.compact!
    mac_addresses.delete_if(&:empty?)
  
    count = 0
    mac_addresses.each do |mac_address|
      resolved_name = @mac_ip_mapping[mac_address] || 'UnknownName'
      count += 1 if insert_mac_record(conn, mac_address, resolved_name)
    end
    count
  end

  # Inserts IP records into the database
  def insert_ip_records(conn, ips)
    ips = ips.to_a
    ips.compact!
    ips.delete_if(&:empty?)
  
    count = 0
    ips.each do |ip_address|
      resolved_name = resolve_ip(ip_address)
      count += 1 if insert_ip_record(conn, ip_address, resolved_name)
    end
    count
  end

  # Inserts or updates a MAC record in the database
  def insert_mac_record(conn, mac_address, resolved_name)
    return false if mac_address == '00:00:00:00:00:00'
    return false if MacObject.exists?(value: mac_address)
  
    MacObject.transaction do
      mac_object = MacObject.new(value: mac_address, object_type: 'UnknownDevice')
      mac_object.assign_attributes(name: resolved_name, user_id: 1, created_at: Time.current, updated_at: Time.current)
      mac_object.save!
    end
    true
  end

  # Inserts or updates an IP record in the database
  def insert_ip_record(conn, ip_address, resolved_name)
    ip_object = NetObject.exists?(value: ip_address)
    return false if ip_object

    NetObject.transaction do
      ip_object = NetObject.new(value: ip_address, object_type: 'UnknownDevice')
      ip_object.assign_attributes(name: resolved_name, user_id: 1, created_at: Time.current, updated_at: Time.current)
      ip_object.save!
    end
    true
  end

  # Calculates the time interval (last 24 hours) for data queries
  def calculate_time_interval
    current_time = Time.now
    twenty_four_hours_ago = current_time - 24 * 3600
    [twenty_four_hours_ago.iso8601(3), current_time.iso8601(3)]
  end

  # Resolves the domain name for an IP address (DNS lookup)
  def resolve_ip(ip)
    Resolv.getname(ip)
  rescue Resolv::ResolvError
    'UnknownName'
  end
end
