#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class EventFiltersController < ApplicationController
  self.view_paths = 'lib/modules/ips/app/views'
  layout "#{Rails.root}/app/views/layouts/application"
  prepend_view_path "#{Rails.root}/app/views"

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  skip_before_action :authenticate_user!, only: :index, raise: false

  before_action only: :index do |controller|
    if controller.request.format.text?
      validate_cert
      set_policy_by_sensor
    else
      authenticate_user!
      set_policy
    end
  end

  before_action :set_policy, except: [:index, :delete]
  before_action :set_rule, except: [
    :index, :edit_details, :update_details, :delete, :new_event_filter,
    :create_event_filter, :import, :import_file
  ]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def index
    if @policy.nil?
      @suppress_events_rules, @filter_events_rules = [[], []]
    else
      respond_to do |format|
        format.html {
          @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
          @suppress_events_rules = @policy.suppress_events_rules
                                   .paginate(
                                     page: params[:page],
                                     per_page: current_user.per_page_count
                                   )
          @filter_events_rules = @policy.filter_events_rules
                                 .paginate(
                                   page: params[:page],
                                   per_page: current_user.per_page_count
                                 )
        }
        format.js {
          msg = (params[:msg] || '')
          if params[:event_filtering]
            @filter_events_rules = @policy.filter_events_rules(msg)
                                   .paginate(
                                     page: params[:page],
                                     per_page: current_user.per_page_count
                                   )
          elsif params[:event_suppression]
            @suppress_events_rules = @policy.suppress_events_rules(msg)
                                     .paginate(
                                       page: params[:page],
                                       per_page: current_user.per_page_count
                                     )
          end
        }
        format.text {
          @filter_events_rules = @policy.filter_events_rules
          @suppress_events_rules = if @sensor.is_sensor_v2?
                                     @policy.suppress_events_rules_for_ips_v2
                                   else
                                     @policy.suppress_events_rules
                                   end
          render layout: false
        }
        format.csv {
          export_csv
        }
      end
    end
  end

  def new
    @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]

    case params[:type]
    when 'filtering'
      @event_filter = EventFilter.new(
        policy: @policy, filter_type: 'threshold', count: 1, seconds: 60,
        rule_sid: @rule.rule_sid, rule_gid: @rule.rule_gid
      )
      render 'new_event_filtering', layout: false
    when 'suppression'
      @event_filter = EventFilter.new(
        policy: @policy, filter_type: 'suppress', rule_sid: @rule.rule_sid,
        rule_gid: @rule.rule_gid, source_ip: params[:source_ip],
        destination_ip: params[:destination_ip]
      )
      render 'new_event_suppression', layout: false
    end
  end

  def new_event_filter
    @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
    rule_db_version_ids = @policy.rule_db_versions.map(&:id)

    @rules =  if params[:msg] && params[:msg].match(/\A\d+\z/)
                Rule.where(rule_db_version_id: rule_db_version_ids)
                .where('LOWER(rules.msg) like ? OR rules.rule_sid = ?',
                       "%#{params[:msg]}%", "#{params[:msg]}")
                .order(msg: :asc)
                .select(:msg, :rule_sid, :rule_gid)
                .paginate(page: params[:page], per_page: 5)
              elsif params[:msg]
                Rule.where(rule_db_version_id: rule_db_version_ids)
                .where('LOWER(rules.msg) like ?',
                       "%#{params[:msg].downcase.split.join('%')}%")
                .order(msg: :asc)
                .select(:msg, :rule_sid, :rule_gid)
                .paginate(page: params[:page], per_page: 5)
              else
                Rule.where(rule_db_version_id: rule_db_version_ids)
                .order(msg: :asc)
                .select(:msg, :rule_sid, :rule_gid)
                .paginate(page: params[:page], per_page: 5)
              end

    respond_to do |format|
      format.html {
        case params[:type]
        when 'filtering'
          @event_filter = EventFilter.new(
            policy: @policy, filter_type: 'threshold', count: 1, seconds: 60
          )
        when 'suppression'
          @event_filter = EventFilter.new(
            policy: @policy, filter_type: 'suppress'
          )
        end
        render layout: false
      }

      format.js
    end
  end

  def create
    @event_filter = EventFilter.new(event_filter_params)
    @event_filter.user = current_user
    @event_filter.policy = @policy

    if @event_filter.save
      flash[:notice] = "#{I18n.t('controllers.event')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} #{I18n.t('controllers.success_saved')}"
    else
      flash[:error] = @event_filter.errors.full_messages
        .map { |msg| "<li>#{msg}</li>" }.join
    end

    render 'event_filter'
  end

  def create_event_filter
    errors = []

    @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]

    params[:selected_rules].split(',').each do |rule|
      @event_filter = EventFilter.new(event_filter_params)
      @event_filter.user = current_user
      @event_filter.policy = @policy
      @event_filter.rule_sid, @event_filter.rule_gid = rule.split('-')

      if !@event_filter.save &&
         params[:event_filter][:filter_type] == 'suppress'

        rule_db_version_ids = @policy.rule_db_versions.map(&:id)
        msg = Rule.find_by(
          rule_db_version_id: rule_db_version_ids,
          rule_sid: @event_filter.rule_sid,
          rule_gid: @event_filter.rule_gid
        ).try('msg')
        errors << "#{@event_filter.errors.full_messages.join('. ')}"
      elsif !@event_filter.save
        rule_db_version_ids = @policy.rule_db_versions.map(&:id)
        msg = Rule.find_by(
          rule_db_version_id: rule_db_version_ids,
          rule_sid: @event_filter.rule_sid,
          rule_gid: @event_filter.rule_gid
        ).try('msg')
        errors << "#{msg}: #{@event_filter.errors.messages.values.join(' ')}"
      end
    end

    if errors.blank?
      flash[:notice] = "#{@event_filter.filter_type == 'suppress' ? 'Suppresion' : 'Filtering'} #{I18n.t('controllers.events_success')}"
    else
      flash[:errors] = errors.uniq.map { |msg| "<li>#{msg}</li>" }.join
    end

    if @sensor
      redirect_to sensor_policy_event_filters_path(@sensor, @policy)
    else
      redirect_to policy_event_filters_path(@policy)
    end
  end

  def edit
    @event_filter = EventFilter.find(params[:id])

    if @event_filter.filter_type == 'suppress'
      render 'new_event_suppression', layout: false
    else
      render 'new_event_filtering', layout: false
    end
  end

  def edit_details
    @event_filter = EventFilter.find(params[:event_filter_id])

    render layout: false
  end

  def update
    @event_filter = EventFilter.find(params[:id])

    if @event_filter.update(event_filter_params)
      flash[:notice] = "#{I18n.t('controllers.event')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} #{I18n.t('controllers.success_updated')}"
    else
      flash[:error] = "#{I18n.t('controllers.error_while')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'}."
    end

    render 'event_filter'
  end

  def update_details
    @event_filter = EventFilter.find(params[:event_filter_id])

    if @event_filter.update(event_filter_params)
      flash[:notice] = "#{I18n.t('controllers.event')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} #{I18n.t('controllers.success_updated')}"
    else
      flash[:error] = @event_filter.errors.full_messages
        .map { |msg| "<li>#{msg}</li>" }.join
    end
  end

  def destroy
    @event_filter = EventFilter.find(params[:id])

    if @event_filter.destroy
      flash[:notice] = "#{I18n.t('controllers.event')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} #{I18n.t('controllers.success_deleted')}"
    else
      flash[:error] = "#{I18n.t('controllers.error_while_delete')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} event."
    end

    render 'event_filter'
  end

  def delete
    @id = params[:id]
    @event_filter = EventFilter.find(params[:id])

    if @event_filter.destroy
      flash[:notice] = "#{I18n.t('controllers.event')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} #{I18n.t('controllers.successfully_created')}"
      render 'delete'
    else
      flash[:error] = "#{I18n.t('controllers.error_while_delete')} #{@event_filter.filter_type == 'suppress' ? 'suppresion' : 'filtering'} event."
      render 'update_details'
    end
  end

  ####################################################################################################
  # Funtion to clean the "new lines" in the CSV File
  ####################################################################################################
  def clean_blank_space_file(file)
    File.write(file, File.readlines(file).reject { |s| s.strip.empty? }.join)
  end

  ####################################################################################################
  # Funtion to clean the "new lines" in the CSV File
  ####################################################################################################
  def clean(input)
    input.gsub(/^[\s]*$\n/, "").strip
  end  

  def check_header(file)
    begin
      headers = CSV.open(file.path, 'r:ISO-8859-1:UTF-8') { |csv| csv.first }
      event_filter_type, signature_sid, signature_gid, count, seconds, source, destination, tracked_type, comment = headers
      if event_filter_type == "Event_Filter_Type" && signature_sid == "Signature_Sid" && signature_gid == "Signature_Gid" && count == "Count" && seconds == "Seconds" && source == "Source" && destination == "Destination" && tracked_type == "Tracked_Type" && comment == "Comment"
        return true
      else 
        return false
      end 
    rescue StandardError => e
      return false
    end  
  end

  def import
    render layout: false
  end

  def import_file
    content_file = params[:policy][:csv_file]

    # Before deciding on adding or replacing variables first it checks if the header of the file is correct
    if check_header(content_file) 
      # Clean the CSV file. Eliminate white space (new lines)
      clean_blank_space_file(content_file.path) 

      if params[:add_rules] == 'true'
        rules_count = @policy.add_event_rules_from_file(
          content_file, current_user.id
        )
        msg = "#{rules_count[0]} #{I18n.t('controllers.event_rules_added')} " \
          "#{rules_count[1]} #{I18n.t('controllers.rules_updated')}"
      else
        rules_count = @policy.replace_event_rules_from_file(
          content_file, current_user.id
        )
        msg = "#{rules_count} #{I18n.t('controllers.rules_added')}"
      end
      redirect_to policy_event_filters_path(@policy), notice: msg  
    else
      error_msg = "#{I18n.t 'controllers.unknown_content_csv'}"
      redirect_to policy_event_filters_path(@policy), alert: error_msg
    end
  end

  private

  def set_policy
    @policy = Policy.find(params[:policy_id])
  end

  def set_rule
    @rule = Rule.find(params[:rule_id])
  end

  def set_policy_by_sensor
    real_sensor = Sensor.find(params[:sensor_id])

    @sensor =
      # Sensor IPS v3
      if real_sensor.is_sensor?
        group_sensor = Sensor.find_by(
          real_parent: real_sensor,
          group_id: params[:group_id].to_i,
          type: Sensor::IPS_GROUP
        )

        Sensor.find_by(
          real_parent: group_sensor,
          group_id: params[:group_id].to_i,
          binding_id: params[:binding_id].to_i,
          type: Sensor::IPS_BINDING
        )
      # Sensor IPS v2
      else
        real_sensor
      end

    @policy = @sensor.policy
  end

  def event_filter_params
    params.require(:event_filter).permit(
      :filter_type, :source_ip, :destination_ip, :count, :seconds,
      :rule_sid, :rule_gid, :tracked_type, :comment, :user_id
    )
  end

  def export_csv
    filename = in_time_machine { Time.now.strftime('%Y%m%d%H%M%S') + '-EventRules-' + @policy.clean_name + '.csv' }
    content = @policy.event_rules_to_csv_format
    send_data content, filename: filename
  end
end
