#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Event DataSource. It takes the data from IPS. The initialization
# traduce multiple parameters like ports, IP Address...
# to user friendly strings.
#
class Event < Redborder::DataSource
  # -------------------------------------------------------------
  # Globals definition
  # -------------------------------------------------------------
  COLUMNS = Redborder::Module.get :ips, :columns
  SESSION_ID = Redborder::Module.get(:ips, :session_id).to_sym
  AGGREGATIONS = Redborder::Module.get(:ips, :aggregations)

  COLUMNS.map do |k, v|
    v[:columns].keys + (v[:complementary_columns].try('keys') || [])
  end.flatten.each do |column|
    attr_reader column
  end

  # Set the columns names to filter
  COLUMN_NAMES = { start_time: 'Start Time',
                   end_time: 'End Time',
                   group_id: 'Group',
                   sensor_uuid: 'Sensor UUID',
                   group_uuid: 'Group UUID',
                   domain: 'Domain' }

  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  AGGREGATIONS.each do |aggregation, _aggregation_data|
    attr_reader aggregation
  end

  attr_reader :timestamp, :result

  # -------------------------------------------------------------
  # Base methods
  # -------------------------------------------------------------

  #
  # Create an Event object with the data returned from Druid.
  #
  # == Parameters:
  # args::
  #   Hash with the data to build an Event model.
  #
  # == Returns:
  # Event object
  #
  def initialize(user, args)
    if args.is_a?(Hash)
      args.each do |key, value|
        instance_variable_set(('@' + key).to_sym, value)
      end

      if args['timestamp'].present?
        @timestamp = args['timestamp'].to_time.in_time_zone
      end
    end

    @sensor_ip = get_sensor_ip(@sensor_name, [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC])

    @src_name = NetObject.fetch(@src, user.id, @src)
    @dst_name = NetObject.fetch(@dst, user.id, @dst)

    @src_port_name = ServiceObject.fetch(@src_port, nil, @src_port)
    @dst_port_name = ServiceObject.fetch(@dst_port, nil, @dst_port)
    @l4_proto_name = ProtocolObject.fetch(@l4_proto, nil, @l4_proto)

    @ethsrc_name = get_mac_vendor_full(@ethsrc)
    @ethdst_name = get_mac_vendor_full(@ethdst)

    @src_country = get_country_name(@src_country_code)
    @dst_country = get_country_name(@dst_country_code)

    # VlanObject's are general values
    @vlan_name = VlanObject.fetch(@vlan, user.id, @vlan)

    # Round the aggregations that must be rounded
    AGGREGATIONS.each do |aggregation, aggregation_data|
      current_value = instance_variable_get("@#{aggregation}")
      next if current_value.nil?

      if aggregation_data[:round]
        instance_variable_set("@#{aggregation}", current_value.round)
      else
        instance_variable_set("@#{aggregation}", current_value.round(3))
      end
    end
  end

  #
  # Return the Symbol identifies the module
  #
  def self.id
    :ips
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.controller
    :events
  end

  #
  # Return the a hash with default action to redirect
  #
  def self.redirect
    { controller: self.controller, action: 'tops' }
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.default_aggregation
    :events
  end

  #
  # Return default columns for RAW. This is used on API when fetch raw data
  # and user doesn't specify columns
  #
  def self.default_columns_raw
    [:sensor_name, :msg, :priority]
  end

  def self.default_columns_incidents
    [:sensor_name, :msg, :priority, :timestamp, :src, :dst, :sig_id]
  end

  def self.default_columns_mitre
    [:sig_id, :priority]
  end

  #
  # *Group_id is not present in druid, so we have to filter by sensor_name*
  #
  # Return all IPS sensors
  #
  # == Parameters:
  # name::
  #   Array or String of names. If its nil, return all flow sensors.
  #
  def self.sensors(name = nil)
    types = [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_GROUP, Sensor::IPS_V2, Sensor::IPS_GENERIC]
    return Sensor.where(type: types, is_deleted: false) if name.nil?
    # No name provided
    Sensor.where(type: types, name: name, is_deleted: false)
  end

  #
  # Check the sensors current user can read.
  #
  # == Parameters:
  # user::
  #   User to check the permissions
  #
  # == Returns:
  # A hash with the group_uuid key contains all readable sensor uuids.
  def self.readable_sensors(user)
    group_uuids =
      user.sensor
      .self_and_descendants
      .where(type: [Sensor::IPS_GROUP, Sensor::IPS_V2, Sensor::IPS_GENERIC], is_deleted: false)
      .map(&:uuid)

    { group_uuid: group_uuids }
  end

  #
  # This method transforms the domains filters on filters
  # based on the sensors
  #
  # == Parameters:
  # domains::
  #   An array of domains
  # sensor_ids::
  #   An array with the name of the sensors
  # user::
  #   Current user
  #
  # == Returns:
  # A hash with the domains filters
  #
  def self.domain_filters(domains, sensor_uuids, user)
    return {} if domains.nil? || domains.empty?

    filters = {}
    filters[:group_uuid] = sensor_uuids || {}

    domains.each do |value, filter_type|
      # Select all the sensors from the domains, based on the type of the
      # filter that the user chose
      selected_domains = domain_filtered_sensors(value, filter_type)

      # Only select sensors that are IPS Group or IPS V2 and that
      # the user can read
      sensors = selected_domains.collect do |d|
        d.all_children.where(type: [RedBorder::SensorIPS::IPS_GROUP, RedBorder::SensorIPS::IPS_V2])
      end.flatten

      sensors = sensors.select do |sensor|
        user.can?(:read, sensor)
      end

      # If there are no sensors on these domains, filter for none
      sensors = [Sensor.new(name: '0', uuid: '0')] if sensors.empty?

      # Apply the selected filter
      sensors.map(&:uuid).each do |s|
        filters[:group_uuid][s.to_s] = { operator: 'is' }
      end
    end

    filters
  end

  # =================================================
  # FILTERS
  # =================================================

  #
  # Add the columns, zones and domains to build filter form
  # It's overrides the filter_to_json from filtrable
  #
  # == Parameters:
  # user::
  #   User who get filters
  # json::
  #   If true, return filters as JSON. True by default.
  #
  # == Returns:
  # A Hash with the basic operators and the domains, columns, sensors and zones
  # to build the filter form
  #
  def self.filter_to_json(user = User.current_user, json = true)
    domains = user
              .sensor
              .self_and_descendants
              .where.not(id: 1)
              .where(type: Sensor::DOMAIN)

    f = super.merge!(
      ips_columns: COLUMN_NAMES.to_a.map do |c|
        { value: c[1],
          id: c[0],
          type: ([:start_time, :end_time].include?(c[0].to_sym) ? :datetime : :text_input) }
      end.compact.sort_by do |h|
        h[:value].to_s
      end,
      ips_sensors: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC])
          .map { |c| { value: c.name, id: c.name } }
      },
      groups: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .where(type: Sensor::IPS_GROUP)
          .map { |c| { value: c.complete_name, id: c.id } }
      },
      domains: {
        type: :dropdown,
        value: domains.map { |c| { value: c.path_name, id: c.id } }
      }
    )
    # Check format
    json ? f.to_json : f
  end
end
