#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module MonitorHelper
  def monitor_cast_num(count)
    cast_num count, ''
  end

  def call_hardwareinfo
    device();
  end

  def monitor_name(monitor)
    sensor_name = monitor.sensor_name
    group_name = monitor.group_name
    instance = monitor.instance
    name = sensor_name

    if instance.present?
      if group_name.present?
        name = (instance + ' (' + name + ' - ' + group_name + ')' )
      else
        name = (instance + ' (' + name + ')' )
      end
    else
      name = (name + ' - ' + group_name) if group_name.present?
    end

    name
  end

  def cast_monitor(count)
    if count > 1e12
      count = (count.to_f / 1e12).round(3)
      count = count.to_s + ' T'
    elsif count > 1e9
      count = (count.to_f / 1e9).round(3)
      count = count.to_s + ' G'
    elsif count > 1e6
      count = (count.to_f / 1e6).round(3)
      count = count.to_s + ' M'
    elsif count > 1e3
      count = (count.to_f / 1e3).round(3)
      count = count.to_s + ' K'
    else
      count = count.round(3).to_s
    end
    count
  end

  #
  # Return the namespace for a given segment
  #
  # == Parameters:
  # segment::
  #   Segment complete source with the data associated
  #
  # == Returns:
  # String with the namespace for the given segment
  #
  def segment_namespace_name(segment)
    uuid = segment.datasource.match(/^rb_\w+_(?<uuid>.+)/)[:uuid]
    current_user.data_sources.find{ |d| d.uuid.eql?(uuid) }.name
  rescue
    'Default'
  end

  #
  # Return the source for a given segment
  #
  # == Parameters:
  # segment::
  #   Segment complete source with the data associated
  #
  # == Returns:
  # String with the source for the given segment
  #
  def segment_source_name(segment)
    match = segment.datasource.match(/^rb_(?<topic>[^_]+)/)
    return 'N/A' unless match
    name = match[:topic]
    if name == 'ips'
      name.upcase
    elsif name == 'loc'
      'Location'
    else
      name.capitalize
    end
  end

  #
  # Return four rows with the identifier, datasource, namespace, type and
  # interval.
  #
  # == Parameters:
  # task::
  #   Task object with the data associated
  #
  # == Returns:
  # [HTML Safe] String
  #
  def task_columns(task)
    data = []
    if task.id.start_with? 'kill'
      type = 'kill'
      interval = task.payload['interval'] || '-'
    elsif task.id.start_with?('index_hadoop') && task.payload['spec'] &&
          task.payload['spec']['dataSchema'] &&
          task.payload['spec']['dataSchema']['granularitySpec'] &&
          task.payload['spec']['dataSchema']['granularitySpec']['intervals']
      type = 'hadoop'
      interval = task.payload['spec']['dataSchema']['granularitySpec']['intervals'].first || '-'
    elsif task.payload['spec'] && task.payload['spec']['ioConfig']
      type = task.payload['spec']['ioConfig']['type']
      interval =
        task.payload['spec']['ioConfig']['firehose'].try(:[], 'interval') || '-'
    else
      type = '-'
      interval = '-'
    end
    # Get the partition and replica
    partition, replica = partition_and_replica(task.id)

    # Get the datasource
    datasource = task.payload['dataSource']
    m = /\Arb_(?<topic>([a-z])+(_[a-z]+){0,1})(?<namespace>_[a-zA-Z0-9-]*)*\z/
        .match(datasource)
    namespace = m[:namespace] ? m[:namespace][1..m[:namespace].size] : 'default'
    data += [['Topic', m[:topic]], ['Namespace', namespace], ['Type', type],
             ['Interval', gsub_interval(interval)], ['Partition', partition],
             ['Replica', replica]]
    # Join and return
    data.map { |name, el| content_tag(:td, el, data: { header: name }) }.join.html_safe
  end

  # Get the partition and replica from a task
  #
  # @param id [String] Identifier of the task
  # @return [Array] The first element is the partition and the second the
  #    replica
  #
  def partition_and_replica(id)
    return ['-', '-'] if id.start_with? 'kill'
    el = id.split('.000Z').last
    m = /\A_(?<partition>[0-9]+)_(?<replica>[0-9]+)\z/.match(el)
    m ? [m[:partition], m[:replica]] : ['-', '-']
  end

  # Remove timezone from interval string
  #
  # @param interval [String] interval to remove timezone
  # @return [String] interval without timezone
  #
  def gsub_interval(interval)
    return if interval == '-'
    interval = interval.gsub(':00.000Z', '')
    interval = interval.gsub('.000Z', '')
    interval = interval.gsub('T', ' ')
    interval.gsub('/', ' / ')
  end

  #
  # Return the status string for resume table based on errors variable
  #
  def check_cluster_status(errors)
    return content_tag(:span, 'OK', class: 'bold green') if errors.empty?
    warning = errors.select { |_, v| v == :warning }.size
    error = errors.select { |_, v| v == :error }.size
    content = []
    if warning > 0
      content <<
        link_to("#{I18n.t('malware_helper.warning')} (#{warning})", '#',
                class: 'bold yellow cluster-filter button',
                data: { name: 'warning', type: 'all' })
    end
    if error > 0
      content <<
        link_to("#{I18n.t('malware_helper.alert')} (#{error})",
                class: 'bold red cluster-filter button',
                data: { name: 'error', type: 'all' })
    end
    # return content
    content.join(' ').html_safe
  end

  def tops_values(val)
    val.is_a?(Numeric) ? cast_monitor(val) : 0.0
  end

  def module_controller_name_monitor
    'monitor'
  end

  def get_oid_from_operation(operation)

     if operation.kind_of?(Array)
       oper = ActiveSupport::JSON.decode(operation.first.to_json)
     else
       oper = ActiveSupport::JSON.decode(operation.to_json)
     end


     return ""

      # if oper.key? "system"
      #   return oper["system"]
      # else
      #   return ""
      # end


  end
end
