#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
###############################################################################

#
# SensorMonitor DataSource. Get the data from the status of cluster nodes. The
# parameters are CPU, Load, Physical Memory...
#
class SensorMonitor < Redborder::DataSource
  COLUMNS = Redborder::Module.get :monitor, :columns
  SESSION_ID = Redborder::Module.get(:monitor, :session_id).to_sym
  TABS = Redborder::Module.get :monitor, :tabs
  AGGREGATIONS = Redborder::Module.get :monitor, :aggregations

  COLUMNS.keys.each do |column|
    attr_reader column
  end

  attr_reader :events, :sum_value, :min, :max, :timestamp, :instance

  COLUMN_NAMES = { start_time: 'Start Time',
                   end_time: 'End Time',
                   group_id: 'Group ID',
                   domain: 'Domain',
                   organization_uuid: 'Organization UUID'
                 }

  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  # Aggregations
  AGGREGATIONS.each do |aggregation, _aggregation_data|
    attr_reader aggregation
  end

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def initialize(_, args)
    return unless args.is_a?(Hash)
    args.each do |key, value|
      self.instance_variable_set(('@' + key).to_sym, value)
    end

    @timestamp = args['timestamp'].to_time.in_time_zone if args['timestamp']
                                                           .present?
  end

  #
  # Override the default conditions.
  #
  # == Returns
  # A Hash with the default conditions.
  #
  def self.default_conditions
    { granularity: :all,
      sort: :timestamp,
      direction: :DESCENDING,
      range: :last_1_hour,
      aggregation: default_aggregation,
      group_by: Redborder::Module.get(id, :columns)
        .select { |_, value| value[:enabled] }.keys,
      remove_zeros: false,
      controller: controller }
  end

  #
  # Return the Symbol identifies the module
  #
  def self.id
    :monitor
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.controller
    :monitor
  end

  #
  # Return the a hash with default action to redirect
  #
  def self.redirect
    { controller: self.controller, action: 'index' }
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.default_aggregation
    :monitors
  end

  #
  # Return all monitored sensors
  #
  # == Parameters:
  # name::
  #   Array or String of names. If its nil, return all flow sensors.
  #
  def self.sensors(name = nil)
    return Sensor.where(type: monitor_sensor_types, is_deleted: false) if
      name.nil?
    Sensor.where(type: monitor_sensor_types, name: name, is_deleted: false)
  end

  # =================================================
  # FILTERS
  # =================================================

  #
  # Add the columns, zones and domains to build filter form
  # It's overrides the filter_to_json from filtrable
  #
  # == Parameters:
  # user::
  #   User who get filters
  # json::
  #   If true, return filters as JSON. True by default.
  #
  # == Returns:
  # A Hash with the basic operators and the domains, columns, sensors and zones
  # to build the filter form
  #
  def self.filter_to_json(user = User.current_user, json = true)
    domains = user
              .sensor
              .self_and_descendants
              .where.not(id: 1)
              .where(type: Sensor::DOMAIN)

    f = super.merge!(
      monitor_columns: COLUMN_NAMES.to_a.map do |c|
        { value: c[1],
          id: c[0],
          type: ([:start_time, :end_time].include?(c[0].to_sym) ? :datetime : :text_input) }
      end.compact.sort_by do |h|
        h[:value].to_s
      end,
      monitor_sensors: {
        type: :dropdown,
        value: get_managers + user
                              .sensor
                              .self_and_descendants
                              .where.not(id: 1)
                              .where(type: monitor_sensor_types)
                              .map { |c| { value: c.name, id: c.name } }
      },
      groups: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .where(type: group_sensor_types)
          .map { |c| { value: c.complete_name, id: c.id } }
      },
      domains: {
        type: :dropdown,
        value: domains.map { |c| { value: c.path_name, id: c.id } }
      }
    )
    # Check format
    json ? f.to_json : f
  end

  def self.monitor_sensor_types(type = [])
    registered_modules = Redborder::Module.registered_modules.keys

    type += [
      Sensor::IPS, Sensor::IPS_V2, Sensor::IPS_PROXY
    ] if registered_modules.include?(:ips)

    type += [
      Sensor::FLOW, Sensor::MSE, Sensor::ALE,
      Sensor::DEVICE, Sensor::MERAKI, Sensor::CLIENT_PROXY
    ] if registered_modules.include?(:flow)

    type
  end

  def self.group_sensor_types(type = [])
    if Redborder::Module.registered_modules.keys.include?(:ips)
      type << RedBorder::SensorIPS::IPS_GROUP
    end
    type
  end

  # Get all managers of the platform
  #
  # @return [Array] Array of Hashes with managers
  #
  def self.get_managers
    return [] unless User.current_user.super_user?
    managers = YAML.load(File.read('config/managers.yml'))
    managers.map { |m| { value: m, id: m } }
  rescue
    Manager.all.map { |m| { value: m.name, id: m.name } }
  end

  def self.sensor_types(tab)
    modules = Redborder::Module.registered_modules.keys
    ips_tab = TABS[tab] && TABS[tab][:instance]
    type = []

    if modules.include?(:ips) && ips_tab
      type << RedBorder::SensorIPS::IPS_GROUP
    else
      type << RedBorder::SensorIPS::IPS if modules.include? :ips
      type << RedBorder::SensorIPS::IPS_V2 if modules.include? :ips
      type << RedBorder::SensorFlow::FLOW if modules.include? :flow
    end

    type
  end

  def self.sensor_fields(tab)
    modules = Redborder::Module.registered_modules.keys
    ips_tab = TABS[tab] && TABS[tab][:instance]
    fields = []

    if modules.include?(:ips) && ips_tab
      fields << :group_id
    else
      fields << :group_id
      fields << :sensor_name
    end

    fields
  end

  def self.filter_param(sensor_types)
    modules = Redborder::Module.registered_modules.keys

    if modules.include?(:ips) &&
       sensor_types.eql?([RedBorder::SensorIPS::IPS_GROUP])
      :group_id
    else
      :sensor_name
    end
  end

  def self.to_csv(content, percentage)
    columns_keys   = [:sensor_name, :min, :monitors, :max]
    columns_values = ['Sensor', 'Min', 'Average', 'Max']

    csv = columns_values.join("|") + "\n"

    content.each do |line|
      aux_array = []
      columns_keys.each do |column|
        val = line.send(column.to_s).to_s
        val += "%" if percentage && [:min, :monitors, :max].include?(column.to_sym)
        aux_array << val
      end
      csv += aux_array.join("|") + "\n"
    end

    csv
  end

  def self.consumption_data(parameters = {})
    # Check if the module is active
    return [] unless Redborder::Module.registered_modules.include? :monitor

    sensor = parameters[:sensor] || Sensor.root

    organization_uuid = sensor.is_root? ? '*' : sensor.uuid

    filter = Redborder::Filter.new(SensorMonitor, {})
    filter.add(:organization_uuid, organization_uuid)

    start_time = parameters[:start_time] || Time.now.utc.beginning_of_month
    end_time = parameters[:end_time] || Time.now.utc.beginning_of_month.next_month

    where(
      query_type: :timeseries,
      namespaces: [Sensor.root.uuid],
      range: :custom,
      start_time: start_time,
      end_time: end_time,
      granularity: :day,
      ignore_user_namespaces: true,
      ignore_user_filters: true,
      monitor_tab: :organization_received_bytes,
      search: filter.get,
      skip_empty: false,
      timezone: 'UTC'
    ).map { |e| { event: { events: e.sum_value }, timestamp: e.timestamp.utc.to_s } }
  end
end
