#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Wireless DataSource. It takes the data from wireless traffic.
#
class Wireless < Redborder::DataSource
  COLUMNS = Redborder::Module.get(:wireless, :columns)
  SESSION_ID = Redborder::Module.get(:wireless, :session_id).to_sym
  AGGREGATIONS = Redborder::Module.get(:wireless, :aggregations)

  # Generate a reader attribute for the columns
  COLUMNS.map{ |k, v| v[:columns].keys + (v[:complementary_columns].try('keys') || []) }.flatten.each do |column|
    attr_reader column
  end

  # Define the column names to filter
  COLUMN_NAMES = { start_time: "#{I18n.t 'wireless2.start_time'}",
                   end_time: "#{I18n.t 'wireless2.end_time'}",
                   domain: "#{I18n.t 'wireless2.domain'}"
                 }
  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  # Aggregations
  AGGREGATIONS.each do |aggregation, _aggregation_data|
    attr_reader aggregation
  end

  attr_reader :timestamp, :result

  #
  # Create a Wireless object with the data returned from Druid.
  #
  # == Parameters:
  # args::
  #   Hash with the data to build a Wireless model.
  #
  # == Returns:
  # Wireless object
  #
  def initialize(_user, args = {})
    if args.is_a?(Hash)
      args.each do |key, value|
        instance_variable_set("@#{key}", value)
      end

      @timestamp = args['timestamp'].to_time.in_time_zone if args['timestamp'].present?

      @client_rssi_num = get_client_rssi_num(@client_rssi)
      @client_mac_name_two = get_mac_vendor_full(@client_mac)
      @client_mac_name  = MacObject.fetch(@client_mac, _user.id, @client_mac) 
      @wireless_station_name = get_ap_name(@wireless_station)
    end

    # Round the aggregations that must be rounded
    AGGREGATIONS.each do |aggregation, aggregation_data|
      current_value = instance_variable_get("@#{aggregation}")
      next if current_value.nil?
      if aggregation_data[:round]
        instance_variable_set("@#{aggregation}", current_value.round)
      else
        instance_variable_set("@#{aggregation}", current_value.round(3))
      end
    end
  end

  #
  # Return the Symbol of id module
  #
  def self.id
    :wireless
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.controller
    :wireless
  end

  #
  # Return the a hash with default action to redirect
  #
  def self.redirect
    { controller: controller, action: 'tops' }
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.default_aggregation
    :events
  end

  #
  # Return default columns for RAW
  #
  def self.default_columns_raw
    [:pri, :hostname]
  end

  #
  # Return all wireless sensors
  #
  # == Parameters:
  # name::
  #   Array or String of names. If its nil, return all flow sensors.
  #
  def self.sensors(name = nil)
    return Sensor.where.not(id: 1).where(is_deleted: false) if name.nil?
    # No name provided
    Sensor.where.not(id: 1).where(name: name, is_deleted: false)
  end

  # =================================================
  # FILTERS
  # =================================================

  #
  # Add the columns and domains to build filter form
  # It's overrides the filter_to_json from filtrable
  #
  # == Parameters:
  # user::
  #   User who get filters
  # json::
  #   If true, return filters as JSON. True by default.
  #
  # == Returns:
  # A Hash with the basic operators and the domains and columns
  # to build the filter form
  #
  def self.filter_to_json(user = User.current_user, json = true)
    domains = user
              .sensor
              .self_and_descendants
              .where.not(id: 1)
              .where(type: Sensor::DOMAIN)

    f = super.merge!(
      wireless_columns: COLUMN_NAMES.to_a.map do |c|
        { value: c[1],
          id: c[0],
          type: ([:start_time, :end_time].include?(c[0].to_sym) ? :datetime : :text_input) }
      end.compact.sort_by do |h|
        h[:value].to_s
      end,
      wireless_sensors: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .map { |c| { value: c.name, id: c.name } }
      },
      domains: {
        type: :dropdown,
        value: domains.map { |c| { value: c.path_name, id: c.id } }
      }
    )
    # Check format
    json ? f.to_json : f
  end

  #
  # Return the filters with zones
  #
  # == Parameters:
  # search::
  #   A hash with the current filters to apply
  # user::
  #   Current user
  #
  # == Returns:
  # A hash with the filters combination of current, user, domain and zones.
  #
  def self.user_filters(search, user)
    return {} if user.nil? || search.nil?
    filters = {}
    # Filters from user
    # TODO: Don't except zones
    # filters.merge! search.except(:domain, :zone)
    filters.merge! search.except(:domain)
    # Filters from domains
    filters.merge! domain_filters(search[:domain], search[:sensor_name], user)
    # Filters from zones
    filters
  end

  private

  def get_client_rssi_num(client_rssi=nil)
    Rails.cache.fetch "rb-webui:client_rssi_num:#{client_rssi}" do
      client_rssi_num = YAML.load(File.read("#{Rails.root}/lib/modules/wireless/config/rssi_values.yml")).symbolize_keys
      client_rssi_num[client_rssi.to_sym]['value']
    end
  rescue
    client_rssi
  end

  def get_ap_name(mac_address)
    Rails.cache.fetch "rb-webui:accesspoint:#{mac_address}", expires_in: 1.day do
      if mac_address.present?
        AccessPoint.find_by(mac_address: mac_address).try(:name) || mac_address
      else
        'N/A'
      end
    end
  end

end
