#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'syslog'
require 'json'

module Redborder
  # This class notificate an alarm to the users using several methods. Now,
  # available methods are:
  #   - Syslog
  #   - GCM push
  #   - Email
  #
  class AlarmNotificator
    attr_reader :alarm, :result, :type

    # Receive the alarm to notify and the result of query
    #
    # @param alarm [Alarm] Alarm to notify
    # @param result [Hash] Result of the query based on alarm conditions
    #
    def initialize(alarm, result)
      @alarm = alarm
      @result = result
      # We can define an alarm with an upper, a bottom or both limits. In the case we
      # define two limits, the alarm will be fired only by one of those limits.
      # That the reason I set here the "type" of the alarm.
      @type = @result[:where].to_s
    end

    # Send a notification based on the options from alarm
    #
    def notify
      notify_email if @alarm.notify_to_email
      notify_syslog if @alarm.notify_to_syslog
      notify_slack if @alarm.notify_to_slack
      notify_vault if @alarm.notify_to_vault
    end

    # Send a notificaiton using the email
    #
    def notify_email
      AlarmMailer.alarm(alarm, result).deliver_now
    end

    # Send a notificaiton using syslog
    #
    def notify_syslog
      ::Syslog.open('redBorder') do |s|
        s.alert(
          "#{I18n.t 'lib_redborder.alarm'} %s #{I18n.t 'lib_redborder.fired'} #{@type} #{I18n.t 'lib_redborder.limit'} %s %s #{I18n.t 'lib_redborder.with'} %s %s #{I18n.t 'lib_redborder.in'} %s <-> %s",
          alarm.name, limit.to_s, units, value, units, result[:start_time].to_s,
          result[:end_time].to_s
        )
      end
    end

    def notify_slack
      require "uri"
      require "json"
      require "openssl"
      require "net/http"
      if @alarm
        begin
          uri = URI(@alarm.slack_webhook)
          http = Net::HTTP.new(uri.host, uri.port)
          http.use_ssl = true
          http.verify_mode = OpenSSL::SSL::VERIFY_PEER
          request = Net::HTTP::Post.new(uri.request_uri)
          request.add_field('Content-Type', 'application/json')
          request.body = { text:"[#{Setting.company}] Alarm #{@alarm.name} #{@alarm.product_type} Severity #{@alarm.severity.name} fired! Domain Alarm Triggered: #{Setting.webpage_url} " }.to_json
          response = http.request(request)

        rescue Exception => e
          raise "#{Time.now} [ERROR] Error on Slack request: #{e}"
        end
      end
    end

    def notify_vault
      sensors_info = get_sensors_info(alarm.condition)
      enrichment_hash = sensors_info.each_with_object({}) do |sensor, hash|
        if sensor[:enrichment]
          hash.merge!(sensor[:enrichment])
        end
      end
      begin
        enrichment_hash["alert_msg"] = "Alarm #{alarm.name} fired #{@type} limit #{limit.to_s} #{units} with #{value} #{units} in #{result[:start_time].to_s} <-> #{result[:end_time].to_s}"
        formatted_string = enrichment_hash.map { |key, value| "#{key}=\"#{value}\"" }.join(" ").to_s

        # Escape % characters
        formatted_string.gsub!('%', '%%')

        ::Syslog.open('alarmsjob') do |s|
          s.alert(formatted_string)
        end
      rescue => e
        Rails.logger.error "Error while sending Syslog alert: #{e.message}"
      end
    rescue => e
      Rails.logger.error "Error in notify_vault: #{e.message} - Backtrace: #{e.backtrace.join("\n")}"
    end

    def get_sensors_info(condition)
      sensors = []
      condition = JSON.parse(condition, symbolize_names: true)

      user_sensors = @alarm.user.sensors_with_permission - @alarm.user.domains_with_permission

      condition.each do |key, value|
        column = value[:column]
        operator = value[:operator]
        regex = value[:value]

        user_sensors.each do |sensor|
          matches_condition = case operator
                                when 'contains'
                                  sensor.name.match?(/#{regex}/i)
                                when 'is'
                                  sensor.name == regex
                                when 'is_not'
                                  sensor.name != regex
                                else
                                  false
                                end
          if matches_condition
            sensor_enrichment = enrich(sensor.chef_node)

            begin 
              sensor_enrichment["sensor_ip"] = sensor.ip if sensor.ip
            rescue
              Rails.logger.error "Cannot access ip of sensor with id #{sensor.id}"
            end

            enriched_sensor = {
              sensor_name: sensor.name,
              enrichment: sensor_enrichment
            }

            begin 
              sensor_enrichment["sensor_ip"] = sensor.ip if sensor.ip
            rescue
              Rails.logger.error "Cannot access ip of sensor with id #{sensor.id}"
            end

            sensors << enriched_sensor
          end
        end
      end
      sensors
    end

    private

    # Get the value of the result
    #
    def value
      @value ||= (result[:value] / (10.0**alarm.send("multiplier_#{@type}"))).round(2)
    end

    # Return the alarm unit
    #
    def units
      @units ||= alarm.units("multiplier_#{@type}".to_sym)
    end

    # Return the limit of the alarm
    #
    def limit
      @limit ||= alarm.send("#{@type}_limit")
    end

    # Return the severity color of the alarm
    #
    def severity_colour
      @alarm.severity.try(:bg_color) || '#777777'
    end

    # Build the message of the alarm
    #
    def mobile_message
      # Base message
      message = {
        name: @alarm.name,
        start_time: @result[:start_time].to_i,
        end_time: @result[:end_time].to_i,
        type: @type,
        value: value,
        server_key: Setting.server_key,
        severity_colour: severity_colour
      }

      # Merge limits
      message.merge(
        "#{@type}_limit".to_sym => limit,
        "#{@type}_units".to_sym => units
      )
    end

    # Enrich from sensor node
    def enrich(resource_node)
      enrichment = {}

      if resource_node['rbname']
        enrichment['sensor_name'] = resource_node['rbname']
      end

      redborder = resource_node['redborder']

      return enrichment unless redborder

      if redborder['sensor_uuid']
        enrichment['sensor_uuid'] = redborder['sensor_uuid']
      end

      if redborder['service_provider']
        enrichment['service_provider'] = redborder['service_provider']
      end

      if redborder['service_provider_uuid']
        enrichment['service_provider_uuid'] = redborder['service_provider_uuid']
      end

      if redborder['namespace']
        enrichment['namespace'] = redborder['namespace']
      end

      if redborder['namespace_uuid']
        enrichment['namespace_uuid'] = redborder['namespace_uuid']
      end

      if redborder['organization']
        enrichment['organization'] = redborder['organization']
      end

      if redborder['organization_uuid']
        enrichment['organization_uuid'] = redborder['organization_uuid']
      end

      if redborder['building']
        enrichment['building_uuid'] = redborder['building_uuid']
      end

      enrichment
    end
  end
end
