#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Audit
    extend ActiveSupport::Concern
    require 'syslog'
    included do
      after_save :customize_comment
      after_destroy :customize_comment
    end

    def customize_comment
      # some objects has 2 audits associated, one for creating and another for updating. TODO: find out why
      if self.audits.length > 1
        #if it has nil in audited changes, it means we are creating it, since when you update this is not nil
        if self.audits.last.try(:audited_changes) && self.audits.last.audited_changes.first &&
          self.audits.last.audited_changes.first[1] && self.audits.last.audited_changes.first[1].kind_of?(Array) && self.audits.last.audited_changes.first[1][0] == nil
          audit = self.audits.first if self.try(:audits)
        else
          audit = self.audits.last if self.try(:audits)
        end
      else
        # select last audit for self object. Audits are ordered by version.
        audit = self.audits.last if self.try(:audits)
      end
      return unless audit

      comment = self.send("audit_comment_on_#{audit.action.to_sym}", audit)
      username = audit.user_id ? User.unscoped.find(audit.user_id).name : 'Administrator'

      send_to_syslog = (comment != audit.comment or username != audit.username) ? true : false

      if audit.update!(comment: comment, username: username)
        notify_audit_to_syslog(audit) if send_to_syslog
      end

    end

    def audit_comment_on_create(audit)
      "#{I18n.t 'lib_redborder.created'} #{audit.auditable_type} #{audit.audited_changes['name']}"
    end

    def audit_comment_on_update(audit)
      object = audit.auditable_type.constantize.where(id: audit.auditable_id)
      name = object.first.try(:name)

      comment = "#{I18n.t 'lib_redborder.updated'} #{name}:"
      comment += audit.audited_changes.collect do |key, values|
        " #{key} #{I18n.t 'lib_redborder.from'} '#{values[0]}' #{I18n.t 'lib_redborder.to'} '#{values[1]}'"
      end.join(', ')
      comment
    end

    def audit_comment_on_destroy(audit)
      "#{I18n.t 'lib_redborder.deleted2'} #{audit.auditable_type} #{audit.audited_changes['name']}"
    end

    def register_license_changes(model, local_file, remote_file)
      comment = "#{I18n.t 'lib_redborder.file'} #{local_file} #{I18n.t 'lib_redborder.remote'} #{model.class.name} #{model.name}, #{remote_file}"
      audit = if model.class.name.eql?("Manager")
        Audited.audit_class.create(auditable_type: model.class.name, action: "update", comment: comment)
      else
        model.audits.create!(auditable_id: model.id, auditable_type: model.class.name, action: "update", comment: comment)
      end
      audit.user = User.current_user
      notify_audit_to_syslog(audit)
    end

    def register_rsa_changes(model)
      comment = " Recreated RSA: #{User.current_user.name} recreated RSA key"
      audit = if model.class.name.eql?("Manager")
        Audited.audit_class.create(auditable_type: model.class.name, action: "update", comment: comment)
      else
        model.audits.create!(auditable_id: model.id, auditable_type: model.class.name, action: "update", comment: comment)
      end
      audit.user = User.current_user
      notify_audit_to_syslog(audit)
    end

    def register_setting_changes(model, message)
      comment = "#{message} by User #{User.current_user.name} with id #{User.current_user.id}"
      audit = if model.class.name.eql?("Manager")
        Audited.audit_class.create(auditable_type: model.class.name, action: "update", comment: comment)
      else
        model.audits.create!(auditable_id: model.id, auditable_type: model.class.name, action: "update", comment: comment)
      end
      audit.user = User.current_user
      notify_audit_to_syslog(audit)
    end

    def register_alarm_changes(model, message)
      user = User.first # Because its call from the job
      comment = "#{message} by User #{user.name} with id #{user.id}"
      audit = if model.class.name.eql?("Manager")
        Audited.audit_class.create(auditable_type: model.class.name, action: "update", comment: comment)
      else
        model.audits.create!(auditable_id: model.id, auditable_type: model.class.name, action: "update", comment: comment)
      end
      audit.user = user
      notify_audit_to_syslog(audit)
    end

    def register_fail_login_changes(model,message)
      user = User.first # Because its call from the job
      comment = message
      audit = if model.class.name.eql?("Manager")
        Audited.audit_class.create(auditable_type: model.class.name, action: "update", comment: comment)
      else
        model.audits.create!(auditable_id: model.id, auditable_type: model.class.name, action: "update", comment: comment)
      end
      audit.user = user
      notify_audit_to_syslog(audit)
    end

    def notify_audit_to_syslog(audit)
      user_name = (audit.user.kind_of?(String)) ? audit.user : audit.user.name
      ::Syslog.open('redBorder') do |s|
        s.alert(audit.comment + " - Username #{user_name}")
      end
    end

  end
end
