require 'rest-client'
require 'json'
require 'active_record'
require 'logger'
require 'yaml'
require 'pg'

class AppLogger
  def self.logger
    @logger ||= Logger.new(STDOUT).tap do |log|
      log.level = Logger::DEBUG
    end
  end
end

class DatabaseConfig
  def self.setup
    config = YAML.load_file("/var/www/rb-rails/config/database.yml")
    ActiveRecord::Base.establish_connection(config['production'])
  rescue StandardError => e
    AppLogger.logger.error("Database configuration error: #{e.message}")
    raise
  end

  def self.verify_connection
    ActiveRecord::Base.connection
    AppLogger.logger.info("Database connection established successfully.")
  rescue ActiveRecord::ConnectionNotEstablished => e
    AppLogger.logger.error("Failed to connect to the database: #{e.message}")
    raise
  end
end

class GLPIClient
  attr_reader :api_url, :logger

  def initialize(api_url, app_token)
    @api_url = api_url
    @app_token = app_token
    @logger = AppLogger.logger
  end

  def initiate_session(user_token)
    response = post_request("#{@api_url}/initSession", { user_token: user_token })
    JSON.parse(response.body)['session_token']
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to start session", e)
    raise
  end

  def end_session(session_token)
    post_request("#{@api_url}/killSession", {}, session_token)
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to end session", e)
    raise
  end

  def get_device_details(device_id, device_type, session_token)
    url = "#{@api_url}/#{device_type}/#{device_id}?expand_dropdowns=true&get_hateoas=true&with_devices=true&with_disks=true&with_softwares=true&with_connections=true&with_networkports=true&with_infocoms=true&with_contracts=true&with_documents=true&with_tickets=true&with_problems=true&with_changes=true&with_notes=true&with_logs=true"
    response = get_request(url, session_token)
    device_details = JSON.parse(response.body)

    mac = get_network_port_mac(device_details, session_token)
    operating_system = get_operating_system(device_id, session_token)
    ip_addresses = search_ip_addresses(device_id, session_token)

    { mac: mac, os: operating_system, ips: ip_addresses, details: device_details }
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to retrieve details for device #{device_id}", e)
    raise
  end

  private

  def headers(session_token = nil)
    {
      content_type: :json,
      accept: :json,
      'App-Token' => @app_token,
      'Session-Token' => session_token
    }
  end

  def post_request(url, payload, session_token = nil)
    RestClient.post(url, payload.to_json, headers(session_token))
  end

  def get_request(url, session_token)
    RestClient.get(url, headers(session_token))
  end

  def get_network_port_mac(device_details, session_token)
    network_ports_link = if device_details["links"]
      link = device_details["links"].find { |link| link["rel"] == "NetworkPort" }
      link ? link["href"] : nil
    end

    return nil unless network_ports_link

    response = get_request(network_ports_link, session_token)
    network_ports_details = JSON.parse(response.body)

    port = network_ports_details.find { |p| p.is_a?(Hash) && p.key?('mac') }
    mac = port ? port['mac'] : nil
    mac
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to retrieve network port details", e)
    raise
  end

  def get_operating_system(device_id, session_token)
    url = "#{@api_url}/Computer/#{device_id}/OperatingSystem"
    response = get_request(url, session_token)
    os_details = JSON.parse(response.body)

    os_details.first['name'] if os_details.is_a?(Array) && !os_details.empty?
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to retrieve operating system details for device #{device_id}", e)
    raise
  end

  def search_ip_addresses(device_id, session_token)
    search_url = "#{@api_url}/search/Computer"
    query_params = {
      "criteria[0][field]" => "2",  # Field ID for the device ID
      "criteria[0][searchtype]" => "equals",
      "criteria[0][value]" => device_id,
      "forcedisplay[0]" => "126",   # Field ID for IP address
      "itemtype" => "Computer",
      "start" => "0",
      "limit" => "50"
    }
    response = RestClient.get(search_url, headers(session_token).merge(params: query_params))
    results = JSON.parse(response.body)
    logger.debug("Search IP addresses raw response: #{results}")

    data = results['data']
    if data
      ip_addresses = data.flat_map { |device_data| device_data["126"] if device_data.is_a?(Hash) }.compact.uniq
    else
      ip_addresses = []
    end
    ip_addresses
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to search IP addresses", e)
    raise
  end

  def log_error(message, exception)
    logger.error("#{message}: #{exception.response}")
  end
end

class DeviceProcessor
  attr_reader :total_devices_retrieved, :total_devices_updated, :total_devices_inserted

  def initialize(client, user_id)
    @client = client
    @user_id = user_id
    @logger = AppLogger.logger
    @total_devices_retrieved = 0
    @total_devices_updated = 0
    @total_devices_inserted = 0
  end

  def retrieve_device_details(session_token, device_type)
    start = 0
    batch_size = 50

    loop do
      range = "#{start}-#{start + batch_size - 1}"
      url = "#{@client.api_url}/#{device_type}?range=#{range}"
      response = @client.send(:get_request, url, session_token)
      devices = JSON.parse(response.body)
      @total_devices_retrieved += devices.size
      logger.debug("Number of devices returned from API: #{devices.size}")

      break if devices.empty?

      object_type_record = find_or_create_object_type(device_type)

      devices.each do |device|
        device_id = device["id"]
        logger.info("Processing device ID: #{device_id}, Name: #{device['name']}")
        previously_existed = process_device(device, device_type, session_token, object_type_record)

        previously_existed ? @total_devices_updated += 1 : @total_devices_inserted += 1
      end

      logger.info("Batch complete: Retrieved #{@total_devices_retrieved}, Updated #{@total_devices_updated}, Inserted #{@total_devices_inserted}")

      content_range = response.headers[:content_range]
      if content_range
        total_devices = content_range.split('/').last.to_i
        start += batch_size
        break if start >= total_devices
      else
        logger.warn("No Content-Range header found in response")
        break
      end

      sleep(10)
    end
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to retrieve #{device_type} data", e)
    raise
  end

  private

  attr_reader :logger

  def find_or_create_object_type(object_type)
    ObjectType.find_or_create_by(name: object_type) do |type|
      type.comment = "Type for #{object_type}"
    end
  rescue ActiveRecord::RecordInvalid => e
    logger.error("Failed to find or create ObjectType: #{e.message}")
    raise
  end

  def find_or_initialize_mac_object(name, value, object_type)
    MacObject.find_or_initialize_by(name: name, value: value).tap do |obj|
      obj.object_type = object_type
      obj.user_id = @user_id
    end
  end

  def process_device(device, device_type, session_token, object_type_record)
    ActiveRecord::Base.transaction do
      begin
        device_details = @client.get_device_details(device["id"], device_type, session_token)
        mac = device_details[:mac]
        os = device_details[:os]
        ips = device_details[:ips] || []
        comment = device["comment"]

        mac_object = find_or_initialize_mac_object(device["name"], mac, object_type_record.name)
        mac_object.comment = comment
        mac_object.type = "MacObject"
        mac_object.object_type = object_type_record.name
        mac_object.operating_system = os
        mac_object.ip_addresses = ips

        mac_object.save!
        logger.info("Added or updated MacObject for device ID: #{device['id']}")
        return !mac_object.new_record?
      rescue ActiveRecord::RecordInvalid => e
        logger.error("Failed to save MacObject: #{e.record.errors.full_messages.join(", ")}")
        return false
      end
    end
  rescue => e
    logger.error("Error processing device ID #{device['id']}: #{e.message}")
    logger.error(e.backtrace.join("\n"))
    raise
  end
end

class GlpiInteraction
  def initialize(user_token, api_url, app_token, user_id)
    @user_token = user_token
    @api_url = api_url
    @app_token = app_token
    @user_id = user_id
    @client = GLPIClient.new(api_url, app_token)
    @device_types = ["Computer", "NetworkEquipment", "Phone", "Peripheral"]
  end

  def execute
    begin
      result = "Starting GLPI interaction.\n"
      AppLogger.logger.info("Starting GLPI interaction.")
      DatabaseConfig.setup
      DatabaseConfig.verify_connection
  
      session_token = @client.initiate_session(@user_token)
      if session_token.nil?
        AppLogger.logger.error("Failed to initiate session. Exiting...")
        raise "Session initiation failed"
      end
  
      @device_types.each do |type|
        device_processor = DeviceProcessor.new(@client, @user_id)
        device_processor.retrieve_device_details(session_token, type)
        result += "Batch complete for #{type}: Retrieved #{device_processor.total_devices_retrieved}, Updated #{device_processor.total_devices_updated}, Inserted #{device_processor.total_devices_inserted}\n"
      end
  
      result += "GLPI interaction completed successfully.\n"
    ensure
      @client.end_session(session_token) if session_token
      ActiveRecord::Base.connection.close
      AppLogger.logger.info("Script is closing all connections.")
    end
  
    result
  rescue => e
    AppLogger.logger.error("GLPI interaction failed: #{e.message}")
    AppLogger.logger.error(e.backtrace.join("\n"))
    raise
  end  
end
