#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    # AIGenerateContentJob is a Delayed Job that handles the generation of AI content and updates a specific field of a model with the generated content.
    # 
    # Usage:
    # - `function_name`: The name of the function to call on the Redborder::AI module to generate the content (e.g., 'generate_incident_title').
    # - `function_params`: Parameters to pass to the AI generation function (e.g., { alerts: ..., incident_name: ... }).
    # - `model_params`: Parameters for updating the model, including:
    #  - `model_name`: The name of the model class as a string (e.g., 'Incident', 'IncidentReports').
    #  - `name_field`: The name of the field to update with the generated content (e.g., 'name', 'description').
    #  - `record_id`: The ID of the record to update.
    class AIGenerateContentJob < Struct.new(:verbose, :user_id, :function_name, :function_params, :model_params)
      include Redborder::Jobs::JobHelper

      def perform
        content = ""
        
        begin
          # Call the appropriate function to generate the content
          response = Redborder::AI.send(function_name, function_params)
          content += "Executing #{function_name} method.\n\n" 

          # Dynamically update the specified field of the given model
          if response.is_a?(String)
            model_class = Object.const_get(model_params[:model_name])
            record = model_class.find(model_params[:record_id])

            if record.update(model_params[:name_field] => response)
              content += "#{model_params[:model_name]} #{record.id} updated with AI-generated #{model_params[:name_field]}: '#{response}'\n"

              # Create an AI generation record
              create_ai_generation_record(model_params) 
            else
              content += "Failed to update #{model_params[:model_name]} #{model_params[:record_id]} with AI-generated content.\n"
              content += "Validation errors: #{record.errors.full_messages.join(', ')}\n"
            end
          else
            content += "Failed to generate content for #{model_params[:model_name]} #{model_params[:record_id]}.\n"
            content += "Error: #{response}\n"
          end
        rescue StandardError => e
          content += "Failed to generate content.\nContent type: #{function_to_execute[:type]}\n"
          content += "Error: #{e.message}\n"
        ensure
          StoredDelayedJob.create(
            job: 'AIGenerateContentJob',
            content: content,
            user_id: user_id
          )
        end
      end

      private

      # Method to create or update a new entry in ai_generation_records table
      def create_ai_generation_record(model_params)
        ai_record = AiGenerationRecords.find_by(
          table_name: model_params[:model_name].tableize,
          field_name: model_params[:name_field],
          record_id: model_params[:record_id],
          record_type: model_params[:model_name]
        )
      
        # If record exists, update the updated_at field
        if ai_record
          ai_record.update(updated_at: Time.current)
        else
          AiGenerationRecords.create(
            table_name: model_params[:model_name].tableize,
            field_name: model_params[:name_field],
            record_id: model_params[:record_id],
            record_type: model_params[:model_name]
          )
        end
      rescue ActiveRecord::RecordInvalid => e
        Rails.logger.error "Failed to create or update AI generation record: #{e.message}"
      end
    end
  end
end
