#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder

  module Jobs

    class GlpiInventoryJob < Struct.new(:user_id, :verbose)
      include Redborder::Jobs::JobHelper
      attr_accessor :content

      def perform
        content = "Node executing this job: #{`hostname`.chomp!}\n\n"
        
        unless Integration.glpi_params_present?
          logit("GLPI parameters are missing. Job will not execute.") if verbose
          return
        end

        begin
          params_glpi = get_params_glpi
          raise "No GLPI settings found in the database" if params_glpi.nil?
      
          
          user_token = params_glpi[:glpi_api_token]
          api_url = params_glpi[:glpi_api_url]
          app_token = params_glpi[:glpi_app_token]

          glpi_interaction = GlpiInteraction.new(user_token, api_url, app_token, user_id)
          

          content += glpi_interaction.execute
          is_job_ready = true
        rescue Exception => e
          e.backtrace.each do |line|
            content += "#{line}\n" 
          end
          e.set_backtrace([])
          logit "#{e.class}: #{e.message}" if verbose
          raise e
        ensure
          StoredDelayedJob.create(job: 'GlpiInventoryJob', user_id: user_id, content: content)
          if is_job_ready
            attempts = Delayed::Job.where(handler: self.to_yaml).pluck(:attempts).first || 0
            next_run_time = reschedule_at(Time.now, attempts)
            Delayed::Job.enqueue(self.class.new(user_id, verbose), run_at: next_run_time)
          end
        end
      end                

      def reschedule_at(current_time, attempts)
        glpi_params = get_params_glpi
        periodicity = glpi_params[:glpi_update_periodicity]
        update_time = glpi_params[:glpi_update_time]

        case periodicity
        when "weekly"
          hour, minute = update_time.split(":").map(&:to_i)
          current_time.beginning_of_week.advance(weeks: 1, hours: hour, minutes: minute)
        when "daily"
          hour, minute = update_time.split(":").map(&:to_i)
          current_time.beginning_of_day.advance(days: 1, hours: hour, minutes: minute)
        when "hourly"
          current_time.advance(hours: 1)
        else
          current_time.advance(hours: 12)  # Default if periodicity is unknown
        end
      end

      private

      def get_params_glpi
        glpi_strategy = Integrations::GlpiStrategy.new
        params = glpi_strategy.saved_params
        required_keys = ["glpi_api_token", "glpi_api_url", "glpi_app_token",
                         "glpi_update_periodicity", "glpi_update_time"]
        required_keys.each do |key|
          raise StandardError, "Error: Missing required GLPI parameter: #{key}" unless params[key.to_sym]
        end
        params
      end

    end

  end

end
