#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
require 'syslog'

module Redborder
  module Jobs

    class IncidentToSyslogJob < Struct.new(:user_id, :verbose)
      attr_accessor :content
      include Redborder::Jobs::JobHelper
      include IncidentsHelper

      def perform
        content = "Node executing this job: #{`hostname`.chomp!}\n\n"
        syslog_configured = false
        begin
          # Fetching the user by user_id, return if no user found
          user = User.find_by(id: user_id)
          return unless user

          syslog_configured = syslog_configured?

          # Get the params to avoid to continue
          unless syslog_configured
            content += 'Syslog is not configured yet\n'
          else          
            incidents = Incident.where(visible: true, sent: false)
            content += "Incidents to sent: #{incidents.count}\n"
            search_params = {}

            # Processing each incident
            incidents.each do |incident|
              incident_data = get_incident_druid_data(incident, search_params, user)
              next unless incident_data && !incident_data.empty?
              content += "Sending incident with id #{incident.id}..\n"
              message = format_incident(incident, incident_data)
              send_to_rsyslog(message)
              incident.update(sent: true)
              content += "\nIncident with id #{incident.id} was sent.\n\n"
            end
          end
        rescue => e
          # Handle any errors during execution, log backtrace and raise the exception
          e.backtrace.each do |line|
            content += "#{line}\n"
          end
          e.set_backtrace([])
          logit "#{e.class}: #{e.message}" if verbose
          raise e
        ensure
          StoredDelayedJob.create(job: "IncidentToSyslogJob", user_id: user_id, content: content)

          if syslog_configured && DelayedJob.where('handler like ?',
            '%!ruby/struct:Redborder::Jobs::IncidentToSyslogJob%').count <= 1
            # Re-enqueue the job to run again after 10 minutes
            Delayed::Job.enqueue(self.class.new(user_id, verbose), run_at: Time.now + 10.minutes)
          end
        end
      end
      
      private

      # Formats the incident data into a structured message string
      #
      # @param incident [Incident] The incident object to be formatted
      # @param incident_data [Array] Additional data related to the incident
      # @return [String] A formatted string representing the incident
      def format_incident(incident, incident_data)
        uuid = incident.uuid
        name = incident.name
        priority_id = incident.priority_id
        incident_status_id = incident.incident_status_id
        description = incident.description

        sources = incident_data.map { |data| data[:source] }.uniq.join(', ')

        assets = []
        observables = []

        incident_data.each do |data|
          assets += data[:assets] if data[:assets]
          observables += data[:observables] if data[:observables]
        end
        
        assets.uniq!
        observables.uniq!

        assets_str = format_assets(assets)
        observables_str = observables.to_a.any? ? observables.map { |obs| obs.map { |key, type| "#{type}:#{key}" }.join(", ") }.join(', ') : ''

        # Format the incident message
        message = "redborder_ndr New incident detected: uuid=\"#{uuid}\", name=\"#{name}\", source=\"#{sources}\", priority_id=#{priority_id}, incident_status_id=#{incident_status_id}, description=\"#{description}\", assets=\"#{assets_str}\", observables=\"#{observables_str}\""
        
        message
      end

      # Sends the formatted message to the remote Syslog server via TCP or UDP
      #
      # @param message [String] The message to be sent to Syslog
      # @return [String] Returns the message if successfully sent
      def send_to_rsyslog(message)
        udp_socket = nil
        tcp_socket = nil
        begin
          # Fetch syslog configuration parameters
          params_syslog = get_params_syslog
          syslog_server = params_syslog[:syslog_server]
          syslog_port = params_syslog[:syslog_port] || 514
          syslog_protocol = params_syslog[:syslog_protocol] || 'udp'
          
          # Get the hostname of the machine executing the job
          hostname = Socket.gethostname
          timestamp = Time.now.strftime('%b %d %H:%M:%S')
          
          facility = 16  # Local use
          severity = 6   # Informational
          priority = (facility * 8) + severity
          
          # Format the complete syslog message
          syslog_message = "<#{priority}>#{timestamp} #{hostname} #{message}"
          
          # Send the message over TCP or UDP based on configuration
          if syslog_protocol.downcase == 'tcp'
            tcp_socket = TCPSocket.new(syslog_server, syslog_port)
            tcp_socket.write(syslog_message)
          else
            udp_socket = UDPSocket.new
            udp_socket.send(syslog_message, 0, syslog_server, syslog_port)
          end
        rescue => e
          Rails.logger.error "Error sending message to remote syslog: #{e.message}"
        ensure
          # Ensure sockets are closed properly
          udp_socket.close if udp_socket && !udp_socket.closed?
          tcp_socket.close if tcp_socket && !tcp_socket.closed?
        end
      end

      # Method to format the assets
      def format_assets(assets)
        return '' unless assets.to_a.any?

        assets.map do |ast|
          ast.map do |key, type|
            case type
            when "NetObject"
              "ip:#{key}"
            when "MacObject"
              "mac:#{key}"
            else
              "#{type}:#{key}"
            end
          end.join(", ")
        end.join(', ')
      end

      # Fetches Syslog configuration parameters from Syslog
      def get_params_syslog
        syslog_strategy = Integrations::SyslogStrategy.new
        params = syslog_strategy.saved_params

        # Check for the required syslog server parameter
        required_keys = ["syslog_server"]
        required_keys.each do |key|
          raise StandardError, "Error: Missing require Syslog parameter: #{key}" unless params[key.to_sym]
        end
        params
      end
    end
  end
end
