#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    # This jobs generates content: titles, description, etc. depending of the function
    class UpdateContentIncidentsJob < Struct.new(:verbose, :user_id)
      include Redborder::Jobs::JobHelper
      include IncidentsHelper

      def perform
        content = ''
        syslog_configured = false
        begin
          content += "Searching for non visible incidents..\n" if verbose
          incidents = Incident.all.where(visible: false)
          content += "#{incidents.count} incidents found.\n" if verbose

          if incidents.count > 0
            syslog_configured = syslog_configured?
          end

          incidents.each do |incident|
            content += "Getting data for incident #{incident.id}...\n" if verbose
            alerts = get_incident_druid_data(incident, {}, user_id)
            next unless alerts && !alerts.empty?
            
            # If redborder-ai is enabled
            if Redborder::AI.is_ai_service_enabled
              content += "\n\nredborder-ai service is enabled. Generating title and description for incident #{incident.id}\n" if verbose
              params = { alerts: alerts, incident_name: incident.name }
              
              title = Redborder::AI.generate_incident_title(params)
              
              if title.is_a?(String)
                incident.name = title[0, 255]
                content += "Title for incident #{incident.id}: #{title}\n\n" if verbose
                create_ai_generation_record('name', incident.id)
              else 
                content +="Error generating title for incident #{incident.id}\n" if verbose
                content +="Title error: #{title}\n\n" if verbose
              end
              
              description = Redborder::AI.generate_incident_description(params)
              if description.is_a?(String)
                incident.description = description[0, 1000]
                content += "Description for incident #{incident.id}: #{description}\n\n" if verbose
                create_ai_generation_record('description', incident.id)
              else
                content += "Error generating description for incident #{incident.id}\n" if verbose
                content += "Description error: #{description}\n" if verbose
              end
            else
              content += "redborder-ai service is not enabled. AI title and description will not be generated.\n" if verbose
            end

            content += "Making incident #{incident.id} visible\n" if verbose
            incident.visible = true

            if incident.save
              content += "Saving incident with ID: #{incident.id}\n" if verbose
              incident.update(failure_count: 0)
            else
              incident.increment!(:failure_count)
              content += "Error Saving incident #{incident.id}. Failure count: #{incident.failure_count}\n" if verbose
              content += "Validation errors: #{incident.errors.full_messages.join(', ')}\n" if verbose

              if incident.failure_count >= 12
                content += "Incident #{incident.id} has failed 12 times and will be deleted\n" if verbose
                incident.destroy
              end
            end
          end
        rescue Exception => e
          content += "#{e.class}: #{e.message}\n"
          e.backtrace.each do |line|
            content += "#{line}\n"
          end
          logit "#{e.class}: #{e.message}"
        ensure
          if syslog_configured && !DelayedJob.incident_to_syslog?
            content += "\nStarting IncidentToSyslogJob..\n"
            # Re-enqueue the job to run again after 10 minutes
            Delayed::Job.enqueue(Redborder::Jobs::IncidentToSyslogJob.new(user_id, verbose), run_at: Time.now + 2.minutes)
          end

          StoredDelayedJob.create(job: "UpdateContentIncidentsJob", user_id: user_id, content: content)
          
          if DelayedJob.where('handler like ?',
            '%!ruby/struct:Redborder::Jobs::UpdateContentIncidentsJob%').count <= 1

            # Enqueue the job again
            Delayed::Job.enqueue(
            Redborder::Jobs::UpdateContentIncidentsJob.new(verbose, user_id), priority: 10, run_at: Time.now + 5.minutes)
          end
        end
      end

      private

      # Method to create a new entry in ai_generation_records table
      def create_ai_generation_record(field_name, incident_id)
        AiGenerationRecords.create(
          table_name: 'incidents',
          field_name: field_name,
          record_id: incident_id,
          record_type: 'Incident'
        )
      rescue ActiveRecord::RecordInvalid => e
        Rails.logger.error "Failed to create AI generation record: #{e.message}"
      end
    end
  end
end
