#!/bin/bash

# print how to use the script and exit
f_usage() {
  echo -en "\033[0;36m"
  echo "usage: $0 -i <manager-ip>"
  echo -en "\033[0m"
  exit
}

#check if we have root permissions
f_is_root() {
  if [ $(whoami) != 'root' ]; then
    echo -en "\033[0;36m"
    echo "This script needs to be execute with root permissions, try with sudo"
    echo -en "\033[0m"
    exit 1 
  fi

}

# check if we are in a rails environment of redborder
f_check_environment() {
  RAILS_FILES=(Gemfile Gemfile.lock Rakefile config/database.yml)
  for file in "${RAILS_FILES[@]}" ; do
    if [ ! -e "$file" ] ; then
      echo -en "\033[0;36m"
      echo "You are not in a redborder rails project!!"
      echo -en "\033[0m"
      f_usage
    fi
  done
}

#
# Find a line that starts with an string. Return the index
#
# $1 : String that start the line
# $2 : Path to the file
#
f_find_line_starts () {
  # Indexes
  local index=1
  local delete=-1
  while read line
  do
    if [[ $line == "$1"* ]]; then
      delete=$index
    fi
    index=$(($index + 1))
  done < "$2"
  echo $delete
}

# retrieve the rsa key from the manager to the local machine so we
# can avoid the password for retrieving more files later
f_bring_rsa() {
  local config="/var/www/rb-rails/config"
  local rsa="$config/rsa"
  local rsapub="$config/rsa.pub"
  local files="$rsa $rsapub"
  for f in $files; do
    scp root@$manager_ip:"$f" ./config
  done
}

f_bring_files () {
  local config="/var/www/rb-rails/config"
  local aws="$config/aws.yml"
  local db="$config/database.yml"
  #local aerospike="$config/aerospike.yml"
  local license="$config/license.key.pub"
  local chef_config="$config/chef_config.yml"
  local druid="$config/rbdruid_config.yml"
  local encrypted_data_bag_secret="/etc/chef/encrypted_data_bag_secret"
  local client="/etc/chef/client.pem"
  local files="$aws $db $managers $license $aerospike $chef_config $druid $encrypted_data_bag_secret $client"
  for f in $files; do
    scp -i ./config/rsa root@$manager_ip:"$f" ./config
  done
}

# check if we can login into manager
f_check_connection_manager() {
  nc -w 5 -z $1 22
  if [ $? -ne 0 ]; then
    echo "cannot ssh into manager $1"
    exit 1
  fi
}

# get the hostname of the manager
f_get_hostname () {
  echo `ssh -i ./config/rsa root@$1 hostname`
}

# get the local system
f_get_SO() {
  echo `uname -s`
}

# MAIN
while getopts "hi:" opt; do
  case $opt in
    i) manager_ip="$OPTARG"
       ;;
    *) f_usage
       ;;
  esac
done

if [ -z $manager_ip ]; then
    f_usage
fi

f_is_root
f_check_environment
f_check_connection_manager $manager_ip

# Info about your local machine
YOUR_SO=$(f_get_SO)

echo "get rsa of manager"
f_bring_rsa $manager_ip

echo -n 'getting remote hostname... '
manager_hostname=$(f_get_hostname $manager_ip)
echo "$manager_hostname"

echo "get manager files to local machine"
f_bring_files $manager_ip

mkdir -p /etc/chef
echo "Moving encrypted_data_bag_secret to /etc/chef/encrypted_data_bag_secret"
mv config/encrypted_data_bag_secret /etc/chef/

echo "Moving client.pem to /etc/chef/client.pem"
mv config/client.pem /etc/chef/

hosts_route=""
resolve_route=""
if [ "$YOUR_SO" == "Darwin" ]; then
  # MAC system
  hosts_route="/private/etc/hosts"
  resolve_route="/etc/hosts"
else
  # Linux system
  hosts_route="/etc/hosts"
  resolve_route="/etc/resolv.conf"
fi

echo "apply changes to your hosts file ${hosts_route} and ${resolve_route}..."
START_COMMENT="##_RedBorder_##"
END_COMMENT="#_RedBorder_#"
LINE_NUMBER=$(f_find_line_starts "${START_COMMENT}" $hosts_route)
echo "line number is $LINE_NUMBER"

if [ $LINE_NUMBER -ne -1 ]; then
  sed -i -e "/$START_COMMENT/,/$END_COMMENT/d" $hosts_route
fi
echo "$START_COMMENT
$manager_ip  $manager_hostname" | tee -a $hosts_route
echo $END_COMMENT | tee -a $hosts_route

grep "redborder.cluster" $resolve_route
if [ $? -ne 0 ]; then
  sed -i '/^search/ s/$/ redborder.cluster/' $resolve_route 
fi

LINE_NUMBER=$(f_find_line_starts "${START_COMMENT}" $resolve_route)
if [ $LINE_NUMBER -ne -1 ]; then
  sed -i -e "/$START_COMMENT/,/$END_COMMENT/d" $resolve_route
fi
grep "$manager_ip" $resolve_route
if [ $? -ne 0 ]; then
  sed -i "/^nameserver/ {
    i $START_COMMENT
    i nameserver $manager_ip
    i $END_COMMENT
    :a;n;ba
  }" $resolve_route
fi
echo ""

echo "connection config"
echo "================="
echo "IP: $manager_ip"
echo "HOSTNAME: $manager_hostname"
echo ""
echo -en "\033[0;34m"
echo ""
cat <<EOF
  ______ _       _     _         _____
 |  ____(_)     (_)   | |      _|  __ \\
 | |__   _ _ __  _ ___| |__   (_) |  | |
 |  __| | | '_ \| / __| '_ \    | |  | |
 | |    | | | | | \__ \ | | |  _| |__| |
 |_|    |_|_| |_|_|___/_| |_| (_)_____/
EOF
echo -en "\033[0m"
echo -en "\033[0;36m"
echo ""
echo "run following command in the manager $manager_ip to allow access from you local machine"
echo "    - firewall-cmd --permanent --zone=public --add-rich-rule='rule family=\"ipv4\" source address=\"<your local ip>/32\" accept' ; firewall-cmd --reload"
echo "    - rb_grant_access -n <your local ip>/32"
echo ""
echo "You save 2 minutes with this script. Go to play ping-pong now!"
echo -en "\033[0m"
echo ""
