require 'test_helper'

#
# Test dashboards controller
#
class DashboardsControllerTest < ActionController::TestCase
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'Overview' do
    setup do
      @user = users(:admin)
      sign_in @user
      @sensor = @user.domains_with_permission.last
      @client_proxy = sensors(:client_proxy)
    end

    should 'see it' do
      get :overview
      assert_response :success
      # Check initalized variables
      %i(domain_zones sensor range reload).each do |var|
        assert_not_nil assigns(var)
      end

      assert_equal('last_1_hour', assigns(:range),
                   'By default, range is last_1_hour')

      overview_widgets_partials.each do |partial|
        assert_template partial
      end
    end

    should 'change the range' do
      get :overview, range: :last_1_hour
      assert_equal('last_1_hour', assigns(:range))
    end

    should 'only allow some ranges' do
      get :overview, range: :last_3_hour
      assert_equal('last_1_hour', assigns(:range))
    end

    should 'see overview of a sensor' do
      get :overview, id: @sensor.id
      assert_response :success
      # Check initalized variables
      %i(domain_zones sensor range reload).each do |var|
        assert_not_nil assigns(var)
      end
      # The sensor must be initialized
      assert_equal @sensor.id, assigns(:sensor).id
    end

    should 'store search type' do
      post :overview_store_search_type, search_type: 'places'
      assert_equal 'places', session[:overview_search_type]
    end
  end

  context 'Overview for a sensor building' do
    setup do
      @user = users(:admin)
      sign_in @user
      @sensor = @user.domains_with_permission
                     .where(domain_type: Sensor::BUILDING).first
    end

    should 'see it' do
      get :overview, id: @sensor.id
      assert_response :success

      # The sensor must be initialized
      assert_equal @sensor.id, assigns(:sensor).id
    end
  end

  context 'A user without dashboard' do
    setup do
      @user = users(:user_domain)
      sign_in @user
    end

    should 'create a dashboard automatically' do
      assert @user.dashboards.blank?

      assert_difference('@user.dashboards.count', 1) do
        get :index
      end
    end

    should 'be redirected to the overview view' do
      get :index
      assert_response :redirect
    end
  end

  context 'A dashboard owner' do
    setup do
      @user = users(:admin)
      sign_in @user
      @dashboard = dashboards(:admin_default)
    end

    should 'see dashboards list and a particular one' do
      get :index
      assert_redirected_to @dashboard

      get :show, id: @dashboard.id
      assert_response :success
    end

    should 'create a dashboard' do
      get :new
      assert_response :success

      assert_difference('Dashboard.count') do
        post :create,
             dashboard: { name: 'New Dashboard', default: true,
                          owner: @user, layout: 'columns' }
      end
      assert_response :found
      assert flash[:notice].include? "#{I18n.t 'controllers.dashboard_created'}"
    end

    should 'clone a dashboard' do
      get :clone, id: @dashboard.id
      assert_response :success

      assert_difference('Dashboard.count') do
        post :create_clone, id: @dashboard.id,
             dashboard: { name: 'Cloned Dashboard'}
      end
      assert_redirected_to assigns(:new_dashboard)
      assert flash[:notice].include? "#{I18n.t 'controllers.dashboard_created'}"
    end

    should 'update a dashboard' do
      patch :update, id: @dashboard.id,
            dashboard: { name: 'New name' }
      assert_equal 'New name', @dashboard.reload.name
      assert_redirected_to @dashboard
    end

    should 'not update a dashboard with invalid params' do
      patch :update, id: @dashboard.id,
            dashboard: { name: "" }
      assert_not_equal "", @dashboard.reload.name
      assert_template :edit
    end

    should 'destroy a dashboard' do
      assert_difference('Dashboard.count', -1) do
        delete :destroy, id: @dashboard.id
      end
      assert_redirected_to dashboards_path
    end

    should 'import a dashboard with a valid file' do
      get :import
      assert_response :success

      post :save_import
      assert_redirected_to assigns(:dashboard)
      assert_not flash[:notice].include? "#{I18n.t 'controllers.dashboard_created'}" if flash[:notice]

      post :save_import, dashboard: { file: fixture_file_upload('files/test_import_dashboard.tar.gz')}
      assert_redirected_to assigns(:dashboard)
      assert flash[:notice].include? "#{I18n.t 'controllers.dashboard_created'}"
    end

    should 'load an image to the dashboard' do
      get :image, id: @dashboard.id
      assert_response :success
    end

    should 'remove image from the dashboard' do
      xhr :patch, :remove_image, format: :js, id: @dashboard.id
      assert_response :success
    end

    should 'set background position' do
      post :set_background_position, id: @dashboard.id, x: 1, y: 2
      assert_response 200
    end

    should 'set locked status' do
      post :set_locked_status, id: @dashboard.id, locked: true
      assert_response 200
    end

    should 'add new sensor to the dashboard' do
      get :new_sensor
      assert_response :success
    end

    should 'add new user to the dashboard' do
      get :new_user
      assert_response :success
    end

    should 'convert the dashboard to report' do
      post :convert_to_report, id: @dashboard.id
      assert_response :redirect
    end

    should 'export the dashboard' do
      get :export, id: @dashboard.id
      assert_response :success
    end

  end

  context 'A non dashboard owner' do
    setup do
      @user = users(:admin_domain)
      sign_in @user
      @dashboard = dashboards(:admin_default)
    end

    should 'destroy the dashboard if has permissions' do
      assert_no_difference('Dashboard.count') do
        delete :destroy, id: @dashboard.id
      end
      assert_redirected_to dashboards_path
    end

    should 'add new sensor to the dashboard' do
      get :new_sensor
      assert_response :success
    end
  end

  context 'An admin user' do
    setup do
      @user = users(:admin)
      sign_in @user
    end

    should 'import a dashboard even if the widgets to add havent got the proper structure' do
      fake_import = fixture_file_upload('files/Bandwidth.tar.gz', 'application/gzip')
      assert_difference('Dashboard.count') do
        post :save_import, dashboard: { file: fake_import }
      end
      assert_redirected_to assigns(:dashboard)
    end
  end

  context 'An user' do
    setup do
      @user = users(:user1)
      @dashboard_shared = dashboards(:admin_abs_dashboard)
      @dashboard_notshared = dashboards(:admin_col_dashboard)
      @dashboard_locked = dashboards(:locked_shared_dashboard)
      sign_in @user
    end

    should 'clone a shared dashboard from another user and set himself as the owner' do
      assert_difference('Dashboard.count') do
        post :create_clone, id: @dashboard_shared.id, dashboard: { name: 'Shared dashboard' }
      end
      new_dashboard = Dashboard.find_by(name: 'Shared dashboard')
      assert_equal(new_dashboard.owner, @user)
      assert_redirected_to dashboard_path(new_dashboard)
    end

    should 'not clone a shared dashboard from another user if he has not got read permissions' do
      assert_no_difference('Dashboard.count') do
        post :create_clone, id: @dashboard_notshared.id, dashboard: { name: 'Not shared dashboard' }
      end
      assert_redirected_to dashboards_path
      assert_match /You don't have enough privileges to do that./, flash[:alert]
    end

    should 'clone a shared dashboard from another user even if it has been locked and he has got read permissions' do
      assert_difference('Dashboard.count') do
        post :create_clone, id: @dashboard_locked.id, dashboard: { name: 'Shared locked dashboard' }
      end
      new_dashboard = Dashboard.find_by(name: 'Shared locked dashboard')
      assert_equal(new_dashboard.owner, @user)
      assert_redirected_to dashboard_path(new_dashboard)
    end

    should 'clone a shared dashboard from another user even after lock/unlock the dashboard' do
      assert_difference('Dashboard.count') do
        post :set_locked_status, id: @dashboard_shared.id, locked: 'true'
        post :set_locked_status, id: @dashboard_shared.id, locked: 'false'
        post :create_clone, id: @dashboard_shared.id, dashboard: { name: 'Shared dashboard' }
      end
      new_dashboard = Dashboard.find_by(name: 'Shared dashboard')
      assert_equal(new_dashboard.owner, @user)
      assert_redirected_to dashboard_path(new_dashboard)
    end
  end

  context 'An organization user' do
    setup do
      @org_admin = users(:admin_domain)
      @dashboard = dashboards(:organization_dashboard)
      sign_in @org_admin
    end

    should 'share its dashboard with a visible user' do
      org_user = users(:user_domain)

      assert_difference '@dashboard.users.count' do
        post :update, id: @dashboard.id, dashboard: { collaborators: [{ 'user_id' => org_user.id, 'can_edit' => 'on' }] }
      end
    end

    should 'not share its dashboard with user from root' do
      super_admin = users(:admin)

      assert_no_difference '@dashboard.users.count' do
        post :update, id: @dashboard.id, dashboard: { collaborators: [{ 'user_id' => super_admin.id, 'can_edit' => 'on' }] }
      end
    end

    should 'share its dashboard with market domain' do
      market = sensors(:market)

      assert_difference '@dashboard.sensors.count' do
        post :update, id: @dashboard.id, dashboard: { sensors_collaborators: {'0' => { 'sensor_id' => market.id, 'can_edit' => 'none' } } }
      end
    end

    should 'not share its dashboard with root domain' do
      root = sensors(:root)

      assert_no_difference '@dashboard.sensors.count' do
        post :update, id: @dashboard.id, dashboard: { sensors_collaborators: {'0' => { 'sensor_id' => root.id, 'can_edit' => 'none' } } }
      end
    end

    should 'not get root\'s overview' do
      root = sensors(:root)
      get :overview, id: root.id
      assert_response :not_found
    end
  end

  # Return all overview widgets
  #
  def overview_widgets_partials
    partials = []
    Redborder::Module.registered_modules.each do |_, prop|
      overview_widgets = prop[:overview_widgets]
      next unless overview_widgets.present?
      # Iterate over widgets
      overview_widgets.each do |w|
        next unless w[:partial]
        partials << "overview/_#{w[:partial]}"
      end
    end
    # Return partials
    partials
  end
end
